/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
/*
 * reader.cc
 * Copyright (C) 2021 MicroEM JSC <support@microem.ru>
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name ``MicroEM JSC'' nor the name of any other
 *    contributor may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 * 
 * CLESCAR IS PROVIDED BY MicroEM JSC ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL MicroEM JSC OR ANY OTHER CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "reader.h"

extern PVOID m_pReader;

Gtk::Button* btn_get_version;
Gtk::Button* btn_get_serial;
Gtk::Entry* edt_version;
Gtk::Entry* edt_serial;
Gtk::Button* btn_beep;
Gtk::SpinButton* spin_beeps;
Gtk::Button* btn_led;
Gtk::ComboBoxText* cmb_color;
Gtk::ComboBoxText* cmb_post_color;

void on_btn_get_version_clicked()
{
	LONG Status;
	int i;
	BYTE Buf[256] = {0};
	char out[256] = {0};
	
	printf("CLSCRF_Get_Mfrc_Version: ");
    Status = CLSCRF_Get_Mfrc_Version( m_pReader, Buf );
	int pos = 0;
    if( Status == 0 )
	{
		printf("OK\n");
        for( i = 0; i < 6; i++ )
            pos += snprintf(out + pos, sizeof(out) - pos, "%02X", Buf[i] );
	}
    else
        printf("Error 0x%08X\n", (UINT)Status );

	edt_version->set_text(out);
}

void on_btn_get_serial_clicked()
{
	LONG Status;
	int i;
	BYTE Buf[256] = {0};
	char out[256] = {0};
	
	printf("CLSCRF_Get_Mfrc_Serial_Number: ");
    Status = CLSCRF_Get_Mfrc_Serial_Number( m_pReader, Buf );
	int pos = 0;
    if( Status == 0 )
	{
		printf("OK\n");
        for( i = 0; i < 4; i++ )
            pos += snprintf(out + pos, sizeof(out) - pos, "%02X", Buf[3-i] );
		edt_serial->set_text(out);
	}
    else
        printf("Error 0x%08X\n", (UINT)Status );

}

void on_btn_beep_clicked()
{
	LONG Status;

	printf("CLSCRF_Sound: ");
    Status = CLSCRF_Sound( m_pReader, spin_beeps->get_value_as_int() );
	if( Status == 0 )
		printf("OK\n");
    else
        printf("Error 0x%08X\n", (UINT)Status);
}

void on_btn_led_clicked()
{
	LONG Status;

	printf("CLSCRF_Led: ");
    Status = CLSCRF_Led( 
                        m_pReader, 
                        cmb_color->get_active_row_number(),
                        spin_beeps->get_value_as_int(), 
                        cmb_post_color->get_active_row_number()
                        );
	if( Status == 0 )
		printf("OK\n");
    else
        printf("Error 0x%08X\n", (UINT)Status );
}
