// Transfer.cpp: implementation of the CTransfer class.
//
//////////////////////////////////////////////////////////////////////

#include "Transfer_unix.h"
#include "SCardErr.h"

#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <termios.h>
#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include <sys/ioctl.h>


#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#define EV_COMPLETE 5

#define LOG_FILE	"/var/log/rfid/tlog.log"

/* Для реализации тайм аутов */
void sig_alarm(int sig)
{
	/* Nop */
//	printf("SIGALARM\n");
}

//////////////////////////////////////////////////////////////////////
//                      class CTransceiveBuffer 
//////////////////////////////////////////////////////////////////////

CTransceiveBuffer::CTransceiveBuffer()
{
    m_Count = 0;
}

CTransceiveBuffer::~CTransceiveBuffer()
{
}

bool CTransceiveBuffer::AddByte( unsigned char Value )
{
    if( m_Count + 1 > TX_BUF_SIZE )
        return false;

    m_Buf[ m_Count ] = Value;
    m_Count++;

    return true;
}

bool CTransceiveBuffer::AddBulk( const unsigned char *Data, DWORD Size )
{
    bool bRet = false;

    if( m_Count + Size > TX_BUF_SIZE )
        Size = TX_BUF_SIZE - m_Count;
    else
        bRet = true;

    memcpy( &m_Buf[ m_Count ], Data, Size );
    m_Count += Size;

    return bRet;
}

bool CTransceiveBuffer::GetBulk( int nIndex, unsigned char *Data, DWORD *pSize )
{
    bool bRet = false;
    DWORD CpySize = *pSize;

    if( nIndex + *pSize > m_Count )
        CpySize = TX_BUF_SIZE - nIndex;

    if( CpySize && CpySize == *pSize )
        bRet = true;

    if( CpySize )
        memcpy( Data, &m_Buf[ nIndex ], CpySize );

    *pSize = CpySize;
    return bRet;
}

BYTE CTransceiveBuffer::operator []( int nIndex ) const
{
    if( nIndex > TX_BUF_SIZE )
        return 0;
    else
        return m_Buf[ nIndex ];
}

//static CTransfer* lpTr;

// 10:10:50.646 R     00 B6 AB F4 B9 49 4E 63 2A 10 B4 8E
//                 FD 00 B6 AB F4 B9 49 4E 63 2A 10 B4 8E 00 FE 
//////////////////////////////////////////////////////////////////////
// Construction/Destruction     class CTransfer
//////////////////////////////////////////////////////////////////////
CTransfer::CTransfer()
{
//#if ADDRESS_RS_485
    m_IO_Addr = 0;
//#endif
    m_Cmd_ID = 0;
    m_Timeout = DEFAULT_IO_TIMEOUT;
    m_PortNumber = -1;
    m_Baudrate = 0;
    fd = -1;
    IsSystemDriver = false;
//    remove(LOG_FILE);
}

CTransfer::~CTransfer()
{
    Close();
}

void CTransfer::_setup_timeout()
{
	struct sigaction act;
	memset(&act, 0, sizeof(act));
	act.sa_handler = sig_alarm;
	act.sa_flags = SA_ONESHOT;
	if (sigaction(SIGALRM, &act, 0) != 0)
	{
		printf("Could not set up timeout\n");
	}
	else
	{
		/* seting up timeout */
		//alarm(TIMEOUT_SEC);
		struct itimerval old, new_;
		new_.it_interval.tv_usec = 0;
		new_.it_interval.tv_sec = 0;
		new_.it_value.tv_usec = 0;
		new_.it_value.tv_sec = (long int) TIMEOUT_SEC;
		if (setitimer (ITIMER_REAL, &new_, &old) < 0)
		{
			printf("Could not set up timeout\n");
		}
	}
}

/* Baud задавать в целочисленном значении скорости */
bool CTransfer::Open()
{
  bool bResult = false;

  libusb_init(&libusb_ctx);
  dev_handle = libusb_open_device_with_vid_pid(libusb_ctx, USB_VID, USB_PID);

  if (dev_handle)
  {
	// Begin insert - Vadim Isaev, 21.07.2017
	if(libusb_kernel_driver_active(dev_handle, 0))
	{
	  IsSystemDriver = true;
	  libusb_detach_kernel_driver(dev_handle, 0);
	}
	libusb_claim_interface(dev_handle, 0);
	// End insert
	
	bResult = true;
  }
  else
  {
	dev_handle = NULL;
  }
  
  return bResult;
}

bool CTransfer::Close( )
{
	//Begin insert - Vadim Isaev, 21.07.2017
	libusb_release_interface(dev_handle, 0);
	if (IsSystemDriver)
	{
//	  libusb_attach_kernel_driver(dev_handle, 0);
	}
	// End insert
	libusb_close(dev_handle);
	libusb_exit(libusb_ctx);
	dev_handle = NULL;
    
    return true;
}

void CTransfer::RX_Reset( void )
{
    m_state.n = 0;  // в т.ч. m_state.b.RX_stuff = 0;
    m_RX_cnt = 0;
    m_RX_FCS = 0xFFFF;
}

void CTransfer::CalculateFCS( unsigned short &FCS, unsigned char b )
{
    unsigned char i;
    unsigned short w;

    w = ( b ^ FCS ) & 0xFF;
    i = 8;
    do{
        if( w & 1 ) 
        {
            w >>= 1;
            w ^= 0x8408;
        } 
        else 
        {
            w >>= 1;
        }
    } while( --i );
    FCS = w ^ ( FCS >> 8 );
}

void CTransfer::PostByte( unsigned char b )
{
    CalculateFCS( m_TX_FCS, b );
    ByteStuff( b );
}

void CTransfer::ByteStuff( unsigned char b )
{
    if( b < B_SPECIAL )
        m_TX_buf[m_TX_cnt++] = b;
    else                            // ByteStuffing
    {      
        m_TX_buf[m_TX_cnt++] = B_STUFF;
        m_TX_buf[m_TX_cnt++] = B_STUFF - b;
    }
}

char CTransfer::OnReceive( unsigned char inp )
{
    unsigned char b;
/*
	if( m_Timer )
	{
		KillTimer( NULL, m_Timer );
		m_Timer = 0;
	}
*/
    switch( inp )
    {
    case B_START:  // Frame begin
        RX_Reset();
        m_state.b.RX_start = 1;
        break;

    case B_STOP:  // Frame end
/*
Для формирования ответа считывателем должны быть приняты как минимум 
стартовый и стоповый байты, идентификатор кадра, код команды и FCS. 
Также должны отсутствовать ошибки байтстаффинга в пределах кадра. 
Иначе никакие ответы не формируются и мастер должен повторить запрос 
по окончании таймаута, выставив соответствующий признак повторного запроса.
*/

        if( m_state.b.RX_start 
            && ( m_RX_cnt >= 4 ))//идентификатор кадра(1), код команды(1) и FCS(2) 
        {
            m_RX_FCS ^= 0xFFFF;
            if( m_RX_FCS == 0x0F47 )
            {
                return 0;
            }
            else
            {
/*
Ответ NACK 1 посылается в случае ошибки FCS. Единственная цель 
такого ответа - ускорить повтор запроса до окончания таймаута.
*/
//                Acknowledgment( NACK1_FCS_ERROR );
                return 2;
            }
        }
        else
        {
            RX_Reset();
            return 3;
        }
        break;

    case B_STUFF:  // Byte stuffing
        if( m_state.b.RX_start )
            m_state.b.RX_stuff = 1;
        break;

    default:    // Usual bytes
        if( ! m_state.b.RX_start )
		{
            return 4;
		}

        if( m_state.b.RX_stuff )
        {
            if( inp > 2 )       // Ошибка байтстаффинга
            {
                RX_Reset();
                return 5;
            }
            m_state.b.RX_stuff = 0;
            b = B_STUFF - inp;
        }
        else
        {
            b = inp;
        }

//#if ADDRESS_RS_485
        if( m_state.b.RX_addr )
        {
        m_RX_buf[m_RX_cnt++] = b;
        }
        else
        {
            if( m_IO_Addr && m_IO_Addr != b )
            {
                RX_Reset();
                return 7;           // оНКСВЕМ НРБЕР НР ОНЯРНПНММЕЦН СЯРПНИЯРБЮ
            }
            m_state.b.RX_addr = 1;
        }
//#else
//            m_RX_buf[m_RX_cnt++] = b;
//#endif
/*
Считыватель вправе игнорировать запрос если количество байт между стартовым 
и стоповым условиями больше чем оговорено самой большой командой.
*/
        if( m_RX_cnt < RX_BUF_SIZE )
        {
            CalculateFCS( m_RX_FCS, b );
        }
        else
        {
            RX_Reset();
            return 6;
        }

    }
    return 1;
}

void CTransfer::OnRX( int Port, BYTE *RXBuff, int Size )
{
    char Error;
    for( int i = 0; i < Size; i++ )
    {
        Error = OnReceive( RXBuff[i] );
        if( Error > 1 )
        {
//            if( m_lpfnError ) 
//                (*m_lpfnError)( Port, SERN_EVENT, Error, m_NotifContext );
            OnError( Port, SERN_EVENT, Error );
            return;
        }
    }
    if( ! Error )
    {
//        if( m_lpfnRX ) 
//            (*m_lpfnRX)( Port, m_RX_buf, m_RX_cnt - 2, m_NotifContext );
        //SetEvent( m_EvIOComplete );
        SetEvent( EV_COMPLETE  );
//        m_state.b.RX_OK = 1;
    }
}

void CTransfer::OnError( int Port, UINT Oper, DWORD Err )
{
//    if( m_lpfnError )
//        (*m_lpfnError)( Port, Oper, Err, m_NotifContext );
//        str.Format( "Порт COM%d   OK", Port+1 );
//	CString str, so, se;
    if( !( (Oper == SERN_EVENT) && (Err == 1) ) )
    {
/*
        switch( Oper )
        {
        case SERN_EVENT:
            so = "EVENT";
            break;

        case SERN_READ:
            so = "READ";
            break;

        case SERN_WRITE:
            so = "WRITE";
            break;

        default:
            so = "NO";
        }

        switch( Err )
        {
        case CE_RXOVER:
            se = "CE_RXOVER";
            break;

        case CE_OVERRUN:
            se = "CE_OVERRUN";
            break;

        case CE_RXPARITY:
            se = "CE_RXPARITY";
            break;

        case CE_FRAME:
            se = "CE_FRAME";
            break;

        case CE_TXFULL:
            se = "CE_TXFULL";
            break;

        case CE_MODE:
            se = "CE_MODE";
            break;

        case ERROR_TIMEOUT:
            se = "ERROR_TIMEOUT";
            break;

        default:
            se.Format( "0x%08X", Err );
        }
//        str.Format( "Порт COM%d   %s   Ошибка    %s", Port+1, so, se );
*/
        m_Error = Err;
//        m_state.b.RX_Err = 1;
    }
//    CCardPage* pPage = (CCardPage*)GetActivePage();
//    pPage->m_StatusBar.SetWindowText( str );
//    m_pCurPage->m_StatusBar.SetWindowText( str );
}
/*
//void CTransfer::SendCommand( unsigned char Oper, unsigned char Size, unsigned char *Param )
void CTransfer::Transmit( unsigned char *Data, int Size )
{
//    m_Cmd_op = Data[0];
    m_TX_cnt = 0;
    m_TX_FCS = 0xffff;
    m_TX_buf[m_TX_cnt++] = B_START;  // Frame begin
    PostByte( m_Cmd_ID++ );
    for( int i = 0; i < Size; i++ )
        PostByte( Data[i] );
    m_TX_FCS ^= 0xffff;
    ByteStuff( m_TX_FCS & 0xFF );
    ByteStuff( ( m_TX_FCS >> 8 ) & 0xFF );
    m_TX_buf[m_TX_cnt++] = B_STOP;  // Frame end
//    m_ComPort.Write( m_TX_buf, m_TX_cnt );
	write(fd, m_TX_buf, m_TX_cnt);

}
*/
//////////////////////////////////////////////////////////////////////
LONG CTransfer::SingleTransceive()
{
    m_Error = 0;
    LONG Status = 0;

    if( ! IsOpened() )
    {
        Status = SCARD_E_NO_READERS_AVAILABLE;
        return Status;
    }

    if( m_TrBuf.GetLength() == 0 )
    {
        Status = SCARD_E_NOT_TRANSACTED;
//        memcpy( pbRecvBuffer, &Status, *pcbRecvLength );
        return Status;
    }

    m_RcBuf.Clear();

    {
        m_TX_cnt = (BYTE)m_TrBuf.GetLength();
        DWORD Count = m_TX_cnt;

	/* Записываем  данные в устройство */
	m_TX_cnt = 0;
        m_TX_FCS = 0xffff;
        m_TX_buf[m_TX_cnt++] = B_START;  // Frame begin
        PostByte( m_IO_Addr );
        PostByte( m_Cmd_ID++ );
        Count = m_TrBuf.GetLength();
        for( DWORD i = 0; i < m_TrBuf.GetLength(); i++ )
            PostByte( m_TrBuf[i] );
        m_TX_FCS ^= 0xffff;
        ByteStuff( m_TX_FCS & 0xFF );
        ByteStuff( ( m_TX_FCS >> 8 ) & 0xFF );
        m_TX_buf[m_TX_cnt++] = B_STOP;  // Frame end

		int transferred;
		int ret = libusb_bulk_transfer(dev_handle, USB_EP_WRITE, m_TX_buf, m_TX_cnt, &transferred, 0);
		
		event = 0;
		m_Error = 0;
		m_RX_cnt = 0;
		BYTE buf[RX_BUF_SIZE];
		/* считываем */
		while (true)
		{
			int ret = libusb_bulk_transfer(dev_handle, USB_EP_READ, buf, RX_BUF_SIZE, &transferred, TIMEOUT_SEC * 1000);
			if (!ret)
			{
				OnRX(m_PortNumber, buf, transferred);
			} else
			{
				/* Сработал таймер  или произошла ошибка */
				/*if (errno == EINTR) */
				m_Error = ERROR_TIMEOUT;
				break;
			}
			
			if (m_Error)
			{
				/* Таймаут или ошибка из OnRX */
				break;
			}

			if (event == EV_COMPLETE)
			{
				/* Прочли до конца */
				break;
			}
		}		

        if( m_Error )               // ( COM-port & Protocol ) Errors (4 bytes)
        {
            Status = m_Error;
        } 
        else
        {
            if( m_RX_buf[2] )       // ( Reader & Card ) Errors (1 byte)
            {
                Status = SCARD_F_INTERNAL_ERROR;
                m_RcBuf.AddByte( m_RX_buf[2] );
            }
            else
            {
                m_RcBuf.AddBulk( &m_RX_buf[2], m_RX_cnt - 4 );
            }
        }
    }
   

    return Status;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void CTransfer::SetTimeout( DWORD NewTimeout )
{
    if( NewTimeout )
        m_Timeout = NewTimeout; 
    else
        m_Timeout = DEFAULT_IO_TIMEOUT; 
}

void CTransfer::SetEvent(int ev)
{
	event = ev;
}


//**********************************************
// Begin insert - Vadim Isaev, 25.07.2017
//**********************************************

// Антенна издаёт звуки
// Входные параметры:
//	nBeepCount - количество звуков
// Возвращаемое значение:
//	0 - отсутствие ошибки, см. SCardError.h
LONG CTransfer::Sound(BYTE nBeepCount)
{
  
    m_TrBuf.Clear();
    m_TrBuf.AddByte( 0x05 );
    m_TrBuf.AddByte( nBeepCount );

    Status = SingleTransceive();

    return Status;
}

// Антенна подмигивает.
// Входные параметры:
//	bBlinkColor - цвет подмигивания;
//	bBlinkCount - количество подмигиваний;
//	bPostColor - цвет после подмигивания.
// Возвращаемое значение:
//	0 - отсутствие ошибки, см. SCardError.h
LONG CTransfer::Led(BYTE bBlinkColor, BYTE bBlinkCount, BYTE bPostColor)
{
    m_TrBuf.Clear();
    m_TrBuf.AddByte( 0x07 );
    m_TrBuf.AddByte( bBlinkColor );
    m_TrBuf.AddByte( bBlinkCount );
    m_TrBuf.AddByte( bPostColor );

    Status = SingleTransceive();

    return Status;
}

// Получение номера карточки.
// Возвращаемое значение:
//	NIL - номера карточки нет,
//	char[17] - номер карточки в шестнадцатиричном виде. Последний байт - '\0'
char * CTransfer::GetNumber()
{
    char * CardNumber = NULL;
    
    BYTE    bFlags = 0x26;
    BYTE    bInventory  = 1;
    BYTE    bAfi = 0;
    BYTE    bMaskLen = 0;
    LPCBYTE pbMaskVal = NULL;
    LPBYTE  pbRecvBuffer = new BYTE[TX_BUF_SIZE];
    LPDWORD pdwRecvLength = new DWORD;
    *pdwRecvLength = TX_BUF_SIZE;

    m_TrBuf.Clear();
    m_TrBuf.AddByte( 0x31 );
    m_TrBuf.AddByte( bFlags );
    m_TrBuf.AddByte( bInventory );
    m_TrBuf.AddByte( bAfi );
    m_TrBuf.AddByte( bMaskLen );
    int SendLength = ( bMaskLen + 7 ) / 8;
    m_TrBuf.AddBulk( pbMaskVal, SendLength );

    Status = SingleTransceive();

    if( Status == SCARD_S_SUCCESS )
    {
        DWORD RecvLength = 0;
        DWORD Size = m_RcBuf.GetLength();
        if( Size > *pdwRecvLength )
        {
            *pdwRecvLength = 0;
            return NULL;
        }

        Size = 2;
        *pdwRecvLength = 0;
        if( ! m_RcBuf.GetBulk( 1, (LPBYTE)pdwRecvLength, &Size ) )
        {
            *pdwRecvLength = 0;
            if( Status == SCARD_S_SUCCESS )
                return NULL;
        }

        if( pdwRecvLength )
        {
            if( ! m_RcBuf.GetBulk( 3, pbRecvBuffer, pdwRecvLength ) )
                Status = SCARD_E_COMM_DATA_LOST;
        }
    }
    else
        *pdwRecvLength = 0;

    if( Status == SCARD_S_SUCCESS )
    {
        CardNumber = new char[17];
        //Преобразование массива байт номера карточки в строковый вид
        sprintf(CardNumber, "%02X%02X%02X%02X%02X%02X%02X%02X", 
        	pbRecvBuffer[11], pbRecvBuffer[10], pbRecvBuffer[9], 
        	pbRecvBuffer[8], pbRecvBuffer[7], pbRecvBuffer[6], 
        	pbRecvBuffer[5], pbRecvBuffer[4]);
        Sound(1);
    }
    delete[] pbRecvBuffer;
    delete pdwRecvLength;

    return CardNumber;
}

// Включение антенны.
// Возвращаемое значение:
//	0 - отсутствие ошибки, см. SCardError.h
LONG CTransfer::ReaderOn()
{
   
    m_TrBuf.Clear();
    m_TrBuf.AddByte( 0x10 );

    Status = SingleTransceive();

    return Status;
}

// Выключение антенны.
// Возвращаемое значение:
//	0 - отсутствие ошибки, см. SCardError.h
LONG CTransfer::ReaderOff()
{
   
    m_TrBuf.Clear();
    m_TrBuf.AddByte( 0x04 );
    m_TrBuf.AddByte( 0x80 );
    m_TrBuf.AddByte( 0x01 );

    Status = SingleTransceive();
    
    return Status;
}

// задание режима считывания карточек ISO\IEC 15693
// Возвращаемое значение:
//	0 - отсутствие ошибки, см. SCardError.h
LONG CTransfer::SetReaderMode(BYTE bRfMode)
{
    
    m_TrBuf.Clear();
    m_TrBuf.AddByte( 0x51 );
    m_TrBuf.AddByte( bRfMode );

    Status = SingleTransceive();
    
    return Status;
}
// Запись сообщения с  датой-временем в лог-файл
// Входные параметры:
//	Msg - текст сообщения.
//	Err - номер ошибки
void CTransfer::WriteLog(const char * Msg, long int Err)
{
    char * timebuffer = GetCurrentDateTime();

    f = fopen(LOG_FILE, "a");
    fputs(timebuffer, f);
    fputs(" ", f);
    fputs(Msg, f);
    fputs("\n", f);
    fprintf(f, ". Ошибка: %lx\n", Err);
    fclose(f);
}

// Получение текущей даты-времени в текстовом виде
// Возвращаемое значение:
//	char[20] в формате "ГГГГ-ММ-ДД чч:мм:сс"
char * CTransfer::GetCurrentDateTime()
{
    time_t t = time(NULL);
    struct tm *tt = localtime(&t);
    char * timebuffer;

    timebuffer = new char[20];
    strftime (timebuffer,20,"%Y-%m-%d %X ",tt);
    
    return timebuffer;
}

// Получение текста об ошибке по её номеру
// Входные параметры:
//	Err - номер ошибки
// Возвращаемое значение:
//	Текст ошибки соответствующий номеру
char * GetErrorMessage(long int Err)
{
  char * Message = new char[100];
  
  strcpy(Message, "And hell knows he's ...\n");
  
  switch (Err)
  {
    case NO_ERROR:
        strcpy(Message, "No error.\n");
        break;
    case ERROR_TIMEOUT:
	strcpy(Message , "Timeout expired.\n");
	break;
    case SCARD_F_INTERNAL_ERROR:
        strcpy(Message , "An internal consistency check failed.\n");
        break;
    case SCARD_E_CANCELLED:
        strcpy(Message , "The action was cancelled by an SCardCancel request.\n");
        break;
    case SCARD_E_INVALID_HANDLE:
        strcpy(Message , "The supplied handle was invalid.\n");
        break;
    case SCARD_E_INVALID_PARAMETER:
        strcpy(Message , "One or more of the supplied parameters could not be properly interpreted.\n");
        break;
    case SCARD_E_INVALID_TARGET:
        strcpy(Message , "Registry startup information is missing or invalid.\n");
        break;
    case SCARD_E_NO_MEMORY:
        strcpy(Message , "Not enough memory available to complete this command.\n");
        break;
    case SCARD_F_WAITED_TOO_LONG:
        strcpy(Message , "An internal consistency timer has expired.\n");
        break;
    case SCARD_E_INSUFFICIENT_BUFFER:
        strcpy(Message , "The data buffer to receive returned data is too small for the returned data.\n");
        break;
    case SCARD_E_UNKNOWN_READER:
        strcpy(Message , "The specified reader name is not recognized.\b");
        break;
    case SCARD_E_TIMEOUT:
        strcpy(Message , "The user-specified timeout value has expired.\n");
        break;
    case SCARD_E_SHARING_VIOLATION:
        strcpy(Message , "The smart card cannot be accessed because of other connections outstanding.\n");
        break;
    case SCARD_E_NO_SMARTCARD:
        strcpy(Message , "The operation requires a Smart Card, but no Smart Card is currently in the device.\n");
        break;
    case SCARD_E_UNKNOWN_CARD:
        strcpy(Message , "The specified smart card name is not recognized.\n");
        break;
    case SCARD_E_CANT_DISPOSE:
        strcpy(Message , "The system could not dispose of the media in the requested manner.\n");
        break;
    case SCARD_E_PROTO_MISMATCH:
        strcpy(Message , "The requested protocols are incompatible with the protocol currently in use with the smart card.\n");
        break;
     case SCARD_E_NOT_READY:
         strcpy(Message , "The reader or smart card is not ready to accept commands.\n");
         break;
     case SCARD_E_INVALID_VALUE:
         strcpy(Message , "One or more of the supplied parameters values could not be properly interpreted.\n");
         break;
     case SCARD_E_SYSTEM_CANCELLED:
         strcpy(Message , "The action was cancelled by the system, presumably to log off or shut down.\n");
         break;
     case SCARD_F_COMM_ERROR:
         strcpy(Message , "An internal communications error has been detected.\n");
         break;
     case SCARD_F_UNKNOWN_ERROR:
         strcpy(Message , "An internal error has been detected, but the source is unknown.\n");
         break;
     case SCARD_E_INVALID_ATR:
         strcpy(Message , "An ATR obtained from the registry is not a valid ATR string.\n");
         break;
     case SCARD_E_NOT_TRANSACTED:
         strcpy(Message , "An attempt was made to end a non-existent transaction.\n");
         break;
     case SCARD_E_READER_UNAVAILABLE:
         strcpy(Message , "The specified reader is not currently available for use.\n");
         break;
     case SCARD_P_SHUTDOWN:
         strcpy(Message , "The operation has been aborted to allow the server application to exit.\n");
         break;
     case SCARD_E_PCI_TOO_SMALL:
         strcpy(Message , "The PCI Receive buffer was too small.\n");
         break;
     case SCARD_E_READER_UNSUPPORTED:
         strcpy(Message , "The reader driver does not meet minimal requirements for support.\n");
         break;
     case SCARD_E_DUPLICATE_READER:
         strcpy(Message , "The reader driver did not produce a unique reader name.\n");
         break;
     case SCARD_E_CARD_UNSUPPORTED:
         strcpy(Message , "The smart card does not meet minimal requirements for support.\n");
         break;
     case SCARD_E_NO_SERVICE:
         strcpy(Message , "The Smart card resource manager is not running.\n");
         break;
     case SCARD_E_SERVICE_STOPPED:
         strcpy(Message , "The Smart card resource manager has shut down.\n");
         break;
     case SCARD_E_UNEXPECTED:
         strcpy(Message , "An unexpected card error has occurred.\n");
         break;
     case SCARD_E_ICC_INSTALLATION:
         strcpy(Message , "No Primary Provider can be found for the smart card.\n");
         break;
     case SCARD_E_ICC_CREATEORDER:
         strcpy(Message , "The requested order of object creation is not supported.\n");
         break;
     case SCARD_E_UNSUPPORTED_FEATURE:
         strcpy(Message , "This smart card does not support the requested feature.\n");
         break;
     case SCARD_E_DIR_NOT_FOUND:
         strcpy(Message , "The identified directory does not exist in the smart card.\n");
         break;
     case SCARD_E_FILE_NOT_FOUND:
         strcpy(Message , "The identified file does not exist in the smart card.\n");
         break;
     case SCARD_E_NO_DIR:
         strcpy(Message , "The supplied path does not represent a smart card directory.\n");
         break;
     case SCARD_E_NO_FILE:
         strcpy(Message , "The supplied path does not represent a smart card file.\n");
         break;
     case SCARD_E_NO_ACCESS:
         strcpy(Message , "Access is denied to this file.\n");
         break;
     case SCARD_E_WRITE_TOO_MANY:
         strcpy(Message , "The smartcard does not have enough memory to store the information.\n");
         break;
     case SCARD_E_BAD_SEEK:
         strcpy(Message , "There was an error trying to set the smart card file object pointer.\n");
         break;
     case SCARD_E_INVALID_CHV:
         strcpy(Message , "The supplied PIN is incorrect.\n");
         break;
     case SCARD_E_UNKNOWN_RES_MNG:
         strcpy(Message , "An unrecognized error code was returned from a layered component.\n");
         break;
     case SCARD_E_NO_SUCH_CERTIFICATE:
         strcpy(Message , "The requested certificate does not exist.\n");
         break;
     case SCARD_E_CERTIFICATE_UNAVAILABLE:
         strcpy(Message , "The requested certificate could not be obtained.\n");
         break;
     case SCARD_E_NO_READERS_AVAILABLE:
         strcpy(Message , "Cannot find a smart card reader.\n");
         break;
     case SCARD_E_COMM_DATA_LOST:
         strcpy(Message , "A communications error with the smart card has been detected.  Retry the operation.\n");
         break;
     case SCARD_E_NO_KEY_CONTAINER:
         strcpy(Message , "The requested key container does not exist on the smart card.\n");
         break;
     case SCARD_W_UNSUPPORTED_CARD:
         strcpy(Message , "The reader cannot communicate with the smart card, due to ATR configuration conflicts.\n");
         break;
     case SCARD_W_UNRESPONSIVE_CARD:
         strcpy(Message , "The smart card is not responding to a reset.\n");
         break;
     case SCARD_W_UNPOWERED_CARD:
         strcpy(Message , "Power has been removed from the smart card, so that further communication is not possible.\n");
         break;
     case SCARD_W_RESET_CARD:
         strcpy(Message , "The smart card has been reset, so any shared state information is invalid.\n");
         break;
     case SCARD_W_REMOVED_CARD:
         strcpy(Message , "The smart card has been removed, so that further communication is not possible.\n");
         break;
     case SCARD_W_SECURITY_VIOLATION:
         strcpy(Message , "Access was denied because of a security violation.\n");
         break;
     case SCARD_W_WRONG_CHV:
         strcpy(Message , "The card cannot be accessed because the wrong PIN was presented.\n");
         break;
     case SCARD_W_CHV_BLOCKED:
         strcpy(Message , "The card cannot be accessed because the maximum number of PIN entry attempts has been reached.\n");
         break;
     case SCARD_W_EOF:
         strcpy(Message , "The end of the smart card file has been reached.\n");
         break;
     case SCARD_W_CANCELLED_BY_USER:
         strcpy(Message , "The action was cancelled by the user.\n");
         break;
     case SCARD_W_CARD_NOT_AUTHENTICATED:
         strcpy(Message , "No PIN was presented to the smart card.\n");
         break;
  }

  return Message;
}
//---------------------------------------------------------------------
// End insert Vadim Isaev
//---------------------------------------------------------------------

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
