
#ifdef __cplusplus
 extern "C" {
#endif // __cplusplus

#include <winbase.h>
#include <usbdi.h>
#include <usbclient.h>
#include <windev.h>
#include "dbglog.h"
#include "uemusb.h"
#include "drvioctl.h"


USB_DRIVER_SETTINGS	UsbDriverSettings = {sizeof (UsbDriverSettings), UEMUSB_VENDOR_ID, UEMUSB_DEVICE_ID, USB_NO_INFO, USB_NO_INFO, USB_NO_INFO, USB_NO_INFO, USB_NO_INFO, USB_NO_INFO, USB_NO_INFO};
DWORD			UemDeviceCount;

BOOL USBInstallDriver (LPCWSTR szDriverLibFile)
{
 BOOL	Result;

 DBGLOG ((TEXT ("USBInstallDriver, szDriverLibFile: %s\r\n"), szDriverLibFile));

 if (Result = RegisterClientDriverID (TEXT (UEMUSB_DRIVER_NAME)))
 {
	Result = RegisterClientSettings (szDriverLibFile, TEXT (UEMUSB_DRIVER_NAME), NULL, &UsbDriverSettings);
 }
 return Result;
}

BOOL USBUnInstallDriver ()
{
 BOOL	Result;

 DBGLOG ((TEXT ("USBUnInstallDriver\r\n")));
 Result = TRUE;
 return Result;
}

BOOL USBDeviceAttach (USB_HANDLE hDevice, LPCUSB_FUNCS lpUsbFuncs, LPCUSB_INTERFACE lpInterface, LPCWSTR szUniqueDriverId, LPBOOL fAcceptControl, LPCUSB_DRIVER_SETTINGS lpDriverSettings, DWORD dwUnused)
{
 BOOL			Result;
 LPCUSB_DEVICE		UsbDevice;
 TCHAR			ClientRegistryPath[MAX_PATH];
 PDEVICE_EXTENSION	DeviceExtension;
 DWORD			i;

 DBGLOG ((TEXT ("==>USBDeviceAttach, hDevice: 0x%lX, lpUsbFuncs: 0x%lX, szUniqueDriverId: %s.\r\n"), hDevice, lpUsbFuncs, szUniqueDriverId));

 if (lpInterface == NULL)
 {
	DBGLOG ((TEXT ("lpInterface == NULL.\r\n")));
	if (UsbDevice = lpUsbFuncs->lpGetDeviceInfo (hDevice))
		if (lpInterface = lpUsbFuncs->lpFindInterface (UsbDevice, 0, 0))
		{
			DBGLOG ((TEXT ("UsbDevice: 0x%lX, DeviceAttach lpInterface: %u, EndPoints: %u, Class: %u, Sub: %u, Prot: %u.\r\n"), UsbDevice, lpInterface->Descriptor.bInterfaceNumber, lpInterface->Descriptor.bNumEndpoints, lpInterface->Descriptor.bInterfaceClass, lpInterface->Descriptor.bInterfaceSubClass, lpInterface->Descriptor.bInterfaceProtocol));
			DeviceExtension = (PDEVICE_EXTENSION) LocalAlloc (LPTR, sizeof (DEVICE_EXTENSION));
			if (GetClientRegistryPath (&ClientRegistryPath[0], sizeof (ClientRegistryPath) / sizeof (ClientRegistryPath[0]), szUniqueDriverId))
			{
				DBGLOG ((TEXT ("ClientRegistryPath: %s.\r\n"), &ClientRegistryPath[0]));
				//if (ActivateDevice (&ClientRegistryPath[0], (DWORD) DeviceExtension))
				if (DeviceExtension->hDevice = ActivateDeviceEx (&ClientRegistryPath[0], NULL, 0, DeviceExtension))
				{
					DBGLOG ((TEXT ("ActivateDevice exits.\r\n")));
					if (lpUsbFuncs->lpRegisterNotificationRoutine (hDevice, USBDeviceNotificationCallback, DeviceExtension))
					{
						DBGLOG ((TEXT ("lpUsbFuncs->lpRegisterNotificationRoutine exits.\r\n")));
						Result = TRUE;
						*fAcceptControl = Result;
					}
					else
					{
						DBGLOG ((TEXT ("lpUsbFuncs->lpRegisterNotificationRoutine returned FALSE.\r\n")));
					}
				}
				else
				{
					DBGLOG ((TEXT ("ActivateDevice returned FALSE.\r\n")));
				}
			}
		}
		else
		{
			DBGLOG ((TEXT ("lpFindInterface failed.\r\n")));
			Result = FALSE;
		}
	else
	{
		DBGLOG ((TEXT ("lpGetDeviceInfo failed.\r\n")));
		Result = FALSE;
	}
 }
 if (lpInterface)
 {
	DeviceExtension->UsbFuncs = lpUsbFuncs;
	for (i = 0; i < lpInterface->Descriptor.bNumEndpoints; i++)
	{
		DBGLOG_USB_ENDPOINT_DESCRIPTOR (&lpInterface->lpEndpoints[i].Descriptor);
		if ((lpInterface->lpEndpoints[i].Descriptor.bmAttributes & USB_ENDPOINT_TYPE_MASK) == USB_ENDPOINT_TYPE_BULK)
			if (USB_ENDPOINT_DIRECTION_IN (lpInterface->lpEndpoints[i].Descriptor.bEndpointAddress))
			{
				//if (DeviceExtension->UsbPipeIn == NULL)
				{
					DeviceExtension->UsbPipeIn = (*lpUsbFuncs->lpOpenPipe) (hDevice, &lpInterface->lpEndpoints[i].Descriptor);
					DBGLOG ((TEXT ("UsbPipeIn: 0x%lX, i: %lu.\r\n"), DeviceExtension->UsbPipeIn, i));
				}
			}
			else if (USB_ENDPOINT_DIRECTION_OUT (lpInterface->lpEndpoints[i].Descriptor.bEndpointAddress))
			{
				//if (DeviceExtension->UsbPipeOut == NULL)
				{
					DeviceExtension->UsbPipeOut = (*lpUsbFuncs->lpOpenPipe) (hDevice, &lpInterface->lpEndpoints[i].Descriptor);
					DBGLOG ((TEXT ("UsbPipeOut: 0x%lX, i: %lu.\r\n"), DeviceExtension->UsbPipeOut, i));
				}
			}
	}
	DBGLOG ((TEXT ("Count of endpoints: %lu.\r\n"), i));
	DeviceExtension->hTransferEvent = CreateEvent (NULL, TRUE, FALSE, NULL);
 }
 else
 {
	DBGLOG ((TEXT ("lpInterface is NULL.\r\n")));
	*fAcceptControl = FALSE;
 }
 DBGLOG_DEVICE_EXTENSION (DeviceExtension);
 Result = TRUE;
 DBGLOG ((TEXT ("<==USBDeviceAttach exits.\r\n")));
 return Result;
}

BOOL USBDeviceNotificationCallback (LPVOID lpvNotifyParameter, DWORD dwCode, LPDWORD* dwInfo1, LPDWORD* dwInfo2, LPDWORD* dwInfo3, LPDWORD* dwInfo4)
{
 DWORD			Result;
 PDEVICE_EXTENSION	DeviceExtension;

 DBGLOG ((TEXT ("==>USBDeviceNotificationCallback, lpvNotifyParameter: 0x%lX, dwCode: %lu.\r\n"), lpvNotifyParameter, dwCode));

 switch (dwCode)
 {
 case USB_CLOSE_DEVICE:
	DeviceExtension = (PDEVICE_EXTENSION) lpvNotifyParameter;
	Result = DeactivateDevice (DeviceExtension->hDevice);
	DBGLOG ((TEXT ("DeactivateDevice with hDevice 0x%lX returns: %lu.\r\n"), DeviceExtension->hDevice, Result));
	Sleep (1000);
	//Result = (LocalFree (DeviceExtension) == NULL) && Result;
	break;
 default:
	Result = FALSE;
	break;
 }
 DBGLOG ((TEXT ("<==USBDeviceNotificationCallback exits.\r\n")));
 return Result;
}

DWORD UEM_Init (DWORD dwContext, DWORD dwBusContext)
{
 DWORD	Result;

 DBGLOG ((TEXT ("==>UEM_Init.\r\n")));

 Result = dwBusContext;
 DBGLOG ((TEXT ("==>UEM_Init exits.\r\n")));
 return Result;
}

BOOL UEM_Deinit (DWORD dwContext)
{
 BOOL	Result;

 DBGLOG ((TEXT ("==>UEM_Deinit.\r\n")));

 Result = TRUE;
 DBGLOG ((TEXT ("<==UEM_Deinit exits.\r\n")));
 return Result;
}

DWORD UEM_Open (DWORD dwData, DWORD dwAccess, DWORD dwShareMode)
{
 DWORD	Result;

 DBGLOG ((TEXT ("==>UEM_Open, dwData: 0x%lX.\r\n"), dwData));

 DBGLOG_DEVICE_EXTENSION (dwData);
 Result = dwData;
 DBGLOG ((TEXT ("<==UEM_Open exits.\r\n")));
 return Result;
}

BOOL UEM_Close (DWORD Handle)
{
 BOOL	Result;

 DBGLOG ((TEXT ("==>UEM_Close, Handle: 0x%lX.\r\n"), Handle));

 DBGLOG_DEVICE_EXTENSION (Handle);
 Result = TRUE;
 DBGLOG ((TEXT ("<==UEM_Close exits.\r\n")));
 return Result;
}

BOOL UEM_IOControl (DWORD Handle, DWORD dwIoControlCode, PBYTE pInBuf, DWORD nInBufSize, PBYTE pOutBuf, DWORD nOutBufSize, PDWORD pBytesReturned)
{
 DWORD			Result;
 PDEVICE_EXTENSION	DeviceExtension;
 USB_ERROR		UsbError;
 DWORD			Temp;

 DBGLOG ((TEXT ("==>UEM_IOControl, Handle: 0x%lX, dwIoControlCode: %lu, nInBufSize: %lu, nOutBufSize: %lu, pBytesReturned: 0x%lX.\r\n"), Handle, dwIoControlCode, nInBufSize, nOutBufSize, pBytesReturned));

 DeviceExtension = (PDEVICE_EXTENSION) Handle;
 DBGLOG_DEVICE_EXTENSION (DeviceExtension);
 Temp = 0;
 switch (dwIoControlCode)
 {
 case IOCTL_UEMUSB_GET_VERSION:
	if (pOutBuf && (nOutBufSize >= sizeof (DWORD)))
	{
		*((PDWORD) pOutBuf) = UEMUSB_CURRENT_VERSION;
		Result = TRUE;
	}
	else
	{
		Result = FALSE;
	}
	break;
 case IOCTL_UEMUSB_TRANSFER:
 case 1:
	//Result = IssueBulkTransfer (DeviceExtension->UsbFuncs, DeviceExtension->UsbPipeOut, NULL, NULL, 0, pInBuf, NULL, nInBufSize, &Temp, 10000, &UsbError);
	Result = Transaction2 (DeviceExtension, pInBuf, nInBufSize, pOutBuf, nOutBufSize, &Temp);
	DBGLOG ((TEXT ("Transaction2 result: %lu, BytesReturned: %lu.\r\n"), Result, Temp));
	if (pBytesReturned)
		*pBytesReturned = Temp;
	break;
 default:
	SetLastError (ERROR_INVALID_PARAMETER);
	Result = FALSE;
	break;
 }
 DBGLOG ((TEXT ("<==UEM_IOControl exits.\r\n")));
 return Result;
}

DWORD UEM_Read (DWORD Handle, LPVOID pBuffer, DWORD dwNumBytes)
{
 DWORD	Result;

 DBGLOG ((TEXT ("UEM_Read\r\n")));

 Result = 0;
 return Result;
}

DWORD UEM_Write (DWORD Handle, LPCVOID pBuffer, DWORD dwNumBytes)
{
 DWORD	Result;

 DBGLOG ((TEXT ("UEM_Write\r\n")));

 Result = 0;
 return Result;
}

DWORD UEM_Seek (DWORD Handle, long lDistance, DWORD dwMoveMethod)
{
 DWORD	Result;

 DBGLOG ((TEXT ("UEM_Seek\r\n")));

 Result = 0;
 return Result;
}

VOID UEM_PowerUp ()
{

 DBGLOG ((TEXT ("UEM_PowerUp\r\n")));
}

VOID UEM_PowerDown ()
{

 DBGLOG ((TEXT ("UEM_PowerDown\r\n")));
}

WORD AddToFsc (WORD Fsc, UCHAR Value)
{
 WORD	Result, Temp, i;

 Temp = (Fsc ^ Value) & 0xFF;
 for (i = 0; i < 8; i++)
	if (Temp & 1)
	{
		Temp = Temp >> 1;
		Temp = Temp ^ 0x8408;
	}
	else
		Temp = Temp >> 1;
 Result = (Fsc >> 8) ^ Temp;
 return Result;
}

DWORD Transaction (PDEVICE_EXTENSION DeviceExtension, PVOID lpInBuffer, DWORD nInBufferSize, PVOID lpOutBuffer, DWORD nOutBufferSize, PDWORD lpBytesReturned)
{
 DWORD		Result, i, Temp, Fsc, j;
 PUCHAR		TempBuffer; // [UEM_SCR_BUFFER_SIZE];
 USB_ERROR	UsbError;

 DBGLOG ((TEXT ("Transaction, lpInBuffer: 0x%lX, nInBufferSize: %lu, lpOutBuffer: 0x%lX, nOutBufferSize: %lu\r\n"), lpInBuffer, nInBufferSize, lpOutBuffer, nOutBufferSize));

 Temp = nInBufferSize + 5;
 for (i = 0; i < nInBufferSize; i++)
	if (((PUCHAR) lpInBuffer)[i] >= 0xFD)
		Temp = Temp + 1;
 Temp = Temp + 2;
 DBGLOG ((TEXT ("Temp: %lu\r\n"), Temp));
 if (Temp <= UEM_SCR_BUFFER_SIZE)
 {
	if (TempBuffer = (PUCHAR) LocalAlloc (LPTR, Temp))
	{
		DBGLOG ((TEXT ("TempBuffer: 0x%lX\r\n"), TempBuffer));
		Fsc = 0xFFFF;
		TempBuffer[0] = 0xFD;
		//TempBuffer[1] = 0x00;
		TempBuffer[1] = HIBYTE (DeviceExtension->FrameId);
		Fsc = AddToFsc (Fsc, TempBuffer[1]);
		TempBuffer[2] = (UCHAR) DeviceExtension->FrameId;
		DeviceExtension->FrameId = DeviceExtension->FrameId + 1;
		Fsc = AddToFsc (Fsc, TempBuffer[2]);
		if (TempBuffer[2] >= 0xFD)
		{
			DBGLOG ((TEXT ("0xFF - TempBuffer[1]\r\n")));
			TempBuffer[3] = 0xFF - TempBuffer[2];
			TempBuffer[2] = 0xFF;
			j = 4;
		}
		else
			j = 3;
		for (i = 0; i < nInBufferSize; i++, j++)
		{
			if (((PUCHAR) lpInBuffer)[i] >= 0xFD)
			{
				DBGLOG ((TEXT ("TempBuffer[j + 1] = 0xFF - ((PUCHAR) lpInBuffer)[i]\r\n")));
				TempBuffer[j] = 0xFF;
				TempBuffer[j + 1] = 0xFF - ((PUCHAR) lpInBuffer)[i];
				j = j + 1;
			}
			else
				TempBuffer[j] = ((PUCHAR) lpInBuffer)[i];
			Fsc = AddToFsc (Fsc, ((PUCHAR) lpInBuffer)[i]);
		}
		Fsc = ~Fsc;
		TempBuffer[j] = Fsc & 0xFF;
		if (TempBuffer[j] >= 0xFD)
		{
			DBGLOG ((TEXT ("TempBuffer[j + 1] = 0xFF - TempBuffer[j]\r\n")));
			TempBuffer[j + 1] = 0xFF - TempBuffer[j];
			TempBuffer[j] = 0xFF;
			j = j + 2;
		}
		else
			j = j + 1;
		TempBuffer[j] = (UCHAR) (Fsc >> 8);
		if (TempBuffer[j] >= 0xFD)
		{
			DBGLOG ((TEXT ("TempBuffer[j + 1] = 0xFF - TempBuffer[j]\r\n")));
			TempBuffer[j + 1] = 0xFF - TempBuffer[j];
			TempBuffer[j] = 0xFF;
			j = j + 2;
		}
		else
			j = j + 1;
		TempBuffer[j] = 0xFE;
		j = j + 1;
		DBGLOG ((TEXT ("j: %lu, TempBuffer: 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X\r\n"), j, TempBuffer[0], TempBuffer[1], TempBuffer[2], TempBuffer[3], TempBuffer[4], TempBuffer[5], TempBuffer[6], TempBuffer[7], TempBuffer[8], TempBuffer[9]));
		Temp = 0;
		Result = IssueBulkTransfer (DeviceExtension->UsbFuncs, DeviceExtension->UsbPipeOut, NULL, NULL, USB_OUT_TRANSFER | USB_SHORT_TRANSFER_OK, &TempBuffer[0], NULL, j, &Temp, 10000, &UsbError);
		DBGLOG ((TEXT ("IssueBulkTransfer returns: 0x%lX, Temp: %lu, UsbError: 0x%lX\r\n"), Result, Temp, UsbError));
		TempBuffer = (PUCHAR) LocalAlloc (LPTR, 256);
		ZeroMemory (TempBuffer, 256);
		Temp = 0;
		DBGLOG ((TEXT ("Before IssueBulkTransfer\r\n")));
		Result = IssueBulkTransfer (DeviceExtension->UsbFuncs, DeviceExtension->UsbPipeIn, NULL, NULL, USB_IN_TRANSFER | USB_SHORT_TRANSFER_OK, &TempBuffer[0], NULL, 256, &Temp, 1000, &UsbError);
		DBGLOG ((TEXT ("IssueBulkTransfer returns: 0x%lX, Temp: %lu, UsbError: 0x%lX\r\n"), Result, Temp, UsbError));
		DBGLOG ((TEXT ("TempBuffer: 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X\r\n"), TempBuffer[0], TempBuffer[1], TempBuffer[2], TempBuffer[3], TempBuffer[4], TempBuffer[5], TempBuffer[6], TempBuffer[7], TempBuffer[8], TempBuffer[9], TempBuffer[10], TempBuffer[11], TempBuffer[12], TempBuffer[13], TempBuffer[14], TempBuffer[15], TempBuffer[16], TempBuffer[17]));
	}
	else
		Result = 0;
 }
 else
	Result = 0;
 return Result;
}

DWORD Transaction2 (PDEVICE_EXTENSION DeviceExtension, PVOID lpInBuffer, DWORD nInBufferSize, PVOID lpOutBuffer, DWORD nOutBufferSize, PDWORD lpBytesReturned)
{
 DWORD		Result, i, j, Temp;
 UCHAR		TempBuffer[UEM_SCR_BUFFER_SIZE];
 WORD		Fsc;
 USB_ERROR	UsbError;

 DBGLOG ((TEXT ("==>Transaction2, lpInBuffer: 0x%lX, nInBufferSize: %lu, lpOutBuffer: 0x%lX, nOutBufferSize: %lu, lpBytesReturned: 0x%lX.\r\n"), lpInBuffer, nInBufferSize, lpOutBuffer, nOutBufferSize, lpBytesReturned));

 if (Result = lpInBuffer && nInBufferSize && (nInBufferSize < UEM_SCR_BUFFER_SIZE))
 {
	j = nInBufferSize + 1 + 2 + 2 + 2 + 1;
	for (i = 0; i < nInBufferSize; i++)
		if (((PUCHAR) lpInBuffer)[i] >= 0xFD)
			j = j + 1;
	DBGLOG ((TEXT ("Buffer length: %lu.\r\n"), j));
	if (Result = j <= UEM_SCR_BUFFER_SIZE)
	{
		Fsc = 0xFFFF;
		TempBuffer[0] = 0xFD;
		TempBuffer[1] = 0x00;
		//TempBuffer[1] = HIBYTE (DeviceExtension->FrameId);
		Fsc = AddToFsc (Fsc, TempBuffer[1]);
		TempBuffer[2] = (UCHAR) DeviceExtension->FrameId;
		DeviceExtension->FrameId = DeviceExtension->FrameId + 1;
		Fsc = AddToFsc (Fsc, TempBuffer[2]);
		if (TempBuffer[2] >= 0xFD)
		{
			TempBuffer[3] = 0xFF - TempBuffer[2];
			TempBuffer[2] = 0xFF;
			j = 4;
		}
		else
			j = 3;
		for (i = 0; i < nInBufferSize; i++, j++)
		{
			if (((PUCHAR) lpInBuffer)[i] >= 0xFD)
			{
				TempBuffer[j] = 0xFF;
				TempBuffer[j + 1] = 0xFF - ((PUCHAR) lpInBuffer)[i];
				j = j + 1;
			}
			else
				TempBuffer[j] = ((PUCHAR) lpInBuffer)[i];
			Fsc = AddToFsc (Fsc, ((PUCHAR) lpInBuffer)[i]);
		}
		Fsc = ~Fsc;
		TempBuffer[j] = LOBYTE (Fsc);
		if (TempBuffer[j] >= 0xFD)
		{
			TempBuffer[j + 1] = 0xFF - TempBuffer[j];
			TempBuffer[j] = 0xFF;
			j = j + 2;
		}
		else
			j = j + 1;
		TempBuffer[j] = HIBYTE (Fsc);
		if (TempBuffer[j] >= 0xFD)
		{
			TempBuffer[j + 1] = 0xFF - TempBuffer[j];
			TempBuffer[j] = 0xFF;
			j = j + 2;
		}
		else
			j = j + 1;
		TempBuffer[j] = 0xFE;
		j = j + 1;
		DBGLOG ((TEXT ("j: %lu, TempBuffer: 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X.\r\n"), j, TempBuffer[0], TempBuffer[1], TempBuffer[2], TempBuffer[3], TempBuffer[4], TempBuffer[5], TempBuffer[6], TempBuffer[7], TempBuffer[8], TempBuffer[9]));
		Result = IssueBulkTransfer (DeviceExtension->UsbFuncs, DeviceExtension->UsbPipeOut, DefaultTransferComplete, DeviceExtension->hTransferEvent, USB_OUT_TRANSFER | USB_SHORT_TRANSFER_OK, &TempBuffer[0], NULL, j, &Temp, 1000, &UsbError) == ERROR_SUCCESS;
		DBGLOG ((TEXT ("IssueBulkTransfer returns: 0x%lX, Temp: %lu, UsbError: 0x%lX.\r\n"), Result, Temp, UsbError));
		if (Result)
		{
			ZeroMemory (&TempBuffer[0], sizeof (TempBuffer));
			Result = IssueBulkTransfer (DeviceExtension->UsbFuncs, DeviceExtension->UsbPipeIn, DefaultTransferComplete, DeviceExtension->hTransferEvent, USB_IN_TRANSFER | USB_SHORT_TRANSFER_OK, &TempBuffer[0], NULL, sizeof (TempBuffer), &Temp, 1000, &UsbError) == ERROR_SUCCESS;
			DBGLOG ((TEXT ("IssueBulkTransfer returns: 0x%lX, Temp: %lu, UsbError: 0x%lX.\r\n"), Result, Temp, UsbError));
			DBGLOG ((TEXT ("TempBuffer: 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X 0x%X.\r\n"), TempBuffer[0], TempBuffer[1], TempBuffer[2], TempBuffer[3], TempBuffer[4], TempBuffer[5], TempBuffer[6], TempBuffer[7], TempBuffer[8], TempBuffer[9], TempBuffer[10], TempBuffer[11], TempBuffer[12], TempBuffer[13], TempBuffer[14], TempBuffer[15], TempBuffer[16], TempBuffer[17]));
			if (Result)
			{
				for (i = 0; i < Temp; i++)
				{
					if (TempBuffer[i] == 0xFF)
					{
						TempBuffer[i] = 0xFF - TempBuffer[i + 1];
						MoveMemory (&TempBuffer[i + 1], &TempBuffer[i + 2], Temp - i - 2);
						Temp = Temp - 1;
					}
				}
				if (lpOutBuffer && (nOutBufferSize >= Temp))
				{
					CopyMemory (lpOutBuffer, &TempBuffer[0], Temp);
					if (lpBytesReturned)
						*lpBytesReturned = Temp;
				}
			}
		}
	}
 }
 DBGLOG ((TEXT ("<==Transaction2 exits.\r\n")));
 return Result;
}

/*

NTSTATUS
Transaction(
	IN PFDO_DATA	fdoData,
	IN PVOID			dout,
	IN ULONG			lout,
	IN PVOID			din,
	IN ULONG			lin,
	IN PULONG			rlin
	) {
    UCHAR   inbuff[SCR_BUFFER_SIZE];
	PUCHAR	pos;
	PUCHAR	buff;
	ULONG	i, j, stage, xlen, rest, alout, urbFlags;
	USHORT FSC;
    PUSBD_INTERFACE_INFORMATION interface;
	PUSBD_PIPE_INFORMATION pipeInformation;
	PURB    urb;
	NTSTATUS status;
    KIRQL   irql;

	*rlin = 0;

	if (fdoData->DeviceState != Working) {
        return STATUS_INVALID_DEVICE_STATE;
    }
    if (fdoData->SSEnable) {
        KeWaitForSingleObject(&fdoData->NoIdleReqPendEvent, 
                              Executive, 
                              KernelMode, 
                              FALSE, 
                              NULL);
    }
    if (fdoData->UsbInterface) {
        interface =  fdoData->UsbInterface;
    } else {
        uem_dbgPrint (1, ("UsbInterface not found\n"));
        return STATUS_INVALID_DEVICE_STATE;
    }
	pipeInformation = &interface->Pipes[1];
    if (UsbdPipeTypeBulk != pipeInformation->PipeType) {
        return STATUS_INVALID_HANDLE;
    }
	if (lout == 0 || lin == 0) {
		return STATUS_SUCCESS;
	}

    KeWaitForMutexObject(&fdoData->Interact, Executive, KernelMode, FALSE, NULL);
	for (i = 0, alout = lout+5; i < lout; i++) {
		if (((UCHAR *)dout)[i] >= 0xFD)
			alout++;
	}
    alout += 2; // for purpose of possible FSC byte-stuffing

	if (alout > SCR_BUFFER_SIZE) {
        KeReleaseMutex(&fdoData->Interact, FALSE);;
		return STATUS_INVALID_PARAMETER;
	}
	buff = (PUCHAR) ExAllocatePool(NonPagedPool, alout);
    if (NULL == buff) {
        KeReleaseMutex(&fdoData->Interact, FALSE);;
		return STATUS_INSUFFICIENT_RESOURCES;
    }

	FSC = 0xFFFF;
	buff[0] = 0xFD; // start
    j = 1;
	buff[j] = (UCHAR)(++(fdoData->FID) & 0xFF);  // Frame ID
    fdoData->FID &= 0xFF;
	FSC = add2FSC(FSC, buff[j]);
    if (buff[j++] >= 0xFD) {
        buff[j] = 0xFF - buff[j-1];
        buff[j-1] = 0xFF;
        j++;
    }
	for (i = 0; i < lout; i++, j++) {
		switch (((UCHAR *)dout)[i]) {
			case 0xFD:
			case 0xFE:
			case 0xFF:
                buff[j++] = 0xFF;
                buff[j] = 0xFF - ((UCHAR *)dout)[i];
				break;
			default:
				buff[j] = ((UCHAR *)dout)[i];
		}
		FSC = add2FSC(FSC, ((UCHAR *)dout)[i]);
	}
	FSC ^= 0xFFFF;
    if ((buff[j++] = FSC & 0xFF) >= 0xFD) {
        buff[j] = 0xFF - buff[j-1];
        buff[j-1] = 0xFF;
        j++;
    }
    if ((buff[j++] = (FSC >> 8) & 0xFF) >= 0xFD) {
        buff[j] = 0xFF - buff[j-1];
        buff[j-1] = 0xFF;
        j++;
    }
	buff[j++] = 0xFE; // Stop byte
    alout = j; // real transfere size
	urb = ExAllocatePool(NonPagedPool,
                         sizeof(struct _URB_BULK_OR_INTERRUPT_TRANSFER));
	if(urb == NULL) {
        ExFreePool(buff);
        KeReleaseMutex(&fdoData->Interact, FALSE);;
        return STATUS_INSUFFICIENT_RESOURCES;
	}
    if (alout > SCR_MAX_TRANSFER_SIZE) {
        stage = SCR_MAX_TRANSFER_SIZE;
    } else {
        stage = alout;
    }
uem_dbgPrint(4, ("Write(%d)  %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x",
             alout, 
             buff[0], buff[1], buff[2], buff[3], buff[4], buff[5], 
             buff[6], buff[7], buff[8], buff[9], buff[10], buff[11]));
    urbFlags = USBD_SHORT_TRANSFER_OK | USBD_TRANSFER_DIRECTION_OUT;
    UsbBuildInterruptOrBulkTransferRequest(
                            urb,
                            sizeof(struct _URB_BULK_OR_INTERRUPT_TRANSFER),
                            pipeInformation->PipeHandle,
                            buff,
                            NULL,
                            stage,
                            urbFlags,
                            NULL);
	xlen = 0;
	pos = buff + stage;
	rest = alout - stage;
	for (;;) {
		status = CallUSBD(fdoData->Self, urb);
		if (NT_SUCCESS(status)) {
			xlen += urb->UrbBulkOrInterruptTransfer.TransferBufferLength;
			if (rest) {
				if (rest > SCR_MAX_TRANSFER_SIZE) {
					stage = SCR_MAX_TRANSFER_SIZE;
				} else {
					stage = rest;
				}
//				urb->UrbBulkOrInterruptTransfer.TransferBufferLength  = stage;
//				urb->UrbBulkOrInterruptTransfer.TransferBuffer = pos;
                UsbBuildInterruptOrBulkTransferRequest(
                            urb,
                            sizeof(struct _URB_BULK_OR_INTERRUPT_TRANSFER),
                            pipeInformation->PipeHandle,
                            pos,
                            NULL,
                            stage,
                            urbFlags,
                            NULL);
				pos += stage;
				rest -= stage;
			} else
				break;
        } else {
            *rlin = 1;
            *((PUCHAR) din) = 0xFF; 
			ExFreePool(buff);
			ExFreePool(urb);
            KeReleaseMutex(&fdoData->Interact, FALSE);;
			return status;
		}
	}
   	ExFreePool(buff);

	// Now read and unpack answer
	pipeInformation = &interface->Pipes[0];
    if (UsbdPipeTypeBulk != pipeInformation->PipeType) {
        KeReleaseMutex(&fdoData->Interact, FALSE);;
        return STATUS_INVALID_HANDLE;
    }
    if (lin > SCR_MAX_TRANSFER_SIZE) {
        stage = SCR_MAX_TRANSFER_SIZE;
    } else {
        stage = lin;
    }
	urbFlags = USBD_SHORT_TRANSFER_OK | USBD_TRANSFER_DIRECTION_IN;
    UsbBuildInterruptOrBulkTransferRequest(
                            urb,
                            sizeof(struct _URB_BULK_OR_INTERRUPT_TRANSFER),
                            pipeInformation->PipeHandle,
                            inbuff,
                            NULL,
                            stage,
                            urbFlags,
                            NULL);
	xlen = 0;
	pos = (PUCHAR)inbuff;
	rest = lin - stage;
	for (;;) {
		status = CallUSBD(fdoData->Self, urb);
		if (NT_SUCCESS(status)) {
uem_dbgPrint(4, ("Portion(%d)   %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x\n",
             stage, 
             pos[0], pos[1], pos[2], pos[3], pos[4], pos[5], 
             pos[6], pos[7], pos[8], pos[9], pos[10], pos[11]));
			xlen += urb->UrbBulkOrInterruptTransfer.TransferBufferLength;
			if (rest && urb->UrbBulkOrInterruptTransfer.TransferBufferLength == stage) {
				if (rest > SCR_MAX_TRANSFER_SIZE) {
					stage = SCR_MAX_TRANSFER_SIZE;
				} else {
					stage = rest;
				}
				pos += stage;
				rest -= stage;
                UsbBuildInterruptOrBulkTransferRequest(
                            urb,
                            sizeof(struct _URB_BULK_OR_INTERRUPT_TRANSFER),
                            pipeInformation->PipeHandle,
                            pos,
                            NULL,
                            stage,
                            urbFlags,
                            NULL);
//				urb->UrbBulkOrInterruptTransfer.TransferBufferLength = stage;
//				urb->UrbBulkOrInterruptTransfer.TransferBuffer = pos;
			} else { // last spit of data
				pos = (PUCHAR)inbuff;
uem_dbgPrint(4, ("Read(%d)   %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x\n",
             xlen, 
             pos[0], pos[1], pos[2], pos[3], pos[4], pos[5], 
             pos[6], pos[7], pos[8], pos[9], pos[10], pos[11]));
				for (i = 1, j = 0; i < xlen-1; i++) {
					if (pos[i] == 0xFF)
						pos[j++] = 0xFF - pos[++i];
					else
						pos[j++] = pos[i];
				}
uem_dbgPrint(4, ("decBS(%d)  %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x\n",
             j, 
             pos[0], pos[1], pos[2], pos[3], pos[4], pos[5], 
             pos[6], pos[7], pos[8], pos[9], pos[10], pos[11]));
				FSC = 0xFFFF;
               	FSC = add2FSC(FSC, pos[0]); // FID
                for (i = 1; i < j - 1; i++) {
                    FSC = add2FSC(FSC, pos[i]);
                    if (i != 1) // cmd echoed
                        ((PUCHAR)din)[i-2] = pos[i];
                }
	            FSC ^= 0xFFFF;
				// can compare FSC with pos[i] [i+1] here
				*rlin = j - 4;  // - FID - CMD - 2CRC
				pos = (PUCHAR)din;
uem_dbgPrint(4, ("  %04x(%d) %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x\n",
             FSC,
             *rlin, 
             pos[0], pos[1], pos[2], pos[3], pos[4], pos[5], 
             pos[6], pos[7], pos[8], pos[9], pos[10], pos[11]));
                break;
			}
        } else {
            *rlin = 1;
            *((PUCHAR) din) = 0xFF; 
			break;
        }
	} // end loop
	ExFreePool(urb);
    KeReleaseMutex(&fdoData->Interact, FALSE);
	return status;
}

*/

#ifdef __cplusplus
 };
#endif // __cplusplus
