// TMH_Demo.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "conio.h"
#include "TMH_Demo.h"
#include "clscrfl.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define TYPE_CONTACT        1
#define TYPE_CONTACTLESS    2
#define BOTH_TYPES          3

#define LOOP_DELAY_MS   2000

#define BUF_SIZE        520

/////////////////////////////////////////////////////////////////////////////
// The one and only application object

CWinApp theApp;
using namespace std;

const BYTE APDU_EXAMPLE[] =         "\x00\xA4\x04\x0C\x07\xA0\x00\x00\x02\x47\x10\x01";
const BYTE T_CL_EXAMPLE[] = "\x0A\x00\x00\xA4\x04\x0C\x07\xA0\x00\x00\x02\x47\x10\x01";

void* m_pReader = NULL;
BYTE  RecvBuffer[BUF_SIZE];
DWORD RecvLength;
LONG  Status;
char TargetCardType = BOTH_TYPES;

void PrintBuf( BYTE* Buffer, DWORD Length )
{
    DWORD i;

    for( i = 0; i < Length; i++ )
        printf( " %02X", Buffer[i] );
    printf( "\n" );
}

int IsCardPresent( void )
{
    LONG  Status = 0;
    BYTE CardPresent;

    printf( "\n" );

// ======================= Is Card Present =====================
    Status = CLSCRF_TDA8029_GetStatus( m_pReader, &CardPresent );
    if( Status )
    {

// =================== Contact interface reset =================
        CLSCRF_TDA8029_Reset( m_pReader, 10 );
        return 0;
    }
    else
        return CardPresent;
}


void ProceedContact( void )
{
// ======================= Card Power On =======================
    RecvLength = BUF_SIZE;
    Status = CLSCRF_TDA8029_Exchange( m_pReader,
                                      0x69,
                                      NULL,
                                      0,
                                      RecvBuffer,
                                      &RecvLength );
    if( Status == 0 )
    {
        printf( "Contact card ATR: " );
        PrintBuf( RecvBuffer, RecvLength );
    
        printf( "APDU-Command:     " );
        PrintBuf( (BYTE*)APDU_EXAMPLE, sizeof( APDU_EXAMPLE ) );

// ========================== Card APDU ========================
        RecvLength = BUF_SIZE;
        Status = CLSCRF_TDA8029_APDU( m_pReader,
                                      APDU_EXAMPLE,
                                      sizeof( APDU_EXAMPLE ),
                                      RecvBuffer,
                                      &RecvLength );
        if( Status == 0 )
        {
            printf( "APDU-Response:    " );
            PrintBuf( RecvBuffer, RecvLength );
        }
    }

// ======================= Card Power Off ======================
    RecvLength = BUF_SIZE;
    Status = CLSCRF_TDA8029_Exchange( m_pReader,
                                      0x4D,
                                      NULL,
                                      0,
                                      RecvBuffer,
                                      &RecvLength );
}



void ProceedContactless( void )
{
    DWORD State;
    DWORD UID_len;
    BYTE ATQ[2];
    BYTE SAK;
    BYTE UID[10];

    printf( "\n" );

// ============== Get Contactless Interface State ==============
    Status = CLSCRF_GetState( m_pReader, &State );
    if( Status )
    {
        printf( "Reader has problems\n" );
        return;
    }
    if( State == 0 )
    {

// ============== Contactless Interface Power On ===============
        Status = CLSCRF_Mfrc_On( m_pReader );
    }
    else
    {

// ===================== Radio Field Reset =====================
        Status = CLSCRF_Mfrc_Rf_Off_On( m_pReader, 10 );
    }

    if( Status == 0 )
    {

// ================= Powering Contactless Card =================
        Sleep( 10 );

// ================= Contactless Card Activate =================
        Status = CLSCRF_Activate_Idle_A( m_pReader,
                                         ATQ,
                                         &SAK,
                                         UID,
                                         &UID_len );
    }
    if( Status == 0 )
    {
        printf( "Contactless card:  " );
        printf("ATQ = %02X%02X   ", ATQ[1], ATQ[0] );
        printf("SAK = %02X   ", SAK );
        printf("UID =" );
        PrintBuf( UID, UID_len );

// ================ ISO14443-4 Activate (T=CL)  ================
        Status = CLSCRF_ISO14443A_4_RATS( m_pReader,
                                          0,
                                          RecvBuffer );
    }
    if( Status == 0 )
    {
        printf( "Contactless card ATS:         " );
        PrintBuf( RecvBuffer, RecvBuffer[0] );            

        printf( "APDU-Command:     " );
        PrintBuf( (BYTE*)T_CL_EXAMPLE, sizeof( T_CL_EXAMPLE ) );

// ==================== Contactless Card APDU ==================
        RecvLength = BUF_SIZE;
        CLSCRF_DirectIO_Card( m_pReader,
                              T_CL_EXAMPLE,
                              sizeof( T_CL_EXAMPLE ),
                              0x17800,
                              RecvBuffer,
                              &RecvLength );
        if( Status == 0 )
        {
            printf( "APDU-Response:    " );
            PrintBuf( RecvBuffer, RecvLength );
        }
    }

// ====================== Radio Field Off ======================
    CLSCRF_Mfrc_Rf_Off_On( m_pReader, 0 );
}



int _tmain(int argc, TCHAR* argv[], TCHAR* envp[])
{

	// initialize MFC and print and error on failure
	if (!AfxWinInit(::GetModuleHandle(NULL), NULL, ::GetCommandLine(), 0))
	{
		// TODO: change error code to suit your needs
		cerr << _T("Fatal Error: MFC initialization failed") << endl;
		return 1;
	}

    DWORD m_PortNumber = 0;
    DWORD m_Baudrate = 9600;
    DWORD State;
    BYTE  m_DeviceAddress = 0;

    if( argc > 1 )
    {
        if( strncmp( argv[1], "COM", 3 ) == 0 )
        {
            m_PortNumber = atoi( &argv[1][3] );
            if( m_PortNumber == 0 )
            {
                printf( "Port %s can't be opened\n", argv[1] );
                Status = 2;
            }
        }
        else
        {
            printf( "Syntax:   Filename.exe [COMx [Baudrate [Address]]]\n" );
            printf( "          (without parameters = USB\n" );
            printf( "          (default Baudrate = 9600 bod)\n" );
            printf( "          (Address from 0 to 255; default 0)\n" );
            printf( "Examples: Filename.exe\n" );
            printf( "          Filename.exe COM1\n" );
            printf( "          Filename.exe COM3 115200\n" );
            printf( "          Filename.exe COM15 9600 17\n" );
            Status = 1;
        }
    }
    if( argc > 2 )
    {
        m_Baudrate = atoi( argv[2] );
    }
    if( argc > 3 )
    {
        m_DeviceAddress = atoi( argv[3] );
    }

    printf("\nDUAL INTERFACE CARD READER's Demo Program\n");
    printf("\n");

    Status = CLSCRF_Create( &m_pReader );

    if( Status == 0 )
        printf("Interface is created successfully\n");
    else
        printf("Interface is not created\n");

    if( m_PortNumber )
        Status = CLSCRF_OpenRS( m_pReader, m_PortNumber-1, m_Baudrate, 0 );
    else
        Status = CLSCRF_OpenUSB( m_pReader, 0, 0 );

    printf("Interface ");
    if( m_PortNumber )
        printf("COM%d ", m_PortNumber);
    else
        printf("USB ");
    if( Status == 0 )
        printf("is opened successfully\n");
    else
        printf("is not opened\n");

    if( Status == 0 )
        Status = CLSCRF_Mfrc_On( m_pReader );
    if( Status == 0 )
        Status = CLSCRF_Mfrc_Rf_Off_On( m_pReader, 0 );
    if( Status == 0 )
        printf("Reader-IC is turn on successfully\n");
    else
        printf("Reader-IC is not turn on\n");

    if( Status == 0 )
    {
        printf("Press:\n");
        printf("       1  - for contact card only\n");
        printf("       2  - for contactless card only\n");
        printf("       3  - for contact and contactles card\n");
        printf("any other - for exit\n");
        while( !_kbhit() )
            Sleep( 100 );
        TargetCardType = _getch() - 0x30;
        if( ( TargetCardType < 1 ) || ( TargetCardType > 3 ) )
            Status = -1;
    }

    if( Status == 0 )
// Loop Begin
        while( 1 )
        {
            if( _kbhit() )
            {
                TargetCardType = _getch() - 0x30;
                if( ( TargetCardType < 1 ) || ( TargetCardType > 3 ) )
                    break;
            }

            if( IsCardPresent() )
            {
                Status = CLSCRF_Sound( m_pReader, 1 );
                if( TargetCardType & TYPE_CONTACT )
                    ProceedContact();
                if( TargetCardType & TYPE_CONTACTLESS )
                    ProceedContactless();
            }

            Sleep( LOOP_DELAY_MS );

        }
// Loop End 
    printf( "\n" );

    if( m_pReader )
    {
        if( CLSCRF_IsOpened( m_pReader ) )
        {
            CLSCRF_GetState( m_pReader, &State );
            if( State )
            {
                Status = CLSCRF_Mfrc_Off( m_pReader );
                if( Status == 0 )
                    printf("Reader-IC is turn off successfully\n");
            }

            Status = CLSCRF_Close( m_pReader );

            printf("Interface ");
            if( m_PortNumber )
                printf("COM%d ", m_PortNumber);
            else
                printf("USB ");
            if( Status == 0 )
                printf("is closed successfully\n");
            else
                printf("is not closed\n");
        }

        Status = CLSCRF_Destroy( &m_pReader );

        if( Status == 0 )
            printf("Interface is destroyed successfully\n");
        else
            printf("Interface is not destroyed\n");
    }

    printf("\n");
    printf( "Press any key ..." );
    _getch();

    return 0;
}

