//==========================================================================//
//  CRC.c : implementation file                                             //
//==========================================================================//

//#include <Config.h>
#include "stdafx.h"

#define RANDOM_MASK  0x83

DWORD m_TableCRC32[256];
WORD  CRC_table[256];
BYTE  m_RandShift;

DWORD Reflect( DWORD v, int bits )
{
    DWORD ret = 0;
    int i;

    --bits;
    for(i = 0; i <= bits; ++i)
    {
        if(v & (1<<i))
            ret |= 1 << (bits-i);
    }
    return ret;
}

void CRC_Init( void )
{
    const DWORD POLYNOMIAL = 0x04C11DB7;
    int i, j;

    const unsigned char POLINOM[] = { 16, 12, 5, 0 };
    unsigned short CRC_Mask = 0;
    unsigned short CRC_Value;

    for(i = 0; i < 0x100; ++i)
    {
        m_TableCRC32[i] = Reflect( (DWORD)i, 8 ) << 24;

        for(j = 0; j < 8; ++j)
            m_TableCRC32[i] = (m_TableCRC32[i] << 1) ^ ( (m_TableCRC32[i] & (1<<31))  ? POLYNOMIAL : 0);

        m_TableCRC32[i] = Reflect(m_TableCRC32[i],32);
    }

// CRC (FCS) http://lattice.itep.ru/old/UNIX/RFC/rfc1331.html
// The HDLC CRC16 polynomial: x**0 + x**5 + x**12 + x**16 (0x8408)
// Highest power (**16) is not used
//     bits no: 0123 4567 89ab cdef
// bin values   1000 0100 0000 1000
// hex values     8    4    0    8
// Work shift >> (to right)
// Initial value CRC = 0xffff
// Invert result
    for( i = 1; i < sizeof( POLINOM ); i++ )
        CRC_Mask |= 0x8000 >> POLINOM[i];

    for( i = 0; i < 256; i++ )
    {
        CRC_Value = i;
        for( j = 0; j < 8; j++ )
            CRC_Value = CRC_Value & 1 ? (CRC_Value >> 1) ^ CRC_Mask : CRC_Value >> 1;
        CRC_table[i] = CRC_Value;
    }
    m_RandShift = 1;
}

void CalculateCRC32( DWORD *CRC32, BYTE b )
{
    *CRC32 = (*CRC32 >> 8) ^ m_TableCRC32[ (*CRC32 & 0xFF) ^ b ];
}

void CalculateFCS( unsigned short *FCS, unsigned char b )
{
    *FCS = ( *FCS >> 8 ) ^ CRC_table[ (*FCS ^ b) & 0xff ];
}

BYTE RandomByte( void )//Galua_LFSR( void )
{
    if( m_RandShift & 0x01 )
        m_RandShift = ( m_RandShift ^ RANDOM_MASK >> 1 ) | 0x80;
    else
        m_RandShift >>= 1;
    return m_RandShift;        
}


