// PCSC_SAM_query.cpp:      .
//

// Preprocessor definitions:
// PCSC		- for PC/SC transactions (PR533 reader) or standard UEM USB commands (CLRC663 reader)
// SAM		- using SAM or not
// MANUAL	- using UEM proprietary commands or not
// PR		- if set, the reader is based on PR533 and uses CCID protocol 

// Authentication key for Mifare Plus Card Sector 0: A5A4A3A2A1A018778E69F0F0F0F0F0F0
// Reader Flash memory block for storing Authentication key: 0x02
// SAM Key Number and Version for storing Authentication Key: 0x01, 0x00

// PC/SC NO SAM ~= 62 ms (1 cycle) period
// PC/SC SAM ~= 420 ms (2 cycles) period
// USB NO SAM ~= 62 ms (1 cycle) period
// USB NO SAM MANUAL ~= 62 ms (1 cycle) period
// USB SAM ~= 218 ms (2 cycles) period
// USB SAM MANUAL ~= 218 ms (2 cycles) period

#include "stdafx.h"
#include "..\..\Clscrfl.h"
#include "..\..\Clscrfl_SAM.h"
#include "aes_128.h"

BYTE lpKey[16] = { 0xA5, 0xA4, 0xA3, 0xA2, 0xA1, 0xA0, 0x18, 0x77, 0x8E, 0x69, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0 };
				   //0xA5, 0xA4, 0xA3, 0xA2, 0xA1, 0xA0, 0x18, 0x77, 0x8E, 0x69, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0,

SCARDCONTEXT		m_SCardContextHandle = NULL;
SCARDHANDLE			m_SCardCardHandle = NULL;
SCARDHANDLE			m_hSCardSAM1 = NULL;
SCARDHANDLE			m_hSCardSAM2 = NULL;
SCARDHANDLE			m_hSCardSAM3 = NULL;
SCARDHANDLE			m_hSCardSAM4 = NULL;
SCARDHANDLE			m_hSCardActiveSAM = NULL;
BOOL				m_fSwitchSAM = FALSE;

LPVOID				m_pReader = NULL;

#ifdef PCSC
//DWORD				dwSmallPause = 100;
DWORD				dwSmallPause = 60;
DWORD				dwLargePause = 1000;
//DWORD				dwLargePause = 0;
DWORD				dwCyclePause = 100;
DWORD				dwActivationPause = 100;
#else
//DWORD				dwSmallPause = 100;
DWORD				dwSmallPause = 0;
DWORD				dwLargePause = 10;
//DWORD				dwLargePause = 0;
DWORD				dwCyclePause = 0;
DWORD				dwActivationPause = 10;
#endif

#ifdef SAM
DWORD				dwSAMPause = 0;
#endif

bool				m_fModeTCL = false;

void printTime();

#ifdef PCSC
LONG EstablishContext()
{
    LONG Status;

	Status = SCardEstablishContext(SCARD_SCOPE_USER, NULL, NULL, &m_SCardContextHandle);

	return Status;
}
#else
LONG EstablishLink()
{
	LONG Status;

	Status = CLSCRF_Create(&m_pReader);
	if (Status != 0) return Status;

	Status = CLSCRF_OpenUSB(m_pReader);
	if (Status != 0) return Status;

	Status = CLSCRF_Mfrc_On(m_pReader);
	if (Status != 0) return Status;

	DWORD dwState = 0;

	Status = CLSCRF_GetState(m_pReader, &dwState);
	if (Status != 0) return Status;

	if ((dwState & 0xFFUL) != 0x00)
	{
		Status = CLSCRF_Mfrc_Set_Rf_Mode(m_pReader, 0x00);
		if (Status != 0) return Status;
	}

	return SCARD_S_SUCCESS;
}
#endif

#ifdef PCSC
LONG ReleaseContext()
{
	LONG Status = SCARD_S_SUCCESS;

	if (m_SCardContextHandle)
	{
		Status = SCardReleaseContext(m_SCardContextHandle);
		m_SCardContextHandle = NULL;
	}

	return Status;
}
#else
LONG ReleaseContext()
{
	LONG Status = SCARD_S_SUCCESS;

	Status = CLSCRF_Close(m_pReader);
	if (Status != 0) return Status;

	Status = CLSCRF_Destroy(&m_pReader);

	return SCARD_S_SUCCESS;
}
#endif

#ifdef PCSC
	#ifdef PR
	LONG ConnectCard()
	{
		LONG Status = SCARD_S_SUCCESS;

		DWORD ActiveProtocol = 0;
		TCHAR* szPCSC = _T("NXP PR533 0");
		/* Establish the connection with the smart card */
		Status = SCardConnect(
			m_SCardContextHandle,
			(LPCTSTR)szPCSC,
			SCARD_SHARE_SHARED,
			SCARD_PROTOCOL_T1,
			&m_SCardCardHandle,
			(LPDWORD)&ActiveProtocol
			);

		if (Status != SCARD_S_SUCCESS)
		{
			szPCSC = _T("MicroEM NFC Reader 0");
			/* Establish the connection with the smart card */
			Status = SCardConnect(
				m_SCardContextHandle,
				(LPCTSTR)szPCSC,
				SCARD_SHARE_SHARED,
				SCARD_PROTOCOL_T1,
				&m_SCardCardHandle,
				(LPDWORD)&ActiveProtocol
				);
		}

		if (Status != SCARD_S_SUCCESS)
		{
			szPCSC = _T("NXP NFC CARD READER 0");
			/* Establish the connection with the smart card */
			Status = SCardConnect(
				m_SCardContextHandle,
				(LPCTSTR)szPCSC,
				SCARD_SHARE_SHARED,
				SCARD_PROTOCOL_T1,
				&m_SCardCardHandle,
				(LPDWORD)&ActiveProtocol
				);
		}

		return Status;
	}
	#else
	LONG ConnectCard()
	{
		LONG Status = SCARD_S_SUCCESS;

		DWORD ActiveProtocol = 0;
		TCHAR* szPCSC = _T("UEM CARD READER 1");
		/* Establish the connection with the smart card */
		Status = SCardConnect(
			m_SCardContextHandle,
			(LPCTSTR)szPCSC,
			SCARD_SHARE_SHARED,
			SCARD_PROTOCOL_T1,
			&m_SCardCardHandle,
			(LPDWORD)&ActiveProtocol
			);

		if (Status != SCARD_S_SUCCESS)
		{
			m_fModeTCL = false;

			return Status;
		}

		szPCSC = _T("UEM SAM1 READER 1");
		/* Establish the connection with the smart card */
		Status = SCardConnect(
			m_SCardContextHandle,
			(LPCTSTR)szPCSC,
			SCARD_SHARE_SHARED,
			SCARD_PROTOCOL_T1,
			&m_hSCardActiveSAM,
			(LPDWORD)&ActiveProtocol
			);

		if (Status == SCARD_S_SUCCESS)
			m_hSCardSAM1 = m_hSCardActiveSAM;

		szPCSC = _T("UEM SAM2 READER 1");
		/* Establish the connection with the smart card */
		Status = SCardConnect(
			m_SCardContextHandle,
			(LPCTSTR)szPCSC,
			SCARD_SHARE_SHARED,
			SCARD_PROTOCOL_T1,
			&m_hSCardActiveSAM,
			(LPDWORD)&ActiveProtocol
			);

		if (Status == SCARD_S_SUCCESS)
		{
			if (m_hSCardSAM1 == NULL)
				m_hSCardSAM1 = m_hSCardActiveSAM;
			else
				m_hSCardSAM2 = m_hSCardActiveSAM;
		}

		szPCSC = _T("UEM SAM3 READER 1");
		/* Establish the connection with the smart card */
		Status = SCardConnect(
			m_SCardContextHandle,
			(LPCTSTR)szPCSC,
			SCARD_SHARE_SHARED,
			SCARD_PROTOCOL_T1,
			&m_hSCardActiveSAM,
			(LPDWORD)&ActiveProtocol
			);

		if (Status == SCARD_S_SUCCESS)
		{
			if (m_hSCardSAM1 == NULL)
				m_hSCardSAM1 = m_hSCardActiveSAM;
			else if (m_hSCardSAM2 == NULL)
				m_hSCardSAM2 = m_hSCardActiveSAM;
			else
				m_hSCardSAM3 = m_hSCardActiveSAM;
		}

		szPCSC = _T("UEM SAM4 READER 1");
		/* Establish the connection with the smart card */
		Status = SCardConnect(
			m_SCardContextHandle,
			(LPCTSTR)szPCSC,
			SCARD_SHARE_SHARED,
			SCARD_PROTOCOL_T1,
			&m_hSCardActiveSAM,
			(LPDWORD)&ActiveProtocol
			);

		if (Status == SCARD_S_SUCCESS)
		{
			if (m_hSCardSAM1 == NULL)
				m_hSCardSAM1 = m_hSCardActiveSAM;
			else if (m_hSCardSAM2 == NULL)
				m_hSCardSAM2 = m_hSCardActiveSAM;
			else if (m_hSCardSAM3 == NULL)
				m_hSCardSAM3 = m_hSCardActiveSAM;
			else
				m_hSCardSAM4 = m_hSCardActiveSAM;
		}

		m_hSCardActiveSAM = m_hSCardSAM1;

		if (m_hSCardActiveSAM == NULL)
		{
			printf(("\nm_hSCardActiveSAM == NULL\n"));

			return SCARD_E_NO_SMARTCARD;
		}

		printf(("\nm_hSCardSAM1 = %08X\n"), m_hSCardSAM1);
		printf(("m_hSCardSAM2 = %08X\n"), m_hSCardSAM2);
		printf(("m_hSCardSAM3 = %08X\n"), m_hSCardSAM3);
		printf(("m_hSCardSAM4 = %08X\n"), m_hSCardSAM4);

		return SCARD_S_SUCCESS;
	}
	#endif
#else
LONG ConnectCard()
{
	return SCARD_S_SUCCESS;
}
#endif

#ifdef PCSC
LONG DisconnectCard()
{
    LONG Status = SCARD_S_SUCCESS;

	if (m_SCardCardHandle)
	{
		//Status = SCardDisconnect(m_SCardCardHandle, SCARD_LEAVE_CARD);
		Status = SCardDisconnect(m_SCardCardHandle, SCARD_RESET_CARD);
		m_SCardCardHandle = NULL;

#ifndef PR
		SCardDisconnect(m_hSCardSAM1, SCARD_LEAVE_CARD);
		m_hSCardSAM1 = NULL;

		SCardDisconnect(m_hSCardSAM2, SCARD_LEAVE_CARD);
		m_hSCardSAM2 = NULL;

		SCardDisconnect(m_hSCardSAM3, SCARD_LEAVE_CARD);
		m_hSCardSAM3 = NULL;

		SCardDisconnect(m_hSCardSAM4, SCARD_LEAVE_CARD);
		m_hSCardSAM4 = NULL;

		m_hSCardActiveSAM = NULL;
#endif
	}

	return Status;
}
#else
LONG DisconnectCard()
{
	return SCARD_S_SUCCESS;
}
#endif

LONG PCSCExchange(PBYTE lpApdu, DWORD dwApduSize, PBYTE lpResponseBuffer, PDWORD pdwResponseBufferLength, SCARDHANDLE hSCard = NULL)
{
	SCARD_IO_REQUEST	IO_Request;
	IO_Request.dwProtocol = SCARD_PROTOCOL_T1;
	IO_Request.cbPciLength = sizeof(SCARD_IO_REQUEST);

	if (hSCard == NULL)
		hSCard = m_SCardCardHandle;

	//printf(("PCSCExchange hSCard = 0x%08X\n"), hSCard);

	memset(lpResponseBuffer, 0x00, *pdwResponseBufferLength);

	LONG Status = SCardTransmit(
		hSCard,
		&IO_Request,
		lpApdu,
		dwApduSize,
		0,
		lpResponseBuffer,
		pdwResponseBufferLength
		);

	if (Status != SCARD_S_SUCCESS)
		return Status;

	BYTE ucSW1 = lpResponseBuffer[*pdwResponseBufferLength - 2];
	BYTE ucSW2 = lpResponseBuffer[*pdwResponseBufferLength - 1];
	
	if ((ucSW1 != 0x90) || ((ucSW2 != 0x00) && (ucSW2 != 0xAF)))
	{
		//printf((" ucSW1 = 0x%02X, ucSW2 = 0x%02X "), ucSW1, ucSW2);

		return SCARD_E_UNEXPECTED;
	}
	
	if (lpResponseBuffer[0] == 0xC0) // PCSC Transparent exchange: Generic error status
	{
		BYTE ucErrorObject = lpResponseBuffer[2];
		BYTE ucInnerSW1 = lpResponseBuffer[3];
		BYTE ucInnerSW2 = lpResponseBuffer[4];

		if ((ucInnerSW1 != 0x90) || ((ucInnerSW2 != 0x00) && (ucInnerSW2 != 0xAF)))
		{
			//printf((" ucInnerSW1 = 0x%02X ucInnerSW2 = 0x%02X "), ucInnerSW1, ucInnerSW2);

			return SCARD_E_UNEXPECTED;
		}
	}

	return Status;
}

#ifdef PCSC
	#ifdef PR
	LONG InitReaderParameters()
	{
		DWORD	dwResponseBufferLength = 255;
		BYTE	lpResponseBuffer[255] = { 0 };

		unsigned char WriteRegisterAPDU[] = {
			0xFF,/* APDU Class */
				0xE1,/* Instruction as defined by PCSC */
				0x02,/* Key structure P1 as defined in PCSC */
				0x01,/* Key structure P2 as defined in PCSC */
				0x03,/* Key structure Lc as defined in PCSC */
				0x61, 0x06, //Address
				0x1B //Value
		};

		LONG Status = PCSCExchange(WriteRegisterAPDU, sizeof(WriteRegisterAPDU), lpResponseBuffer, &dwResponseBufferLength);

		return SCARD_S_SUCCESS;// Status;
	}
	#else
	LONG InitReaderParameters()
	{
		return SCARD_S_SUCCESS;
	}
	#endif
#else
LONG InitReaderParameters()
{
	#ifndef SAM
		#ifndef MANUAL

	LONG Status = CLSCRF_WriteFlashValue(
		m_pReader,
		2,
		lpKey);

	return Status;

		#endif
	#endif

	return SCARD_S_SUCCESS;
}
#endif

#if defined(PCSC) && defined(PR)
LONG SwitchOnTDA()
{
	DWORD	dwResponseBufferLength = 255;
	BYTE	lpResponseBuffer[255] = { 0 };

	unsigned char TDAONAPDU[] = {
		0xFF,/* APDU Class */
		0xE1,/* Instruction as defined by PCSC */
		0x03,/* Key structure P1 as defined in PCSC */
		0x00,/* Key structure P2 as defined in PCSC */
		0x01,/* Key structure Lc as defined in PCSC */
		0x38
	};

	LONG Status = PCSCExchange(TDAONAPDU, sizeof(TDAONAPDU), lpResponseBuffer, &dwResponseBufferLength);

	return Status;
}
#else
LONG SwitchOnTDA()
{
	return SCARD_S_SUCCESS;
}
#endif

#ifdef PCSC
	#ifdef PR
	LONG SwitchSAM(BOOL fP34)
	{
		union UP {
			struct {
				struct {
					BYTE fRFU1:1;
					BYTE fRFU2:1;
					BYTE fP32:1;
					BYTE fRFU3:1;
					BYTE fP34:1;
					BYTE fP35:1;
					BYTE fNU:1;
					BYTE fVal1:1;
				};
				struct {
					BYTE fP70:1;
					BYTE fRFU1:1;
					BYTE fRFU2:1;
					BYTE fRFU3:1;
					BYTE fRFU4:1;
					BYTE fRFU5:1;
					BYTE fRFU6:1;
					BYTE fVal2:1;
				};
			};
			BYTE RAW[2];
		} P = {0};

		P.fVal1 = 1;
		P.fP35 = 1;
		P.fP34 = fP34;
		P.fP32 = 1;
		P.fVal2 = 0;
		P.fP70 = 0;

		DWORD	dwResponseBufferLength = 255;
		BYTE	lpResponseBuffer[255] = { 0 };

		unsigned char WriteGPIOAPDU[] = {
			0xFF,/* APDU Class */
				0xE1,/* Instruction as defined by PCSC */
				0x01,/* Key structure P1 as defined in PCSC */
				0x01,/* Key structure P2 as defined in PCSC */
				0x02,/* Key structure Lc as defined in PCSC */
				P.RAW[0], P.RAW[1] //Value
		};

		LONG Status = PCSCExchange(WriteGPIOAPDU, sizeof(WriteGPIOAPDU), lpResponseBuffer, &dwResponseBufferLength);

		return Status;
	}
	#else
	LONG SwitchSAM(BOOL fP34)
	{
		if (fP34 == FALSE)
		{
			if (m_hSCardSAM1)
				m_hSCardActiveSAM = m_hSCardSAM1;
		}
		else
		{
			if (m_hSCardSAM2)
				m_hSCardActiveSAM = m_hSCardSAM2;
		}

		return SCARD_S_SUCCESS;
	}
	#endif
#else
LONG SwitchSAM(BOOL fP34)
{
	CLSCRF_SAM_Reset(m_pReader);
	return CLSCRF_SAM_SetCurrentHolder(m_pReader, (BYTE)fP34 + 1);
}
#endif

BYTE CalcLRC(PBYTE buffer, int iBufferLen)
{
	BYTE LRC = 0x00;
	for (int i = 0; i < iBufferLen; i++)
		LRC ^= buffer[i];
	return LRC;
}

#ifdef PCSC
	#ifdef PR
	LONG ActivateSAM()
	{
		DWORD	dwResponseBufferLength = 255;
		BYTE	lpResponseBuffer[255] = { 0 };

		unsigned char TDAActivateAPDU[] = {
			0xFF,/* APDU Class */
			0xF4,/* Instruction as defined by PCSC */
			0x60,/* Key structure P1 as defined in PCSC */
			0x00,/* Key structure P2 as defined in PCSC */
			0x01,/* Key structure Lc as defined in PCSC */
			0x6E,
			0x00,
			0x0F
		};

		LONG Status = PCSCExchange(TDAActivateAPDU, sizeof(TDAActivateAPDU), lpResponseBuffer, &dwResponseBufferLength);

		if ((Status == SCARD_S_SUCCESS) && (dwResponseBufferLength >= 7))
		{
			BYTE ucCommandType = lpResponseBuffer[3];
			WORD wLength = ((WORD)lpResponseBuffer[1] << 8) | (WORD)lpResponseBuffer[2];
			if (lpResponseBuffer[0] == 0x60) //ACK
			{
				if (0x00 == CalcLRC(lpResponseBuffer, dwResponseBufferLength - 2))
				{
					if (ucCommandType == 0x6E)
					{
						Status = SCARD_S_SUCCESS;
						//AddReport( true, "TDA ALPAR OK   ATR =", &pResponseBuffer[4], wLength, 0 );
						printf(("0x%08X"), Status);
						printTime();
						printf("ATR = ");
						int iPos = 4;
						for( int i = 0; i < wLength; i++ )
							printf("%02X", lpResponseBuffer[iPos++]);
					}
					else
					{
						Status = SCARD_E_UNEXPECTED;
						//AddReport( true, "Incorrect TDA ALPAR command type =", &ucCommandType, 1, 0 );
					}
				}
				else
				{
					Status = SCARD_E_UNEXPECTED;
					//AddReport( true, "Incorrect TDA ALPAR LRC = ", &pResponseBuffer[4 + wLength], 1, 0 );
				}
			}
			else
			{
				Status = SCARD_E_UNEXPECTED;
				//AddReport( true, "TDA ALPAR Command failed, error code =", &pResponseBuffer[4], wLength, 0 );
			}
		}
		else
			Status = SCARD_E_UNEXPECTED;

		return Status;
	}
	#else
	LONG ActivateSAM()
	{
		return SCARD_S_SUCCESS;
	}
	#endif
#else
LONG ActivateSAM()
{
	return CLSCRF_SAM_Reset(m_pReader);
}
#endif

#ifdef PCSC
	#ifdef PR
	LONG CommandSAM(PBYTE lpSAMAPDU, DWORD dwLenSAMAPDU, PBYTE pRespSAMAPDU, DWORD* pdwRespLenSAMAPDU)
	{
		DWORD	dwResponseBufferLength = 255;
		BYTE	lpResponseBuffer[255] = { 0 };

		PBYTE TDACommunicationAPDU = new BYTE[7+dwLenSAMAPDU];
		TDACommunicationAPDU[0] = 0xFF;
		TDACommunicationAPDU[1] = 0xF4;
		TDACommunicationAPDU[2] = 0x60;
		TDACommunicationAPDU[3] = (BYTE)(dwLenSAMAPDU >> 8);
		TDACommunicationAPDU[4] = (BYTE)dwLenSAMAPDU;
		TDACommunicationAPDU[5] = 0x00;
		memcpy(&TDACommunicationAPDU[6], lpSAMAPDU, dwLenSAMAPDU);
		TDACommunicationAPDU[6 + dwLenSAMAPDU] = CalcLRC(&TDACommunicationAPDU[2], 4 + dwLenSAMAPDU);

		LONG Status = PCSCExchange(TDACommunicationAPDU, 6 + dwLenSAMAPDU + 1, lpResponseBuffer, &dwResponseBufferLength);

		delete[] TDACommunicationAPDU;

		if ((Status == SCARD_S_SUCCESS)&&(dwResponseBufferLength >= 7))
		{
			BYTE ucCommandType = lpResponseBuffer[3];
			WORD wLength = ((WORD)lpResponseBuffer[1] << 8) | (WORD)lpResponseBuffer[2];
			if (lpResponseBuffer[0] == 0x60) //ACK
			{
				if (0x00 == CalcLRC(lpResponseBuffer, dwResponseBufferLength - 2))
				{
					if (ucCommandType == 0x00)
					{
						if (wLength >= 2)
						{
							BYTE ucSW1 = lpResponseBuffer[4 + wLength - 2];
							BYTE ucSW2 = lpResponseBuffer[4 + wLength - 1];

							if ((ucSW1 = 0x90)&&((ucSW2 = 0x00) || (ucSW2 = 0xAF)))
								Status = SCARD_S_SUCCESS;
							else
								Status = SCARD_E_UNEXPECTED;
					
							//CString item;
							int iPos = 4;
							//m_strTDAReceive = "";
							for( int i = 0; i < wLength; i++ )
							{
								//item.Format( " %02X", pResponseBuffer[iPos++] );
								pRespSAMAPDU[i] = lpResponseBuffer[iPos++];
								//m_strTDAReceive += item;
							}

							*pdwRespLenSAMAPDU = wLength;
						}
						else
							Status = SCARD_E_UNEXPECTED;
					}
					else
					{
						Status = SCARD_E_UNEXPECTED;
						//AddReport( true, "Incorrect TDA ALPAR LRC = ", &pResponseBuffer[4 + wLength], 1, 0 );
					}
				}
				else
				{
					Status = SCARD_E_UNEXPECTED;
					//AddReport( true, "TDA ALPAR Command failed, error code =", &pResponseBuffer[4], wLength, 0 );
				}
			}
			else
			{
				Status = SCARD_E_UNEXPECTED;
				//AddReport( true, "TDA ALPAR Command failed, error code =", &pResponseBuffer[4], wLength, 0 );
			}
		}
		else
		{
			Status = SCARD_E_UNEXPECTED;
			//AddReport( true, "TDA ALPAR Command failed, error code =", &pResponseBuffer[4], wLength, 0 );
		}

		return Status;
	}
	#else
	LONG CommandSAM(PBYTE lpSAMAPDU, DWORD dwLenSAMAPDU, PBYTE pRespSAMAPDU, DWORD* pdwRespLenSAMAPDU)
	{
		//printf(("\nCommandSAM"));

		//printf("\n");
		//for (int i = 0; i < dwLenSAMAPDU; i++)
		//	printf("%02X", lpSAMAPDU[i]);
		//printf("\n");

		LONG Status = PCSCExchange(lpSAMAPDU, dwLenSAMAPDU, pRespSAMAPDU, pdwRespLenSAMAPDU, m_hSCardActiveSAM);

		if (Status != SCARD_S_SUCCESS)
			return Status;

		if (*pdwRespLenSAMAPDU >= 2)
		{
			BYTE ucSW1 = pRespSAMAPDU[*pdwRespLenSAMAPDU - 2];
			BYTE ucSW2 = pRespSAMAPDU[*pdwRespLenSAMAPDU - 1];

			if ((ucSW1 = 0x90) && ((ucSW2 = 0x00) || (ucSW2 = 0xAF)))
				Status = SCARD_S_SUCCESS;
			else
				Status = SCARD_E_UNEXPECTED;
		}
		else
			Status = SCARD_E_UNEXPECTED;

		return Status;
	}
	#endif
#else
LONG CommandSAM(PBYTE lpSAMAPDU, DWORD dwLenSAMAPDU, PBYTE pRespSAMAPDU, DWORD* pdwRespLenSAMAPDU)
{
	LONG Status = CLSCRF_SAM_APDU(
		m_pReader,
		lpSAMAPDU,
		dwLenSAMAPDU,
		pRespSAMAPDU,
		pdwRespLenSAMAPDU
		);

	if (Status == SCARD_S_SUCCESS)
	{
		if (*pdwRespLenSAMAPDU >= 2)
		{
			BYTE ucSW1 = pRespSAMAPDU[*pdwRespLenSAMAPDU - 2];
			BYTE ucSW2 = pRespSAMAPDU[*pdwRespLenSAMAPDU - 1];

			if ((ucSW1 = 0x90) && ((ucSW2 = 0x00) || (ucSW2 = 0xAF)))
				Status = SCARD_S_SUCCESS;
			else
				Status = SCARD_E_UNEXPECTED;
		}
		else
			Status = SCARD_E_UNEXPECTED;
	}

	return Status;
}
#endif

#ifdef PCSC
LONG OpenTransparentSession()
{
	DWORD	dwResponseBufferLength = 255;
	BYTE	lpResponseBuffer[255] = { 0 };

	//FF C2 00 00 02 81 00 00
	unsigned char OPENSESSIONAPDU[] = {
		0xFF,/* APDU Class */
		0xC2,
		0x00,
		0x00,
		0x02,
		0x81,
		0x00,
		0x00
	};

	LONG Status = PCSCExchange(OPENSESSIONAPDU, sizeof(OPENSESSIONAPDU), lpResponseBuffer, &dwResponseBufferLength);

	return Status;
}
#else
LONG OpenTransparentSession()
{
	return SCARD_S_SUCCESS;
}
#endif

#ifdef PCSC
LONG CloseTransparentSession()
{
	DWORD	dwResponseBufferLength = 255;
	BYTE	lpResponseBuffer[255] = { 0 };

	//FF C2 00 00 02 82 00 00
	unsigned char CLOSESESSIONAPDU[] = {
		0xFF,/* APDU Class */
		0xC2,
		0x00,
		0x00,
		0x02,
		0x82,
		0x00,
		0x00
	};

	LONG Status = PCSCExchange(CLOSESESSIONAPDU, sizeof(CLOSESESSIONAPDU), lpResponseBuffer, &dwResponseBufferLength);

	return Status;
}
#else
LONG CloseTransparentSession()
{
	return SCARD_S_SUCCESS;
}
#endif

#ifdef PCSC
LONG EnterTCL()
{
	//if (m_fModeTCL)
	//	return SCARD_S_SUCCESS;

	DWORD	dwResponseBufferLength = 255;
	BYTE	lpResponseBuffer[255] = { 0 };

	//FF C2 00 02 04 8F 02 00 04 00
	unsigned char RATSPPSAPDU[] = {
		0xFF,/* APDU Class */
		0xC2,
		0x00,
		0x02,
		0x04,
		0x8F,
		0x02,
		0x00,
		0x04,
		0x00
	};

	LONG Status = PCSCExchange(RATSPPSAPDU, sizeof(RATSPPSAPDU), lpResponseBuffer, &dwResponseBufferLength);

	if (Status == SCARD_S_SUCCESS)
		m_fModeTCL = true;

	return Status;
}
#else
LONG EnterTCL()
{
	if (m_fModeTCL)
		return SCARD_S_SUCCESS;

	LONG Status;

	BYTE ATS[1024];

	Status = CLSCRF_ISO14443A_4_RATS(m_pReader, 0, ATS);
	if (Status != 0) return Status;

	Status = CLSCRF_ISO14443A_4_PPS(m_pReader, 0, 0);
	if (Status != 0) return Status;

	//Status = CLSCRF_Mfrc_Set_Rf_Mode(m_pReader, 0);
	//if (Status != 0) return Status;

	m_fModeTCL = true;

	return SCARD_S_SUCCESS;
}
#endif

#ifdef PCSC
LONG AuthenticateMFP_Part1(PBYTE lpResponseBuffer, PDWORD pdwResponseBufferLength)
{
	//13:12:57.625 T Command 48 06 00 0A 00 70 00 40 00 00 00 11 01
	//13:12:57.625 R 00 13 00 0A 00 90 C2 EA C1 42 73 15 05 1A C9 73 AE C9 A0 79 13 30

	//FF C2 00 01 0B 90 02 00 00 95 05 02 70 00 40 00 00
#ifdef PR
	unsigned char AUTHAPDU[] = {
		0xFF,/* APDU Class */
		0xC2,
		0x00,
		0x01,
		0x0B,
		0x90,
		0x02,
		0x00,
		0x00,
		0x95,
		0x05,
		0x02,
		0x70,
		//0x00,
		//0x40,//Sector 0 Block 0 //0x40,
		0x02,
		0x40,//Sector 1 Block 0 //0x40,
		0x00,
		0x00
	};
#else
	unsigned char AUTHAPDU[] = {
		0xFF,/* APDU Class */
		0xC2,
		0x00,
		0x01,
		0x0A,
		0x90,
		0x02,
		0x00,
		0x00,
		0x95,
		0x04,
		0x70,
		//0x00,
		//0x40,//Sector 0 Block 0 //0x40,
		0x02,
		0x40,//Sector 1 Block 0 //0x40,
		0x00,
		0x00
	};
#endif

	LONG Status = PCSCExchange(AUTHAPDU, sizeof(AUTHAPDU), lpResponseBuffer, pdwResponseBufferLength);

	return Status;
}
#else
LONG AuthenticateMFP_Part1(PBYTE lpResponseBuffer, PDWORD pdwResponseBufferLength)
{
	BYTE lpRequest[6] = { 0x0a, 0x00, 0x70, 0x00, 0x40, 0x00 };

	LONG Status = CLSCRF_DirectIO_Card(
		m_pReader,
		lpRequest,
		sizeof(lpRequest),
		1000,
		lpResponseBuffer,
		pdwResponseBufferLength
		);

	return Status;
}
#endif

#ifdef PCSC
LONG AuthenticateMFP_Part2(PBYTE lpResponseBuffer, PDWORD pdwResponseBufferLength)
{
	//13:17:34.843 T Command 48 23 00 0B 00 72 F3 CF 50 E6 6F 29 D0 A6 48 3E 64 A7 30 B5 AA 50 8B 57 90 AC B1 77 94 A7 BB 54 A0 F9 27 FB B3 9F 00 00 11 01
	//13:17:34.859 R 00 23 00 0B 00 90 BB 28 50 7B E7 61 1A FF 2B B1 DB D5 2A 85 6D F6 F1 93 D1 EE B2 66 01 36 0C 0A 73 91 0B 97 75 58

#ifdef PR
	unsigned char AUTHAPDU2[] = {
		0xFF,/* APDU Class */
		0xC2,
		0x00,
		0x01,
		0x28,
		0x90,
		0x02,
		0x00,
		0x00,
		0x95,//0x05,//0x95?
		0x22,
		0x03,
		0x72,

		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,

		0x00
	};
	memcpy(&AUTHAPDU2[13], lpResponseBuffer, 32);
#else
	unsigned char AUTHAPDU2[] = {
		0xFF,/* APDU Class */
		0xC2,
		0x00,
		0x01,
		0x27,
		0x90,
		0x02,
		0x00,
		0x00,
		0x95,//0x05,//0x95?
		0x21,
		0x72,

		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,

		0x00
	};
	memcpy(&AUTHAPDU2[12], lpResponseBuffer, 32);
#endif

	LONG Status = PCSCExchange(AUTHAPDU2, sizeof(AUTHAPDU2), lpResponseBuffer, pdwResponseBufferLength);

	return Status;
}
#else
LONG AuthenticateMFP_Part2(PBYTE lpResponseBuffer, PDWORD pdwResponseBufferLength)
{
	BYTE lpRequest[35] = { 0x0b, 0x00, 0x72, 
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		};

	memcpy(&lpRequest[3], lpResponseBuffer, 32);

	LONG Status = CLSCRF_DirectIO_Card(
		m_pReader,
		lpRequest,
		sizeof(lpRequest),
		1000,
		lpResponseBuffer,
		pdwResponseBufferLength
		);

	return Status;
}
#endif

LONG GetTranscieveFromTLV(PBYTE lpTLVBuffer, DWORD dwTLVBufferLength, PBYTE lpTranscieveBuffer, PDWORD pdwTranscieveLength)
{
	DWORD dwLengthTLV = 0;
	PBYTE lpStart = lpTLVBuffer;

	//int iPos = 0;
	while (lpTLVBuffer - lpStart < dwTLVBufferLength)
		//while (iPos < dwTLVBufferLength)
	{
		BYTE ucTypeTLV = *lpTLVBuffer;

		if (lpTLVBuffer[1] <= 0x7F)
		{
			dwLengthTLV = lpTLVBuffer[1];
			lpTLVBuffer += 1;
		}
		else if (lpTLVBuffer[1] == 0x81)
		{
			CopyMemory(
				&dwLengthTLV,
				&lpTLVBuffer[2],
				1
				);
			lpTLVBuffer += 2;
		}
		else if (lpTLVBuffer[1] == 0x82)
		{
			dwLengthTLV = lpTLVBuffer[2] * 256 + lpTLVBuffer[3];
			lpTLVBuffer += 3;
		}
		else if (lpTLVBuffer[1] == 0x83)
		{
			dwLengthTLV = (DWORD)lpTLVBuffer[2] * 256 * 256 + (DWORD)lpTLVBuffer[3] * 256 + (DWORD)lpTLVBuffer[4];
			lpTLVBuffer += 4;
		}
		else if (lpTLVBuffer[1] == 0x84)
		{
			dwLengthTLV = (DWORD)lpTLVBuffer[2] * 256 * 256 * 256 + (DWORD)lpTLVBuffer[3] * 256 * 256 + (DWORD)lpTLVBuffer[4] * 256 + (DWORD)lpTLVBuffer[5];
			lpTLVBuffer += 5;
		}
		else
		{
			return SCARD_E_UNSUPPORTED_FEATURE;
		}

		switch (ucTypeTLV)
		{
		case 0x92:
			lpTLVBuffer += 1 + dwLengthTLV;
			break;
		case 0x96:
			lpTLVBuffer += 1 + dwLengthTLV;
			break;
		case 0x97:
			if (dwLengthTLV > *pdwTranscieveLength)
				return SCARD_E_INSUFFICIENT_BUFFER;

			CopyMemory(
				lpTranscieveBuffer,
				&lpTLVBuffer[1],
				dwLengthTLV
				);
			*pdwTranscieveLength = dwLengthTLV;
			lpTLVBuffer += 1 + dwLengthTLV;
			return SCARD_S_SUCCESS;
			break;
		default:
			return SCARD_E_UNSUPPORTED_FEATURE;
		}
	}

	return SCARD_E_NOT_TRANSACTED;
}

#ifdef SAM
	
	#define SAM_BUF_SIZE 256
	BYTE SAM_Plain[SAM_BUF_SIZE];
	BYTE SAM_Coded[SAM_BUF_SIZE];
	BYTE SessionKeyEnc[16];
	BYTE SessionKeyMAC[16];
	#define UEM_SAM_BYTECOUNTERR		0xC5
	#define UEM_SAM_APDU_ERR			0xC8
	#define UEM_SAM_INVALID_CARD_MAC	0xC7
	#define UEM_SAM_AUTHENTICATION_ERR	0xC6
	#define MI_WRONG_PARAMETER_VALUE	0xC4

	char SAM_AuthenticateHost(
		BYTE LogicalChannel,
		BYTE* Kx,
		BYTE KeyNo,
		BYTE KeyVer,
		BYTE HostMode,
		BYTE* SW)
	{
		int i, j;
		DWORD SendLen, RecvLen = SAM_BUF_SIZE;
		//    volatile 
		BYTE* ToSM;
		BYTE* From;
		BYTE  SV1[16];
		BYTE  SV2[16];
		BYTE  MAC16[16];
		BYTE  MAC24[24];
		BYTE  Kxe[24];
		char status = 0;

		ToSM = SAM_Plain;
		From = SAM_Coded;
		ToSM[0] = 0x80 | LogicalChannel;
		ToSM[1] = 0xa4;
		ToSM[2] = 0x00; // P1
		ToSM[3] = 0x00; // P2
		ToSM[4] = 0x03; // Lc
		ToSM[5] = KeyNo;
		ToSM[6] = KeyVer;
		ToSM[7] = HostMode;
		SendLen = 8;
		ToSM[SendLen++] = 0x00;    // Le

		RecvLen = SAM_BUF_SIZE;

		status = CommandSAM(
			ToSM, 
			SendLen, 
			From, 
			&RecvLen);

		memcpy(SW, &From[RecvLen - 2], 2);
		if (status)
			return status;
		if (RecvLen < 2)
			return (char)UEM_SAM_BYTECOUNTERR;
		if ((From[RecvLen - 2] != 0x90) || (From[RecvLen - 1] != 0xAF))
			return (char)UEM_SAM_APDU_ERR;

		memcpy(&SV1[5], &From[7], 5);
		memcpy(&SV1[10], &From[0], 5);

		From[12] = HostMode;
		From[13] = 0;
		From[14] = 0;
		From[15] = 0;
	
		AES_128_CBC_En_MAC(From, 16, Kx, MAC16);
	
		ToSM = &SAM_Plain[5];

		for (i = 0, j = 1; i < 8; i++, j += 2)
			ToSM[i] = MAC16[j];
	
		// Rnd1 p.6 to SAM
		ToSM[8] = 0x00;
		ToSM[9] = 0x01;
		ToSM[10] = 0x02;
		ToSM[11] = 0x03;
		ToSM[12] = 0x04;
		ToSM[13] = 0x05;
		ToSM[14] = 0x06;
		ToSM[15] = 0x07;
		ToSM[16] = 0x08;
		ToSM[17] = 0x09;
		ToSM[18] = 0x0A;
		ToSM[19] = 0x0B;

		memcpy(&SV1[0], &ToSM[15], 5);
		for (i = 0; i < 5; i++)
			SV1[10 + i] ^= ToSM[8 + i];
		SV1[15] = 0x91;

		aes_128_encrypt16(SV1, Kx, Kxe);
	
		// p.7
		ToSM = SAM_Plain;
		ToSM[0] = 0x80 | LogicalChannel;
		ToSM[1] = 0xa4;
		ToSM[2] = 0x00; // P1
		ToSM[3] = 0x00; // P2
		ToSM[4] = 0x14; // Lc
		SendLen = 5 + 20;
		ToSM[SendLen++] = 0x00;    // Le

		RecvLen = SAM_BUF_SIZE;
		status = CommandSAM(
			ToSM,
			SendLen,
			From,
			&RecvLen);

		memcpy(SW, &From[RecvLen - 2], 2);
		if (status)
			return status;
		if (RecvLen < 2)
			return (char)UEM_SAM_BYTECOUNTERR;
		if ((From[RecvLen - 2] != 0x90) || (From[RecvLen - 1] != 0xAF))
			return (char)UEM_SAM_APDU_ERR;

		ToSM = &SAM_Plain[13];
		ToSM[12] = HostMode;
		ToSM[13] = 0;
		ToSM[14] = 0;
		ToSM[15] = 0;

		AES_128_CBC_En_MAC(ToSM, 16, Kx, MAC16);

		for (i = 0, j = 1; i < 8; i++, j += 2)
			ToSM[i] = MAC16[j];

		if (memcmp(ToSM, From, 8))
			return (char)UEM_SAM_INVALID_CARD_MAC;

		aes_128_decrypt16(&From[8], Kxe, &From[24]);

		// Load RndB for Ke & Km
		memcpy(&SV1[5], &From[24], 5);
		memcpy(&SV1[10], &From[31], 5);
		memcpy(&SV2[5], &From[28], 5);
		memcpy(&SV2[10], &From[35], 5);

		// RndB''
		From[40] = From[24];
		From[41] = From[25];
		From = &SAM_Coded[26];

		// RndA p.14 to SAM
		From = &SAM_Coded[10];
		From[0] = 0x00;
		From[1] = 0x01;
		From[2] = 0x02;
		From[3] = 0x03;
		From[4] = 0x04;
		From[5] = 0x05;
		From[6] = 0x06;
		From[7] = 0x07;
		From[8] = 0x08;
		From[9] = 0x09;
		From[10] = 0x0A;
		From[11] = 0x0B;
		From[12] = 0x0C;
		From[13] = 0x0D;
		From[14] = 0x0E;
		From[15] = 0x0F;
		// Load RndA for Ke & Km
		memcpy(SV1, &From[0], 5);
		for (i = 0; i < 5; i++)
			SV1[10 + i] ^= From[7 + i];
		SV1[15] = 0x81;
		memcpy(SV2, &From[4], 5);
		for (i = 0; i < 5; i++)
			SV2[10 + i] ^= From[11 + i];
		SV2[15] = 0x82;

		for (i = 0; i < 16; i++)
			MAC16[i] = 0;
		ToSM = &SAM_Plain[5];
		aes_128_encrypt_cbc(
			From,
			32,
			Kxe,
			MAC16,
			ToSM);

		ToSM = &SAM_Plain[0];
		ToSM[0] = 0x80 | LogicalChannel;
		ToSM[1] = 0xa4;
		ToSM[2] = 0x00; // P1
		ToSM[3] = 0x00; // P2
		ToSM[4] = 0x20; // Lc
		SendLen = 5 + 32;
		ToSM[SendLen++] = 0x00;    // Le
		From = &SAM_Coded[28];

		RecvLen = SAM_BUF_SIZE;
		status = CommandSAM(
			ToSM,
			SendLen,
			From,
			&RecvLen);

		memcpy(SW, &From[RecvLen - 2], 2);
		if (status)
			return status;
		if (RecvLen < 2)
			return (char)UEM_SAM_BYTECOUNTERR;
		if ((From[RecvLen - 2] != 0x90) || (From[RecvLen - 1] != 0x00))
			return (char)UEM_SAM_APDU_ERR;

		aes_128_decrypt16(From, Kxe, MAC16);

		From = &SAM_Coded[10];
		From[16] = From[0];
		From[17] = From[1];
	
		if (memcmp(&From[2], MAC16, 16))
			return (char)UEM_SAM_AUTHENTICATION_ERR;

		aes_128_encrypt16(SV1, Kx, SessionKeyEnc);
		aes_128_encrypt16(SV2, Kx, SessionKeyMAC);

		return status;
	}//SAM_AuthenticateHost

	LONG SAM_Test_PKI(PBYTE lpResponseBuffer, PDWORD pdwResponseBufferLength)
	{
		BYTE SW[2] = { 0 };
		BYTE Key[16] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
		SAM_AuthenticateHost(
			0,//BYTE LogicalChannel,
			Key,
			0,//BYTE KeyNo,
			0,//BYTE KeyVer,
			0,//BYTE HostMode,
			SW);//BYTE* SW);

		unsigned char SAMAPDU[] = { 0x80, 0x15, 0x01, 0x00, 0x0E, 0x00, 0x00, 0x63, 0x00, 0x00, 0xFF, 0x00, 0x30, 0x00, 0x04, 0x00, 0x01, 0x00, 0x01 };

		getchar();

		LONG Status = CommandSAM(SAMAPDU, sizeof(SAMAPDU), lpResponseBuffer, pdwResponseBufferLength);

		getchar();

		return Status;
	}

	LONG SAM_AuthenticateMFP_Part1(PBYTE lpResponseBuffer, PDWORD pdwResponseBufferLength)
	{
		//SAM_Test_PKI(lpResponseBuffer, pdwResponseBufferLength);
		//13:14:56.765 T SAM_APDU 80 A3 0C 00 12 01 00 C2 EA C1 42 73 15 05 1A C9 73 AE C9 A0 79 13 30 00
		//13:14:56.781 R F3 CF 50 E6 6F 29 D0 A6 48 3E 64 A7 30 B5 AA 50 8B 57 90 AC B1 77 94 A7 BB 54 A0 F9 27 FB B3 9F 90 AF
		LONG Status;

		unsigned char SAMAPDU[] = {
			0x80,// APDU Class
			0xA3,
			0x0C,
			0x00,
			0x12,
			0x01,//KeyNr //0x03, 
			0x00,//KeyVer //0xFF,

			0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
			0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,

			0x00
		};

	#ifdef PCSC
		BYTE lpTranscieveBuffer[64] = { 0 };
		DWORD dwTranscieveLength = sizeof(lpTranscieveBuffer);

		Status = GetTranscieveFromTLV(&lpResponseBuffer[5], *pdwResponseBufferLength - 5, lpTranscieveBuffer, &dwTranscieveLength);

		if (Status != SCARD_S_SUCCESS)
			return Status;
		//memcpy(&SAMAPDU[7], &lpResponseBuffer[18], 16);
		#ifdef PR
			memcpy(&SAMAPDU[7], &lpTranscieveBuffer[2], 16);
		#else
			memcpy(&SAMAPDU[7], &lpTranscieveBuffer[1], 16);
		#endif
	#else
		memcpy(&SAMAPDU[7], &lpResponseBuffer[3], 16);
	#endif
		
		
		/*unsigned char SAMAPDU[] = {
			0x80,// APDU Class
			0x15,
			0x01,
			0x00,
			0x0E,
			0x00,//KeyNr //0x03, 
			0x00,//KeyVer //0xFF,
			0x63, 0x00, 0x00, 0xFF, 0x00, 0x30, 0x00, 0x04, 0x00, 0x01, 0x00, 0x01
		};*/

		*pdwResponseBufferLength = 255;

		Status = CommandSAM(SAMAPDU, sizeof(SAMAPDU), lpResponseBuffer, pdwResponseBufferLength);

		return Status;
	}

	LONG SAM_AuthenticateMFP_Part2(PBYTE lpResponseBuffer, PDWORD pdwResponseBufferLength)
	{
		//13:19:35.625 T SAM_APDU 80 A3 00 00 20 BB 28 50 7B E7 61 1A FF 2B B1 DB D5 2A 85 6D F6 F1 93 D1 EE B2 66 01 36 0C 0A 73 91 0B 97 75 58 00
		//13:19:35.640 R 00 00 00 00 00 00 00 00 00 00 00 00 90 00
		LONG Status;

		unsigned char SAMAPDU2[] = {
			0x80,/* APDU Class */
			0xA3,
			0x00,
			0x00,
			0x20,

			0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
			0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
			0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
			0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,

			0x00
		};

	#ifdef PCSC
		BYTE lpTranscieveBuffer[64] = { 0 };
		DWORD dwTranscieveLength = sizeof(lpTranscieveBuffer);

		Status = GetTranscieveFromTLV(&lpResponseBuffer[5], *pdwResponseBufferLength - 5, lpTranscieveBuffer, &dwTranscieveLength);

		if (Status != SCARD_S_SUCCESS)
			return Status;

		//memcpy(&SAMAPDU2[5], &lpResponseBuffer[1], 32);
		#ifdef PR
			memcpy(&SAMAPDU2[5], &lpTranscieveBuffer[2], 32);
		#else
			memcpy(&SAMAPDU2[5], &lpTranscieveBuffer[1], 32);
		#endif
	#else
		memcpy(&SAMAPDU2[5], &lpResponseBuffer[3], 32);
	#endif

		*pdwResponseBufferLength = 255;

		Status = CommandSAM(SAMAPDU2, sizeof(SAMAPDU2), lpResponseBuffer, pdwResponseBufferLength);

		return Status;
	}
#else
	void MfPlusAddressFromType(BYTE Type, BYTE Sector, BYTE Block, BYTE* MfPAddr)
	{
		BYTE BlockAddr;

		MfPAddr[1] = Type & 0xf0;
		if (Type == 0x00)
		{
			BlockAddr = 4 * Sector;
			if (Sector >= 0x20)
				BlockAddr += 12 * (Sector - 0x20);
			MfPAddr[0] = BlockAddr + Block;
		}
		else if (Type < 0x50)
		{
			MfPAddr[0] = Sector * 2 + (Type & 1);
		}
		else
		{
			if ((Type == 0xa2) || (Type == 0xa3))
				MfPAddr[0] = 0x80 | (Type & 1);
			else
				MfPAddr[0] = Type & 0x0f;
		}
	}

	DWORD TransactID;
	WORD W_Ctr;
	WORD R_Ctr;

	void MfPlusAES128InitVector(BYTE Type, BYTE* InV)
	{
		int i;
		if( Type == 0 )         // First Authenticate
		{
			for( i = 0; i < 16; i++ )
				InV[i] = 0;
		}
		else if (Type == 1)    // for Command
		{
			memcpy(&InV[0], &TransactID, 4);
			memcpy(&InV[4], &R_Ctr, 2);
			memcpy(&InV[6], &W_Ctr, 2);
			memcpy(&InV[8], &R_Ctr, 2);
			memcpy(&InV[10], &W_Ctr, 2);
			memcpy(&InV[12], &R_Ctr, 2);
			memcpy(&InV[14], &W_Ctr, 2);
		}
		else // if( Type == 2 ) // for Response
		{
			memcpy(&InV[0], &R_Ctr, 2);
			memcpy(&InV[2], &W_Ctr, 2);
			memcpy(&InV[4], &R_Ctr, 2);
			memcpy(&InV[6], &W_Ctr, 2);
			memcpy(&InV[8], &R_Ctr, 2);
			memcpy(&InV[10], &W_Ctr, 2);
			memcpy(&InV[12], &TransactID, 4);
		}
	}

	//BYTE lpKey[16] = { 0 };
	BYTE lpInitVector[16] = { 0 };
	BYTE lpRndA[16] = { 0 };

	LONG Direct_AuthenticateMFP_Part1(PBYTE lpResponseBuffer, PDWORD pdwResponseBufferLength)
	{
		TransactID = 0;
		W_Ctr = 0;
		R_Ctr = 0;

		BYTE lpRndB[16] = { 0 };
		BYTE lpRndBB[16] = { 0 };
		BYTE lpRndARndBB[32] = { 0 };

		MfPlusAES128InitVector(2, lpInitVector);

#ifdef PCSC
		BYTE lpTranscieveBuffer[64] = { 0 };
		DWORD dwTranscieveLength = sizeof(lpTranscieveBuffer);

		LONG Status = GetTranscieveFromTLV(&lpResponseBuffer[5], *pdwResponseBufferLength - 5, lpTranscieveBuffer, &dwTranscieveLength);

		if (Status != SCARD_S_SUCCESS)
			return Status;

		AES_128_CBC_De(
			//&lpResponseBuffer[18],
	#ifdef PR
			&lpTranscieveBuffer[2],
	#else
			&lpTranscieveBuffer[1],
	#endif
			1,
			lpKey,    // Key
			lpInitVector,    // InV
			lpRndB);
#else
		AES_128_CBC_De(
			&lpResponseBuffer[3],
			1,
			lpKey,    // Key
			lpInitVector,    // InV
			lpRndB);
#endif

		srand(GetTickCount());
		for (int i = 0; i < 16; i++)
			lpRndA[i] = rand();

		memcpy(lpRndBB, lpRndB + 1, 15);
		lpRndBB[15] = lpRndB[0];
		memcpy(lpRndARndBB, lpRndA, 16);
		memcpy(lpRndARndBB + 16, lpRndBB, 16);

		MfPlusAES128InitVector(1, lpInitVector);
		AES_128_CBC_En(
			lpRndARndBB, 
			2,
			lpKey,    // Key
			lpInitVector,    // InV
			lpResponseBuffer);

		*pdwResponseBufferLength = 32;

		return 0;
	}

	LONG Direct_AuthenticateMFP_Part2(PBYTE lpResponseBuffer, PDWORD pdwResponseBufferLength)
	{
#ifdef PCSC
		BYTE lpTranscieveBuffer[64] = { 0 };
		DWORD dwTranscieveLength = sizeof(lpTranscieveBuffer);

		LONG Status = GetTranscieveFromTLV(&lpResponseBuffer[5], *pdwResponseBufferLength - 5, lpTranscieveBuffer, &dwTranscieveLength);

		if (Status != SCARD_S_SUCCESS)
			return Status;

		AES_128_CBC_De(
			//&lpResponseBuffer[18],
	#ifdef PR
			&lpTranscieveBuffer[2],
	#else
			&lpTranscieveBuffer[1],
	#endif
			2,
			lpKey,    // Key
			lpInitVector,    // InV
			lpResponseBuffer);
#else
		AES_128_CBC_De(
			&lpResponseBuffer[3],
			2,
			lpKey,    // Key
			lpInitVector,    // InV
			lpResponseBuffer);
#endif

		BYTE ResRndAA[16] = { 0 };
		BYTE ResRndA[16] = { 0 };

		memcpy(ResRndAA, lpResponseBuffer + 4, 16);
		memcpy(ResRndA + 1, ResRndAA, 15);
		ResRndA[0] = ResRndAA[15];

		*pdwResponseBufferLength = 32;

		if (memcmp(ResRndA, lpRndA, 16) != 0)
			return SCARD_E_UNEXPECTED;

		//memcpy(&TransactID, &CryptoCoded[1], 4);
		//CryptoCoded[4] = CryptoCoded[20];
		//n = 4;
		//for (int i = 0; i < 16; i++)
		//	if (CryptoCoded[i + n] != i + 0xa1)
		//		return MI_MFP_AUTHENTICATION_ERR;

		return 0;
	}
#endif

#ifdef PCSC // PCSC Reader
	#ifdef SAM // PCSC SAM
	LONG AuthenticateMFP()
	{
		LONG Status = 0;

		//A5A4A3A2A1A018778E69F0F0F0F0F0F0
	
		DWORD	dwResponseBufferLength = 255;
		BYTE	lpResponseBuffer[255] = { 0 };

		printf(("\n"));

		printf(("\tAuthenticateMFP_Part1...\t"));
		Status = AuthenticateMFP_Part1(lpResponseBuffer, &dwResponseBufferLength);
		printf(("0x%08X"), Status);
		printTime();

		if (Status != SCARD_S_SUCCESS)
		{
			printf(("\tAuthentication failed.\t\t"));

			return SCARD_E_UNEXPECTED;
		}
	
		BYTE lpTranscieveBuffer[64] = { 0 };
		DWORD dwTranscieveLength = sizeof(lpTranscieveBuffer);

		Status = GetTranscieveFromTLV(&lpResponseBuffer[5], dwResponseBufferLength - 5, lpTranscieveBuffer, &dwTranscieveLength);

		if (Status != SCARD_S_SUCCESS)
		{
			printf(("\tAuthentication failed.\t\t"));

			return Status;
		}

//#ifdef PR
//		if (lpTranscieveBuffer[0x01] == 0x06)
//#else
//		if (lpTranscieveBuffer[0x00] == 0x06)
//#endif
//			return SCARD_E_UNEXPECTED;

		//if (lpResponseBuffer[0x11] != 0x90)
#ifdef PR
		if (lpTranscieveBuffer[0x01] != 0x90)
#else
		if (lpTranscieveBuffer[0x00] != 0x90)
#endif
			return SCARD_E_UNEXPECTED;


		//printf("\n");
		//printf(("dwTranscieveLength = %d "), dwTranscieveLength);
		//for (int i = 0; i < dwTranscieveLength; i++)
		//	printf("%02X", lpTranscieveBuffer[i]);
		//printf("\n");

		printf(("\tSAM_AuthenticateMFP_Part1...\t"));
		Status = SAM_AuthenticateMFP_Part1(lpResponseBuffer, &dwResponseBufferLength);
		printf(("0x%08X"), Status);
		printTime();

		if (Status != SCARD_S_SUCCESS)
		{
			printf(("\tAuthentication failed.\t\t"));

			return SCARD_E_UNEXPECTED;
		}

		//printf("\n");
		//printf(("dwResponseBufferLength = %d "), dwResponseBufferLength);
		//for (int i = 0; i < dwResponseBufferLength; i++)
		//	printf("%02X", lpResponseBuffer[i]);
		//printf("\n");

		dwResponseBufferLength = 255;

		printf(("\tAuthenticateMFP_Part2...\t"));
		Status = AuthenticateMFP_Part2(lpResponseBuffer, &dwResponseBufferLength);
		//if ((Status == SCARD_S_SUCCESS) && (lpResponseBuffer[0x11] == 0x06))
		if (Status == SCARD_S_SUCCESS)
		{
			BYTE lpTranscieveBuffer[64] = { 0 };
			DWORD dwTranscieveLength = sizeof(lpTranscieveBuffer);

			LONG Status = GetTranscieveFromTLV(&lpResponseBuffer[5], dwResponseBufferLength - 5, lpTranscieveBuffer, &dwTranscieveLength);

			if (Status != SCARD_S_SUCCESS)
			{
				printf(("\tAuthentication failed.\t\t"));

				return Status;
			}

//			//if (lpResponseBuffer[0x11] == 0x06)
//#ifdef PR
//			if (lpTranscieveBuffer[0x01] == 0x06)
//#else
//			if (lpTranscieveBuffer[0x00] == 0x06)
//#endif
//				Status = SCARD_E_UNEXPECTED;

			//if (lpResponseBuffer[0x11] != 0x90)
#ifdef PR
			if (lpTranscieveBuffer[0x01] != 0x90)
#else
			if (lpTranscieveBuffer[0x00] != 0x90)
#endif
			{
				printf(("dwTranscieveLength = %d"), dwTranscieveLength);
				printTime();

				printf("\n");
				for (int i = 0; i < dwTranscieveLength; i++)
					printf("%02X", lpTranscieveBuffer[i]);
				printf("\n");

				return SCARD_E_UNEXPECTED;
			}
		}
		printf(("0x%08X"), Status);
		printTime();

		if (Status != SCARD_S_SUCCESS)
		{
			printf(("\tAuthentication failed.\t\t"));

			return SCARD_E_UNEXPECTED;
		}

		//dwResponseBufferLength = 255;

		printf(("\tSAM_AuthenticateMFP_Part2...\t"));
		Status = SAM_AuthenticateMFP_Part2(lpResponseBuffer, &dwResponseBufferLength);
		printf(("0x%08X"), Status);
		printTime();

		if (Status != SCARD_S_SUCCESS)
		{
			printf(("\tAuthentication failed.\t\t"));

			return SCARD_E_UNEXPECTED;
		}

		printf(("\tAuthentication successful!\t"));

		return Status;
	}
	#else // PCSC WITHOUT SAM
	LONG AuthenticateMFP()
	{
		LONG Status = 0;

		//A5A4A3A2A1A018778E69F0F0F0F0F0F0

		DWORD	dwResponseBufferLength = 255;
		BYTE	lpResponseBuffer[255] = { 0 };

		printf(("\n"));

		printf(("\tAuthenticateMFP_Part1...\t"));
		Status = AuthenticateMFP_Part1(lpResponseBuffer, &dwResponseBufferLength);
		//if ((Status == SCARD_S_SUCCESS) && (lpResponseBuffer[0x11] == 0x06))
		if (Status == SCARD_S_SUCCESS)
		{
			BYTE lpTranscieveBuffer[64] = { 0 };
			DWORD dwTranscieveLength = sizeof(lpTranscieveBuffer);

			LONG Status = GetTranscieveFromTLV(&lpResponseBuffer[5], dwResponseBufferLength - 5, lpTranscieveBuffer, &dwTranscieveLength);

			if (Status != SCARD_S_SUCCESS)
			{
				printf(("\tAuthentication failed.\t\t"));

				return Status;
			}

			//printf(("\nlpTranscieveBuffer:"));
			//for (int i = 0; i < 16; i++)
			//	printf("%02X", lpTranscieveBuffer[i++]);
			//printf(("\n"));

			//if (lpResponseBuffer[0x11] == 0x06)
#ifdef PR
			if (lpTranscieveBuffer[0x01] == 0x06)
#else
			if (lpTranscieveBuffer[0x00] == 0x06)
#endif
				Status = SCARD_E_UNEXPECTED;
		}
		printf(("0x%08X"), Status);
		printTime();

		if (Status != SCARD_S_SUCCESS)
		{
			printf(("\tAuthentication failed.\t\t"));

			return Status;
		}

		dwResponseBufferLength = 255;

		printf(("\tDirect_AuthenticateMFP_Part1...\t"));
		Status = Direct_AuthenticateMFP_Part1(lpResponseBuffer, &dwResponseBufferLength);
		printf(("0x%08X"), Status);
		printTime();

		if (Status != SCARD_S_SUCCESS)
		{
			printf(("\tAuthentication failed.\t\t"));

			return Status;
		}

		dwResponseBufferLength = 255;

		printf(("\tAuthenticateMFP_Part2...\t"));
		Status = AuthenticateMFP_Part2(lpResponseBuffer, &dwResponseBufferLength);
		//if ((Status == SCARD_S_SUCCESS) && (lpResponseBuffer[0x11] == 0x06))
		if (Status == SCARD_S_SUCCESS)
		{
			BYTE lpTranscieveBuffer[64] = { 0 };
			DWORD dwTranscieveLength = sizeof(lpTranscieveBuffer);

			LONG Status = GetTranscieveFromTLV(&lpResponseBuffer[5], dwResponseBufferLength - 5, lpTranscieveBuffer, &dwTranscieveLength);

			if (Status != SCARD_S_SUCCESS)
			{
				printf(("\tAuthentication failed.\t\t"));

				return Status;
			}
			//if (lpResponseBuffer[0x11] == 0x06)
#ifdef PR
			if (lpTranscieveBuffer[0x01] == 0x06)
#else
			if (lpTranscieveBuffer[0x00] == 0x06)
#endif
				Status = SCARD_E_UNEXPECTED;
		}
		printf(("0x%08X"), Status);
		printTime();

		if (Status != SCARD_S_SUCCESS)
		{
			printf(("\tAuthentication failed.\t\t"));

			return SCARD_E_UNEXPECTED;
		}

		dwResponseBufferLength = 255;

		printf(("\tDirect_AuthenticateMFP_Part2...\t"));
		Status = Direct_AuthenticateMFP_Part2(lpResponseBuffer, &dwResponseBufferLength);
		printf(("0x%08X"), Status);
		printTime();

		if (Status != SCARD_S_SUCCESS)
		{
			printf(("\tAuthentication failed.\t\t"));

			return SCARD_E_UNEXPECTED;
		}

		printf(("\tAuthentication successful!\t"));

		return Status;
	}
	#endif
#else // STANDARD UEM USB NON-PCSC READER 
	#ifdef SAM // NON-PCSC READER SAM
		#ifdef MANUAL // NON-PCSC READER SAM MANUAL
		LONG AuthenticateMFP()
		{
			LONG Status = 0;

			//A5A4A3A2A1A018778E69F0F0F0F0F0F0

			DWORD	dwResponseBufferLength = 255;
			BYTE	lpResponseBuffer[255] = { 0 };

			printf(("\n"));

			printf(("\tAuthenticateMFP_Part1...\t"));
			Status = AuthenticateMFP_Part1(lpResponseBuffer, &dwResponseBufferLength);
			printf(("0x%08X"), Status);
			printTime();

			if (Status != SCARD_S_SUCCESS)
			{
				printf(("\tAuthentication failed.\t\t"));

				return SCARD_E_UNEXPECTED;
			}

			dwResponseBufferLength = 255;

			printf(("\tSAM_AuthenticateMFP_Part1...\t"));
			Status = SAM_AuthenticateMFP_Part1(lpResponseBuffer, &dwResponseBufferLength);
			printf(("0x%08X"), Status);
			printTime();

			if (Status != SCARD_S_SUCCESS)
			{
				printf(("\tAuthentication failed.\t\t"));

				return SCARD_E_UNEXPECTED;
			}

			dwResponseBufferLength = 255;

			printf(("\tAuthenticateMFP_Part2...\t"));
			Status = AuthenticateMFP_Part2(lpResponseBuffer, &dwResponseBufferLength);
			if ((Status == SCARD_S_SUCCESS) && (lpResponseBuffer[0x02] != 0x90))
			{
				Status = SCARD_E_UNEXPECTED;
			}
			printf(("0x%08X"), Status);
			printTime();

			if (Status != SCARD_S_SUCCESS)
			{
				printf(("\tAuthentication failed.\t\t"));

				return SCARD_E_UNEXPECTED;
			}

			dwResponseBufferLength = 255;

			printf(("\tSAM_AuthenticateMFP_Part2...\t"));
			Status = SAM_AuthenticateMFP_Part2(lpResponseBuffer, &dwResponseBufferLength);
			printf(("0x%08X"), Status);
			printTime();

			if (Status != SCARD_S_SUCCESS)
			{
				printf(("\tAuthentication failed.\t\t"));

				return SCARD_E_UNEXPECTED;
			}

			printf(("\tAuthentication successful!\t"));

			return Status;
		}
		#else // NON-PCSC READER SAM BY UEM PROPROETARY COMMAND
		LONG AuthenticateMFP()
		{
			BYTE ucSW1, ucSW2;
			BYTE PCDcap2[6] = { 0 };
			BYTE Div[256] = { 0 };

			printf(("\n"));

			printf(("\tSAM_MifarePlus_Authenticate...\t"));
			LONG Status = CLSCRF_SAM_MifarePlus_Authenticate(
				m_pReader,
				0x00,
				0x01,
				0x03,
				0x4A,
				0x00,
				0x01,
				0x00,
				0x00,
				PCDcap2,
				0x00,
				Div,
				&ucSW1,
				&ucSW2
				);
			printf(("0x%08X"), Status);
			printTime();

			if (Status != SCARD_S_SUCCESS)
			{
				printf(("\tAuthentication failed.\t\t"));

				return SCARD_E_UNEXPECTED;
			}

			printf(("\tAuthentication successful!\t"));

			return Status;
		}
		#endif
	#else // NON-PCSC READER WITHOUT SAM
		#ifdef MANUAL // NON-PCSC READER WITHOUT SAM MANUAL
		LONG AuthenticateMFP()
		{
			LONG Status = 0;

			//A5A4A3A2A1A018778E69F0F0F0F0F0F0

			DWORD	dwResponseBufferLength = 255;
			BYTE	lpResponseBuffer[255] = { 0 };

			printf(("\n"));

			printf(("\tAuthenticateMFP_Part1...\t"));
			Status = AuthenticateMFP_Part1(lpResponseBuffer, &dwResponseBufferLength);
			printf(("0x%08X"), Status);
			printTime();

			if (Status != SCARD_S_SUCCESS)
			{
				printf(("\tAuthentication failed.\t\t"));

				return SCARD_E_UNEXPECTED;
			}

			dwResponseBufferLength = 255;

			printf(("\tSAM_AuthenticateMFP_Part1...\t"));
			Status = Direct_AuthenticateMFP_Part1(lpResponseBuffer, &dwResponseBufferLength);
			printf(("0x%08X"), Status);
			printTime();

			if (Status != SCARD_S_SUCCESS)
			{
				printf(("\tAuthentication failed.\t\t"));

				return SCARD_E_UNEXPECTED;
			}

			dwResponseBufferLength = 255;

			printf(("\tAuthenticateMFP_Part2...\t"));
			Status = AuthenticateMFP_Part2(lpResponseBuffer, &dwResponseBufferLength);
			if ((Status == SCARD_S_SUCCESS) && (lpResponseBuffer[0x02] != 0x90))
			{
				Status = SCARD_E_UNEXPECTED;
			}
			printf(("0x%08X"), Status);
			printTime();

			if (Status != SCARD_S_SUCCESS)
			{
				printf(("\tAuthentication failed.\t\t"));

				return SCARD_E_UNEXPECTED;
			}

			dwResponseBufferLength = 255;

			printf(("\tSAM_AuthenticateMFP_Part2...\t"));
			Status = Direct_AuthenticateMFP_Part2(lpResponseBuffer, &dwResponseBufferLength);
			printf(("0x%08X"), Status);
			printTime();

			if (Status != SCARD_S_SUCCESS)
			{
				printf(("\tAuthentication failed.\t\t"));

				return SCARD_E_UNEXPECTED;
			}

			printf(("\tAuthentication successful!\t"));

			return Status;
		}
		#else // NON-PCSC READER WITHOUT SAM BY UEM PROPROETARY COMMAND
		LONG AuthenticateMFP()
		{
			BYTE PCDcap2[6] = { 0 };

			printf(("\n"));

			printf(("\tMifarePlus_Authenticate...\t"));
			LONG Status = CLSCRF_MifarePlus_Authenticate(
				m_pReader,
				0x01,
				0x4A,
				0x00,
				0x02,
				0x00,
				PCDcap2);
			printf(("0x%08X"), Status);
			printTime();

			if (Status != SCARD_S_SUCCESS)
			{
				printf(("\tAuthentication failed.\t\t"));

				return SCARD_E_UNEXPECTED;
			}

			printf(("\tAuthentication successful!\t"));

			return Status;
		}
		#endif
	#endif
#endif

#ifdef PCSC
LONG ReadMFP()
{
	DWORD	dwResponseBufferLength = 255;
	BYTE	lpResponseBuffer[255] = { 0 };

	//FF C2 00 01 0B 90 02 00 00 95 05 02 36 00 00 01 00
#ifdef PR
	unsigned char READAPDU[] = {
		0xFF,/* APDU Class */
		0xC2,
		0x00,
		0x01,
		0x0B,
		0x90,
		0x02,
		0x00,
		0x00,
		0x95,
		0x05,
		0x02,
		0x36,
		//0x00, // Sector 0, block 0
		0x04, // Sector 1, block 0
		0x00,
		0x01,
		0x00
	};
#else
	unsigned char READAPDU[] = {
		0xFF,/* APDU Class */
		0xC2,
		0x00,
		0x01,
		0x0A,
		0x90,
		0x02,
		0x00,
		0x00,
		0x95,
		0x04,
		0x36,
		//0x00, // Sector 0, block 0
		0x04, // Sector 1, block 0
		0x00,
		0x01,
		0x00
	};
#endif

	LONG Status = PCSCExchange(READAPDU, sizeof(READAPDU), lpResponseBuffer, &dwResponseBufferLength);

	if (Status != SCARD_S_SUCCESS)
		return Status;

	//printf(("dwResponseBufferLength = %d"), dwResponseBufferLength);
	//printTime();

	//printf("\n");
	//for (int i = 0; i < dwResponseBufferLength; i++)
	//	printf("%02X", lpResponseBuffer[i]);
	//printf("\n");

	//if (dwResponseBufferLength < 0x0F + 2)
	if (dwResponseBufferLength < 0x05 + 2)
		return SCARD_E_UNEXPECTED;

	//printf(("dwResponseBufferLength OK: 0x%08X"), Status);
	//printTime();

	BYTE lpTranscieveBuffer[64] = { 0 };
	DWORD dwTranscieveLength = sizeof(lpTranscieveBuffer);

	Status = GetTranscieveFromTLV(&lpResponseBuffer[5], dwResponseBufferLength - 5, lpTranscieveBuffer, &dwTranscieveLength);

	if (Status != SCARD_S_SUCCESS)
		return Status;

	//printf(("GetTranscieveFromTLV passed\n"));
	//printTime();

	//if (lpResponseBuffer[0x11] != 0x90)
#ifdef PR
	if (lpTranscieveBuffer[0x01] != 0x90)
#else
	if (lpTranscieveBuffer[0x00] != 0x90)
#endif
		return SCARD_E_UNEXPECTED;

	//BYTE Data[256] = { 0 };

	//memcpy(Data, &pResponseBuffer[0x10], pResponseBuffer[0x0F]);

	printf("Data = ");
	/*
	int iPos = 0x12;
	for (int i = 0; i < lpResponseBuffer[0x0F] - 2; i++)
		printf("%02X", lpResponseBuffer[iPos++]);
	*/
#ifdef PR
	int iPos = 0x02;
#else
	int iPos = 0x01;
#endif
	for (int i = 0; i < 16; i++)
		printf("%02X", lpTranscieveBuffer[iPos++]);
	printf("\n");

	return Status;
}
#else // NON-PC/SC
	#ifdef MANUAL // NON-PC/SC MANUAL
	LONG ReadMFP()
	{
		BYTE lpRequest[6] = { 0x0a, 0x00, 0x36, 0x00, 0x00, 0x01 };
		BYTE Data[255] = { 0 };
		DWORD dwRespLen = 255;

		LONG Status = CLSCRF_DirectIO_Card(
			m_pReader,
			lpRequest,
			sizeof(lpRequest),
			1000,
			Data,
			&dwRespLen
			);
		
		if (Status != SCARD_S_SUCCESS)
			return Status;

		if (dwRespLen < 3)
			return SCARD_E_UNEXPECTED;

		if (Data[2] != 0x90)
			return SCARD_E_UNEXPECTED;

		printf(("0x%08X"), Status);
		printTime();

		printf("Data = ");
		for (int i = 0; i < 16; i++)
			printf("%02X", Data[3 + i]);
		printf("\n");

		return Status;
	}	
	#else // NON-PC/SC UEM PROPRIETARY COMMAND
		#ifdef SAM // NON-PC/SC UEM PROPRIETARY COMMAND SAM
		LONG ReadMFP()
		{
			BYTE ucSW1, ucSW2;
			BYTE Data[48] = { 0 };

			LONG Status = CLSCRF_SAM_MifarePlus_ReadData(
				m_pReader,
				0x00,
				0x00,
				0x00,
				0x00,
				0x00,
				0x01,
				Data,
				&ucSW1,
				&ucSW2
				);

			if (Status != SCARD_S_SUCCESS)
				return Status;

			printf("Data = ");
			for (int i = 0; i < 16; i++)
				printf("%02X", Data[i]);
			printf("\n");

			return Status;
		}
		#else // NON-PC/SC UEM PROPRIETARY COMMAND NO SAM
		LONG ReadMFP()
		{
			BYTE Data[48] = { 0 };

			LONG Status = CLSCRF_MifarePlus_ReadData(
				m_pReader,
				0x00,
				0x00,
				0x00,
				0x00,
				0x01,
				Data
				);

			if (Status != SCARD_S_SUCCESS)
				return Status;

			printf("Data = ");
			for (int i = 0; i < 16; i++)
				printf("%02X", Data[i]);
			printf("\n");

			return Status;
		}
		#endif
	#endif
#endif

#ifdef PCSC
LONG CheckCardState()
{
	WCHAR           szReader[200];
	DWORD           cch = 200;
	BYTE            bAttr[32];
	DWORD           cByte = 32;
	DWORD           dwState, dwProtocol;
	LONG            lReturn;

	// Determine the status.
	// hCardHandle was set by an earlier call to SCardConnect.
	lReturn = SCardStatus(m_SCardCardHandle,
		szReader,
		&cch,
		&dwState,
		&dwProtocol,
		(LPBYTE)&bAttr,
		&cByte);

	if (SCARD_S_SUCCESS != lReturn)
	{
		printf("SCardStatus = %08X\n", lReturn);

		return SCARD_E_NO_SMARTCARD;
		//exit(1);     // or other appropriate action
	}

	// Examine retrieved status elements.
	// Look at the reader name and card state.
	//printf("%S\n", szReader);
	switch (dwState)
	{
	case SCARD_ABSENT:
		printf("Card absent.\n");
		return SCARD_E_NO_SMARTCARD;
		break;
	case SCARD_PRESENT:
		//printf("Card present.\n");
		break;
	case SCARD_SWALLOWED:
		printf("Card swallowed.\n");
		return SCARD_E_COMM_DATA_LOST;
		break;
	case SCARD_POWERED:
		//printf("Card has power.\n");
		break;
	case SCARD_NEGOTIABLE:
		printf("Card reset and waiting PTS negotiation.\n");
		return SCARD_E_COMM_DATA_LOST;
		break;
	case SCARD_SPECIFIC:
		//printf("Card has specific communication protocols set.\n");
		break;
	default:
		printf("Unknown or unexpected card state.\n");
		return SCARD_E_UNEXPECTED;
		break;
	}

	return SCARD_S_SUCCESS;
}
#else // NOT PC/SC
LONG ActivateCard()
{
	LONG Status = SCARD_S_SUCCESS;

	Status = CLSCRF_Mfrc_Rf_Off_On(m_pReader, 10);
	if (Status != 0) return Status;

	Sleep(dwActivationPause);

	BYTE ATQ[2] = { 0 }, SAK = 0;
	BYTE UID[10] = { 0 };
	DWORD dwUIDLength = sizeof(UID);

	Status = CLSCRF_Activate_Idle_A(m_pReader, ATQ, &SAK, UID, &dwUIDLength);
	if (Status != 0) return Status;

	return SCARD_S_SUCCESS;
}
#endif

void printTime()
{
	static DWORD dwLastTickCount = GetTickCount();

	printf(("\tTM: %u ms\n"), GetTickCount() - dwLastTickCount);

	dwLastTickCount = GetTickCount();
}

int _tmain(int argc, _TCHAR* argv[])
{
	LONG Status = 0;

#ifdef PCSC
	printf(("Establishing PCSC context...\t"));
	Status = EstablishContext();
#else
	printf(("Establishing USB Link...\t"));
	Status = EstablishLink();
#endif
	printf(("0x%08X"), Status);
	printTime();

	if (Status == SCARD_S_SUCCESS)
	while(true)
	{
		printf(("\nConnecting card...\t"));
		Status = ConnectCard();
		printf(("0x%08X"), Status);
		printTime();

		if (Status != SCARD_S_SUCCESS)
		{
			printf(("Waiting %ums"), dwCyclePause);
			Sleep(dwCyclePause);
			printTime();
			
			continue;
		}

		printf(("Setting registers...\t\t"));
		Status = InitReaderParameters();
		printf(("\t0x%08X"), Status);
		printTime();

		DWORD dwLastRoundTickCount;

		dwLastRoundTickCount = GetTickCount();

		while (true)
		{
#ifdef PCSC
			if (SCardIsValidContext(m_SCardContextHandle) != SCARD_S_SUCCESS)
				return 0;
#endif
			//SCardGetStatusChange()
			//SCardReconnect()
			//SCardState()
			//SCardStatus()
			
			if (m_fSwitchSAM == TRUE)
			{
				printf(("Round time (ms): %u ms\n\n"), GetTickCount() - dwLastRoundTickCount);

				printf(("Pause %ums..."), dwLargePause);

				Sleep(dwLargePause);

				printTime();
				printf(("\n\n"));

				dwLastRoundTickCount = GetTickCount();

#ifdef SAM
				m_fSwitchSAM = FALSE;
#endif
#ifdef PCSC
				if (CheckCardState() != SCARD_S_SUCCESS)
					break;
#else // NOT PC/SC
				if (ActivateCard() != SCARD_S_SUCCESS)
					break;
#endif
			}
			else
			{
//#ifdef PCSC
//				if (CheckCardState() != SCARD_S_SUCCESS)
//					break;
//#else // NOT PC/SC
//				if (ActivateCard() != SCARD_S_SUCCESS)
//					break;
//#endif

#ifdef SAM
				printf(("SAM Pause %ums...\t\t\t\t"), dwSAMPause);

				Sleep(dwSAMPause);

				printTime();

				printf(("\n"));
#endif
//#ifdef SAM
				m_fSwitchSAM = TRUE;
//#endif
			}

#ifdef SAM
			printf(("Switching SAM to %d...\t\t\t"), m_fSwitchSAM);
			Status = SwitchSAM(m_fSwitchSAM);
			printf(("0x%08X"), Status);
			printTime();

			if (Status != SCARD_S_SUCCESS)
			{
				continue;
			}

			printf(("Waiting TDA relay %ums...\t\t\t"), dwSmallPause);
			Sleep(dwSmallPause);
			printTime();

			printf(("Switching TDA ON...\t\t\t"));
			Status = SwitchOnTDA();
			printf(("0x%08X"), Status);
			printTime();

			if (Status != SCARD_S_SUCCESS)
			{
				continue;
			}

			printf(("Activating SAM...\t\t\t"));
			Status = ActivateSAM();

			if (Status != SCARD_S_SUCCESS)
			{
				printf(("0x%08X"), Status);
				printTime();
				printf(("\n"));

				continue;
			}

			printf(("\n"));
#endif

			printf(("Waiting MFP power %ums...\t\t\t"), dwLargePause);
			Sleep(dwLargePause);
			printTime();

#ifdef PCSC
			printf(("Opening Session...\t\t\t"));
			Status = OpenTransparentSession();
			printf(("0x%08X"), Status);
			printTime();

			if (Status != SCARD_S_SUCCESS)
			{
				printf(("Disconnecting card...\t\t\t"));
				Status = DisconnectCard();
				printf(("0x%08X"), Status);
				printTime();
				continue;
			}
#endif

			printf(("Entering T=CL...\t\t\t"));
			Status = EnterTCL();
			printf(("0x%08X"), Status);
			printTime();

			if (Status != SCARD_S_SUCCESS)
			{
#ifdef PCSC
				printf(("Closing Session...\t\t\t"));
				Status = CloseTransparentSession();
				printf(("0x%08X"), Status);
				printTime();
				printf(("Disconnecting card...\t\t\t"));
				Status = DisconnectCard();
				printf(("0x%08X"), Status);
				printTime();
#endif
				continue;
			}
			printf(("Authenticating MFP...\t"));
			Status = AuthenticateMFP();
			printf(("0x%08X"), Status);
			printTime();

			if (Status != SCARD_S_SUCCESS)
			{
#ifdef PCSC
				printf(("Closing Session...\t\t\t"));
				Status = CloseTransparentSession();
				printf(("0x%08X"), Status);
				printTime();
				printf(("Disconnecting card...\t\t\t"));
				Status = DisconnectCard();
				printf(("0x%08X"), Status);
				printTime();
#endif
				continue;
			}

			printf(("Reading MFP...\t\t\t\t"));
			Status = ReadMFP();
			printf(("\tReadMFP:\t\t\t0x%08X"), Status);
			printTime();

			if (Status != SCARD_S_SUCCESS)
			{
#ifdef PCSC
				printf(("Closing Session...\t\t\t"));
				Status = CloseTransparentSession();
				printf(("0x%08X"), Status);
				printTime();
				printf(("Disconnecting card...\t\t\t"));
				Status = DisconnectCard();
				printf(("0x%08X"), Status);
				printTime();
#endif
				continue;
			}
#ifdef PCSC
			printf(("Closing Session...\t\t\t"));
			Status = CloseTransparentSession();
			printf(("0x%08X"), Status);
			printTime();
			printf(("Disconnecting card...\t\t\t"));
			Status = DisconnectCard();
			printf(("0x%08X"), Status);
			printTime(); 
#endif
		}

		printf(("Disconnecting card...\t\t\t"));
		Status = DisconnectCard();
		printf(("0x%08X"), Status);
		printTime();

		printf(("Waiting %ums"), dwCyclePause);
		Sleep(dwCyclePause);
		printTime();
	}

	ReleaseContext();

	return 0;
}

