//==========================================================================//
//  MfPlus.c : implementation file                                          //
//==========================================================================//

#include "Config.h"
#if (FEATURES & B_MIFARE_PLUS_LPC)

#include <String.h>
#include <MfErrNo.h>
#include <Crypto\MfPlus.h>
#include <Crypto\AES_128.h>
#include <Crypto\CRC.h>
#include <Lib500\RICReg.h>
#include <Lib500\MfRc_ARM.h>
#include <Lib500\Inface.h>
#include <Host\HostIO.h>
#include <Modules\Flash.h>

extern CRYPTO_SUITE m_CryptoSuite;
DWORD TransactID;
WORD W_Ctr;
WORD R_Ctr;
WORD RR_MAC_Fill;
WORD RR_Prev_MACed;
BYTE ReadResponse_MAC_Work[16];
BYTE ReadResponse_MAC_Rest[16];
BYTE MfP_K_ENC[16];
BYTE MfP_K_MAC[16];
//BYTE MfP_CRYPTO1[16];

//extern DWORD byte4;
extern DWORD TCL_Timeout;
extern WORD byte2;
extern BYTE *MSndBuffer; // pointer to the transmit buffer
extern BYTE *MRcvBuffer; // pointer to the receive buffer
extern BYTE CryptoPlain[];
extern BYTE CryptoCoded[];
extern BYTE Answer[];
extern BYTE TCL_PCB;
extern BYTE TCL_CID;
extern BYTE TCL_PPS1;
/*
void xor_128( BYTE* a, BYTE* b, BYTE* out)
{
    int i;
    for( i = 0; i < 16; i++ )
        out[i] = a[i] ^ b[i];
}

void leftshift_onebit( BYTE* input, BYTE* output )
{
    int i;
    BYTE overflow = 0;

    for( i = 15; i >= 0; i-- ) 
    {
        output[i] = input[i] << 1;
        output[i] |= overflow;
        overflow = (input[i] & 0x80)?1:0;
    }
}

void generate_subkey( BYTE* key, BYTE* K1, BYTE* K2 )
{
    int i;
	BYTE const_Rb[16] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0x87 };
	BYTE L[16];
    BYTE Z[16];
    BYTE tmp[16];

    for( i = 0; i < 16; i++ ) 
        Z[i] = 0;

    aes_128_encrypt16( Z, key, L );

    if( (L[0] & 0x80) == 0 )    // If MSB(L) = 0, then K1 = L << 1
    { 
        leftshift_onebit( L, K1 );
    } 
    else                        // Else K1 = ( L << 1 ) (+) Rb
    { 
        leftshift_onebit( L, tmp );
        xor_128( tmp, const_Rb, K1 );
    }

    if ( (K1[0] & 0x80) == 0 ) 
    {
        leftshift_onebit( K1, K2 );
    } 
    else 
    {
        leftshift_onebit( K1, tmp );
        xor_128( tmp, const_Rb, K2 );
    }
}

void AES_128_CBC_En_MAC( BYTE* pInp, int ByteSize, BYTE* pKey, BYTE* pOut )
{
// Parameters: BYTE* pInp, int ByteSize, BYTE* pKey, BYTE* pOut
    int i, j, BlockQty;
    BYTE* pPlain;
    BYTE* pSubKey;
    BYTE wrk[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    BYTE pK1[16];
    BYTE pK2[16];

    if( ByteSize & 0x0f )
    {
        pSubKey = pK2;
        pInp[ByteSize] = 0x80;
        for( i = ByteSize + 1; ( i & 0x0f ) > 0; i++ )
            pInp[i] = 0;
    }
    else
        pSubKey = pK1;
    BlockQty = ( ByteSize + 15 ) >> 4;
    generate_subkey( pKey, pK1, pK2 );
    pPlain = pInp;
    for( i = 0; i < BlockQty; i++ )
    {
        for( j = 0; j < 16; j++ )
        {
            wrk[j] ^= pPlain[j]; 
            if( i == BlockQty - 1 )
//                wrk[j] ^= pK2[j]; 
                wrk[j] ^= pSubKey[j]; 
        }
        aes_128_encrypt16( wrk, pKey, pOut );
        memcpy( wrk, pOut, 16 );
        pPlain += 16;
    }
}
*/
void MfPlusAES128InitVector( BYTE Type,  BYTE* InV )
{
    int i;
    if( Type == 0 )         // First Authenticate
    {
        for( i = 0; i < 16; i++ )
            InV[i] = 0;
    }
    else if( Type == 1 )    // for Command
    {
        memcpy( &InV[0], &TransactID, 4 );
        memcpy( &InV[4], &R_Ctr, 2 );
        memcpy( &InV[6], &W_Ctr, 2 );
        memcpy( &InV[8], &R_Ctr, 2 );
        memcpy( &InV[10], &W_Ctr, 2 );
        memcpy( &InV[12], &R_Ctr, 2 );
        memcpy( &InV[14], &W_Ctr, 2 );
    }
    else // if( Type == 2 ) // for Response
    {
        memcpy( &InV[0], &R_Ctr, 2 );
        memcpy( &InV[2], &W_Ctr, 2 );
        memcpy( &InV[4], &R_Ctr, 2 );
        memcpy( &InV[6], &W_Ctr, 2 );
        memcpy( &InV[8], &R_Ctr, 2 );
        memcpy( &InV[10], &W_Ctr, 2 );
        memcpy( &InV[12], &TransactID, 4 );
    }
}

void MfPlusMacOnReadCommand( BYTE* Cmd, WORD Size, BYTE* MAC8 )
{
/*
1 CC    0       1 PCB   4
2 R_Ctr 1       1 CID   5
4 TI    3       1 CC    6       1  PCB  7
2 BNr   7       2 BNr   7       1  CID  8
1 Len   9       1 Len   9       1  SC   9
6 Pad  10       8 MAC  10       X  Dat 10
*/
    int i, j;
	BYTE MAC16[16];
#if MFP_DEBUG
    for( i = 0; i < 16; i++ )
        Cmd[Size+4+i] = i | 0xf0;
#endif
    Cmd[0] = Cmd[6];
    memcpy( &Cmd[1], &R_Ctr, 2 );
    memcpy( &Cmd[3], &TransactID, 4 );

    AES_128_CBC_En_MAC( Cmd, Size + 4, MfP_K_MAC, MAC16 );

    for( i = 0, j = 1; i < 8; i++, j+=2 )
        MAC8[i] = MAC16[j];
}

void MfPlusMacOnReadResponse( BYTE ThisMACed, BYTE* ReadAddr, BYTE ReadLen, BYTE* InpResp, WORD InpSize, BYTE* OutMAC8 )
{
// Parameters: BYTE MACed,
//                  InpResp
//1 SC    0
//2 R_Ctr 1
//4 TI    3       1  PCB  7
//2 BNr   7       1  CID  8
//1 Len   9       1  SC   9
//X Dat  10       X  Dat 10

// Parameters: BYTE* pInp, int ByteSize, BYTE* pKey, BYTE* pOut
    int i, j, CurInp, EndInp;
    BYTE* pSubKey;
    BYTE pK1[16];
    BYTE pK2[16];

    if( RR_Prev_MACed )
    {
        for( i = 0; i < 16; i++ )
            ReadResponse_MAC_Work[i] = 0;
        ReadResponse_MAC_Rest[0] = InpResp[9];
        memcpy( &ReadResponse_MAC_Rest[1], &R_Ctr, 2 );
        memcpy( &ReadResponse_MAC_Rest[3], &TransactID, 4 );
        RR_MAC_Fill = 7;
    }
    InpResp[7] = ReadAddr[0];
    InpResp[8] = ReadAddr[1];
    InpResp[9] = ReadLen;
    CurInp = 7;
    EndInp = ( InpSize & 0xfff0 ) + 10;
    do
    {
        ReadResponse_MAC_Rest[ RR_MAC_Fill++ ] = InpResp[ CurInp++ ];
        if( RR_MAC_Fill == 16 )
        {
            RR_MAC_Fill = 0;
            if( ( CurInp < EndInp ) || ( ThisMACed == 0 ) )
            {
                for( i = 0; i < 16; i++ )
                    ReadResponse_MAC_Work[i] ^= ReadResponse_MAC_Rest[i];
                aes_128_encrypt16( ReadResponse_MAC_Work, MfP_K_MAC, ReadResponse_MAC_Rest );
                memcpy( ReadResponse_MAC_Work, ReadResponse_MAC_Rest, 16 );
            }
        }
    } while( CurInp < EndInp );

    if( ThisMACed == 0 )
    {
        RR_Prev_MACed = 0;
        return;
    }

    generate_subkey( MfP_K_MAC, pK1, pK2 );
    if( RR_MAC_Fill )
    {
        pSubKey = pK2;
        ReadResponse_MAC_Rest[RR_MAC_Fill] = 0x80;
        for( i = RR_MAC_Fill + 1; ( i & 0x0f ) > 0; i++ )
            ReadResponse_MAC_Rest[i] = 0;
    }
    else
        pSubKey = pK1;

    for( i = 0; i < 16; i++ )
    {
        ReadResponse_MAC_Work[i] ^= ReadResponse_MAC_Rest[i];
        ReadResponse_MAC_Work[i] ^= pSubKey[i];
    }
    aes_128_encrypt16( ReadResponse_MAC_Work, MfP_K_MAC, ReadResponse_MAC_Rest );

    for( i = 0, j = 1; i < 8; i++, j+=2 )
        OutMAC8[i] = ReadResponse_MAC_Rest[j];

    RR_Prev_MACed = 1;
}

void MfPlusMacOnWriteCommand( BYTE* Cmd, WORD Size, BYTE* MAC8 )
{
/*
1 CC    0      1 PCB  4
2 W_Ctr 1      1 CID  5
4 TI    3      1 CC   6
2 BNr   7      2 BNr  7
X Dat   9      X Dat  9
  Pad  10      8 MAC 10 
*/
    int i, j;
	BYTE MAC16[16];
#if MFP_DEBUG
    for( i = 0; i < 16; i++ )
        Cmd[Size+4+i] = i | 0xf0;
#endif
    Cmd[0] = Cmd[6];
    memcpy( &Cmd[1], &W_Ctr, 2 );
    memcpy( &Cmd[3], &TransactID, 4 );

    AES_128_CBC_En_MAC( Cmd, Size + 4, MfP_K_MAC, MAC16 );

    for( i = 0, j = 1; i < 8; i++, j+=2 )
        MAC8[i] = MAC16[j];
}

void MfPlusMacOnWriteResponse( BYTE* Dat, BYTE* MAC8 )
{
/*
0 PCB   1
1 CID   1
2 SC    1
3 W_Ctr 2
5 TI    4
9 Pad   9
*/
    int i, j;
	BYTE MAC16[16];
#if MFP_DEBUG
    for( i = 0; i < 16; i++ )
        Dat[3+i] = i | 0xf0;
#endif
    memcpy( &Dat[3], &W_Ctr, 2 );
    memcpy( &Dat[5], &TransactID, 4 );

    AES_128_CBC_En_MAC( &Dat[2], 7, MfP_K_MAC, MAC16 );

    for( i = 0, j = 1; i < 8; i++, j+=2 )
        MAC8[i] = MAC16[j];
}

char MfPlusStatusCode( BYTE Error )
{
    if( Error == 0x90 )
        return 0;
    else
        return ( Error | 0xa0 );
}

void MfPlusAddressFromType( BYTE Type, BYTE Sector, BYTE Block, BYTE* MfPAddr )
{
    BYTE BlockAddr;

    MfPAddr[1] = Type & 0xf0;
    if( Type == 0x00 )
    {
        BlockAddr = 4 * Sector;
        if( Sector >= 0x20 )
            BlockAddr += 12 * ( Sector - 0x20 );
        MfPAddr[0] = BlockAddr + Block;
    }
    else if( Type < 0x50 )
    {
        MfPAddr[0] = Sector * 2 + ( Type & 1 );
    }
    else
    {
        if( (Type == 0xa2) || (Type == 0xa3) )
            MfPAddr[0] = 0x80 | ( Type & 1 );
        else
            MfPAddr[0] = Type & 0x0f;
    }
}

//===================================== PERSONALIZATION ================================================

//char MfPlusWritePerso( BYTE KeyType, BYTE Sector, BYTE Block, WORD KeyNumber16 )
char MfPlusWritePerso( BYTE* Param )
{
 //0 0A 00
 //2 A8 
 //3 00 90 
 //5 A5 A4 A3 A2 A1 A0 18 77 8E 69 F0 F0 F0 F0 F0 F0
 // 00 2F 00 00
    BYTE* pBuf;
    WORD KeyNumber16;
    char Status = 0;

    pBuf = CryptoPlain;
    pBuf[0] = TCL_PCB;
    pBuf[1] = TCL_CID;
    pBuf[2] = 0xa8;
    MfPlusAddressFromType( Param[0], Param[1], Param[2], &pBuf[3] );
    memcpy( &KeyNumber16, &Param[3], 2 );
    if( KeyNumber16 == 0xffff )
        memcpy( &pBuf[5], &Param[5], 16 );
    else
        Status = LPC_FlashRead16( KeyNumber16, &pBuf[5] );

    if( Status )
        return Status;

    Status = Mf500PiccExchangeBlock( pBuf,
                                     21,
                                     Answer,
                                     &byte2,
                                     TCL_Timeout );
    TCL_PCB ^= 0x01;
    if( Status )
        return Status;
    Status = MfPlusStatusCode( Answer[2] );
    if( Status )
        return Status;

    return Status;
}

char MfPlusCommitPerso( void )
{
    BYTE* pBuf;
    char Status = 0;

    pBuf = CryptoPlain;
    pBuf[0] = TCL_PCB;
    pBuf[1] = TCL_CID;
    pBuf[2] = 0xaa;
    Status = Mf500PiccExchangeBlock( pBuf,
                                     3,
                                     Answer,
                                     &byte2,
                                     TCL_Timeout );
    TCL_PCB ^= 0x01;
    if( Status )
        return Status;
    Status = MfPlusStatusCode( Answer[2] );
    if( Status )
        return Status;

    return Status;
}

//==================================== AUTHENTICATION ================================================

char MfPlusAuthenticate( BYTE* Param )
{// Param: AuthType[1]; KeyType[1]; Sector[1]; KeyNumber[2]; LenCap[1]; [PCDCap2[0..6]]
 //  A0       01          93          00          0000         00
#if MFP_DEBUG
    const BYTE cRndA[16] = { 0xCF, 0xC1, 0x0C, 0x4F, 0x63, 0x05, 0x3E, 0x15, 0x08, 0x25, 0xC8, 0xC7, 0xE8, 0x05, 0xF3, 0x02 };
    const BYTE cRndB[16] = { 0xB0, 0xE4, 0x0C, 0x79, 0x7C, 0x50, 0xE1, 0xE4, 0x8E, 0x88, 0xBE, 0xD0, 0x4C, 0x9F, 0x95, 0x79 };
    const BYTE cKeyX[16] = { 0x8D, 0xDF, 0xF1, 0x51, 0xA6, 0xEF, 0x6A, 0x7F, 0xE6, 0xD0, 0x33, 0x3A, 0x42, 0xBE, 0x21, 0xEE };
#endif
    int i, n, CmdOffset;
    BYTE* pBuf;
	BYTE RndA[16];
    BYTE RndB[16];
    char Status = 0;
    BYTE AuthType;
// 0 - Reset
// 1 - First
// 2 - SL2
// else - Following

#if MFP_DEBUG
    memcpy( RndA, cRndA, 16 );
    memcpy( RndB, cRndB, 16 );
    memcpy( (BYTE*)&m_CryptoSuite.k.AesKey[1], cKeyX, 16 );
#else

    AuthType = Param[0];
    RR_Prev_MACed = 1;
// 1      0x70
// 2      KeyNumber
// 1      LenCap ( default == 0 )
// 0 .. 6 PCDcap2

    pBuf = CryptoPlain;
    pBuf[0] = TCL_PCB;
    pBuf[1] = TCL_CID;

    if( TCL_CID > MAX_VALID_CID )
        CmdOffset = 2;
    else
        CmdOffset = 0;

    if( AuthType == 0 )
    {
        pBuf[2] = 0x78;
        byte2 = 3;
    }
    else 
    {
        MfPlusAddressFromType( Param[1], Param[2], 0, &pBuf[3] );
        if( AuthType == 1 )
        {
            pBuf[2] = 0x70;
            byte2 = Param[5];
            pBuf[5] = byte2;
            if( byte2 > 0 )
                memcpy( &pBuf[6], &Param[6], byte2 );
            byte2 += 6;
        }
        else
        {
            pBuf[2] = 0x76;
            byte2 = 5;
        }
    }
    Status = Mf500PiccExchangeBlock( &pBuf[CmdOffset],          // 111
                                     byte2-CmdOffset,
                                     &Answer[CmdOffset],
                                     &byte2,
                                     TCL_Timeout );
    TCL_PCB ^= 0x01;
    if( Status )
        return Status;
    Status = MfPlusStatusCode( Answer[2] );
    if( Status )
        return Status;

    if( AuthType == 0 )
        return Status;

/*
    if( AuthType == 1 )
    {
        MfPlusAES128InitVector( 0, m_CryptoSuite.k.b );
    }
    else
    {
        MfPlusAES128InitVector( 2, m_CryptoSuite.k.b );
    }
*/
    if( ( AuthType == 1 ) || ( TCL_CID > MAX_VALID_CID ) )
    {
        TransactID = 0;
        W_Ctr = 0;
        R_Ctr = 0;
    }
    MfPlusAES128InitVector( 2, m_CryptoSuite.k.b );

    memcpy( &byte2, &Param[3], 2 );
    Status = LPC_FlashRead16( byte2, (BYTE*)&m_CryptoSuite.k.AesKey[1] );
    if( Status )
        return Status;

    AES_128_CBC_De( &Answer[3], 1, 
                    (BYTE*)&m_CryptoSuite.k.AesKey[1],    // Key
                    (BYTE*)&m_CryptoSuite.k.AesKey[0],    // InV
                    &CryptoCoded[15] );

    memcpy( RndB, &CryptoCoded[15], 16 );
    CryptoCoded[31] = CryptoCoded[15];
    for( i = 0; i < 16; i++ )
        CryptoCoded[i] = i + 0xa1;
    memcpy( RndA, &CryptoCoded[0], 16 );
// 1      0x72
// 32     E(Kx, RndA || RndB' )

    pBuf[0] = TCL_PCB;
    pBuf[1] = TCL_CID;
    pBuf[2] = 0x72;

//    if( AuthType > 1 )
        MfPlusAES128InitVector( 1, m_CryptoSuite.k.b );
    AES_128_CBC_En( CryptoCoded, 2,
                    (BYTE*)&m_CryptoSuite.k.AesKey[1],    // Key
                    (BYTE*)&m_CryptoSuite.k.AesKey[0],    // InV
                    &pBuf[3] );

    Status = Mf500PiccExchangeBlock( &pBuf[CmdOffset],          // 222
                                     35-CmdOffset,
                                     &Answer[CmdOffset],
                                     &byte2,
                                     TCL_Timeout );
    TCL_PCB ^= 0x01;
    if( Status )
        return Status;
    Status = MfPlusStatusCode( Answer[2] );
    if( Status )
        return Status;

// RndA' = ( RndA << 8 )
    if( AuthType == 1 )
        n = 2;
    else
    {
        n = 1;
        MfPlusAES128InitVector( 2, m_CryptoSuite.k.b );
    }
    AES_128_CBC_De( &Answer[3], n, 
                    (BYTE*)&m_CryptoSuite.k.AesKey[1],    // Key
                    (BYTE*)&m_CryptoSuite.k.AesKey[0],    // InV
                    &CryptoCoded[1] );
//    memcpy( &Answer[3], &CryptoCoded[1], 32 );      // All Answer
// Verify RndA'
    if( AuthType == 1 )
    {
        memcpy( &TransactID, &CryptoCoded[1], 4 );
        CryptoCoded[4] = CryptoCoded[20];
        n = 4;
    }
    else
    {
        CryptoCoded[0] = CryptoCoded[16];
        n = 0;
    }
    for( i = 0; i < 16; i++ )
        if( CryptoCoded[i+n] != i + 0xa1 )
            return MI_MFP_AUTHENTICATION_ERR;
#endif

//-------------------------------------------------------

    memcpy( &CryptoCoded[0], &RndA[11], 5 );
    memcpy( &CryptoCoded[5], &RndB[11], 5 );
    for( i = 4; i < 9; i++ )
        CryptoCoded[i+6] = RndA[i] ^ RndB[i];
    if( TCL_CID > MAX_VALID_CID )
        CryptoCoded[15] = 0x33;
    else
        CryptoCoded[15] = 0x11;
    aes_128_encrypt16( CryptoCoded, (BYTE*)&m_CryptoSuite.k.AesKey[1], MfP_K_ENC );

    memcpy( &CryptoCoded[0], &RndA[7], 5 );
    memcpy( &CryptoCoded[5], &RndB[7], 5 );
    for( i = 0; i < 5; i++ )
        CryptoCoded[i+10] = RndA[i] ^ RndB[i];
    CryptoCoded[15] = 0x22;
    aes_128_encrypt16( CryptoCoded, (BYTE*)&m_CryptoSuite.k.AesKey[1], MfP_K_MAC );

    m_CryptoSuite.f = 1;
    
#if MFP_DEBUG
    memcpy( Answer, MfP_K_ENC, 16 );
    memcpy( &Answer[16], MfP_K_MAC, 16 );
    byte2 = 32;
#endif
//BYTE MfP_CRYPTO1[16];
    return Status;
}

//===================================== READ == & == WRITE ============================================

char MfPlusReadData( BYTE* Param )
{// ReadMode(1) ValueType(1) SectorNumber(1) BlockNumber(1) BlockCount(1)
 // a4 06 b3 00 00 01
// ReadMode(1) :
//Bit0: ({MAC on response})
//Bit1: (!{Encrypted})
//Bit2: (!{MAC on command})
//Bit3-7: 00000
#if MFP_DEBUG
                                                    // Example 9.8.6    a4 01 00 04 02 01
                                                    // 00   0E 00    0A   0C   31 12 00 01   41 9D 25 AD D3 8F C6 FC
                                                    // Plain Data    60 11 16 00 9F EE E9 FF 60 11 16 00 12 ED 12 ED
    const BYTE Kenc[16] = { 0x59, 0x41, 0x6a, 0xd5, 0x75, 0xa2, 0x76, 0x4e, 0x8d, 0x0b, 0x18, 0xfc, 0xc2, 0xe4, 0xe2, 0x4f };
    const BYTE Kmac[16] = { 0xba, 0x90, 0xf2, 0x04, 0x41, 0x72, 0x6a, 0x3d, 0xdb, 0x57, 0x7f, 0x7d, 0x1e, 0xc9, 0xc1, 0x18 };
    const BYTE cAns[30] = { 0x0a, 0x0c, 0x90, 0x8A, 0xDC, 0xC4, 0xC6, 0x45, 0x1A, 0x23, 0xBB, 0x6C, 0x1D, 0xED, 0xE0, 
                            0x2D, 0xF3, 0xAB, 0x72, 0x07, 0x48, 0xAF, 0x73, 0x14, 0x48, 0x96, 0x70, 0x00, 0x00, 0x00 };
#endif

    BYTE* pBuf;
    WORD Size;
	BYTE InV[16];
    BYTE ReadMode;
    char Status = 0;

#if MFP_DEBUG
    byte2 = 0;
    TCL_PCB = 0x0a;
    TCL_CID = 0x0c;
    TransactID = 0x24ccbbaa;
    W_Ctr = 2;
    R_Ctr = 1;
    memcpy( MfP_K_ENC, Kenc, 16 );
    memcpy( MfP_K_MAC, Kmac, 16 );
#endif

    ReadMode = 0x30 | ( Param[0] & 7 );
    pBuf = CryptoPlain;
    pBuf[6] = ReadMode;
    MfPlusAddressFromType( Param[1], Param[2], Param[3], &pBuf[7] );
    pBuf[9] = Param[4];
    Size = 6;
// 1. Calculate the MAC
    if( ( ReadMode & 4 ) == 0 )
    {
        MfPlusMacOnReadCommand( pBuf, Size, &pBuf[10] );
        Size += 8;
    }
/*
1 CC    0       1 PCB   4
2 R_Ctr 1       1 CID   5
4 TI    3       1 CC    6
2 BNr   7       2 BNr   7
1 Len   9       1 Len   9
6 Pad  10       8 MAC  10 
*/
// 2. Send the command to the PICC
    pBuf[4] = TCL_PCB;
    pBuf[5] = TCL_CID;
    pBuf[6] = ReadMode;
#if MFP_DEBUG
    memcpy( &Answer[byte2], &Size, 2 );
    byte2 += 2;
    memcpy( &Answer[byte2], &pBuf[4], Size );
    byte2 += Size;
//    return 0;
#else
    Status = Mf500PiccExchangeBlock( &pBuf[4],          // 111
                                     Size,
                                     &CryptoCoded[7],
                                     &byte2,
                                     TCL_Timeout );
    TCL_PCB ^= 0x01;
    if( Status )
        return Status;
    Status = MfPlusStatusCode( CryptoCoded[9] );
    if( Status )
        return Status;
#endif
/*
1 CC    0 
2 R_Ctr 1 
4 TI    3     1  PCB  7
2 BNr   7     1  CID  8
1 Len   9     1  SC   9
6 Pad  10     X  Dat 10
*/
#if MFP_DEBUG
    byte2 = 27;
    memcpy( &CryptoCoded[7], cAns, byte2 );
#endif

// 3. Increment the R_Ctr
    R_Ctr++;
// 4. Receive the response from the PICC
// 5. Verify the response and check the MAC
    memcpy( Answer, &CryptoCoded[byte2-1], 8 );
    MfPlusMacOnReadResponse( ReadMode & 1, &pBuf[7], pBuf[9], CryptoCoded, byte2, &CryptoCoded[byte2-1] );
    if( ReadMode & 1 )
    {
// MfPlusMacOnReadResponse( BYTE ThisMACed, BYTE* ReadAddr, BYTE ReadLen, BYTE* InpResp, WORD InpSize, BYTE* OutMAC8 );
//        MfPlusMacOnReadResponse( CryptoCoded, byte2-8, &CryptoCoded[byte2-1], &pBuf[7], pBuf[9] );
/*
    const BYTE cAns[30] = { 
0x0a, 0x0c, 
0x90, 
0x8A, 0xDC, 0xC4, 0xC6, 0x45, 0x1A, 0x23, 0xBB, 0x6C, 0x1D, 0xED, 0xE0, 0x2D, 0xF3, 0xAB, 0x72, 
0x07, 0x48, 0xAF, 0x73, 0x14, 0x48, 0x96, 0x70
*/
        if( memcmp( Answer, &CryptoCoded[byte2-1], 8 ) )
            return MI_MFP_INVALID_CARD_MAC;

        byte2 -= 8;
    }
// 6. Decrypt the data
    byte2 -= 3;
    if( ( ReadMode & 2 ) == 0 ) // Decryption
    {
        MfPlusAES128InitVector( 2, InV );
        AES_128_CBC_De( &CryptoCoded[10], byte2 >> 4,
                        MfP_K_ENC,      // Key
                        InV,            // InV
                        &Answer[2] );
        memcpy( Answer, &byte2, 2 );
        byte2 += 2;
    }
    else                        // Plain data
    {
        memcpy( Answer, &byte2, 2 );
        memcpy( &Answer[2], &CryptoCoded[10], byte2 );
        byte2 += 2;
    }
    return Status;
}

char MfPlusWriteData( BYTE* Param )
{// IN: WriteMode(1) ValueType(1) SectorNumber(1) BlockNumber(1) BlockCount(1) Data(16/32/48)
 // A5 02 4A 00 00 01 A5A4A3A2A1A018778E69F0F0F0F0F0F0
//Bit0: ({MAC on response})
//Bit1: (!{Encrypted})
//Bit2-7: 000000
#if MFP_DEBUG
                                                    // Example 9.8.3    a5 01 00 02 02 01 02BB2A18BA7C4B54FC9F8CAE377F5C1A
                                                    // 
    const BYTE Kenc[16] = { 0xD1, 0x3C, 0xDB, 0x09, 0xCC, 0xD2, 0xE4, 0x2C, 0xE0, 0x5E, 0xD7, 0xB8, 0xB6, 0xEB, 0xC6, 0x87 };
    const BYTE Kmac[16] = { 0x72, 0xA8, 0x2A, 0xEF, 0x1A, 0x1E, 0xA4, 0xB8, 0x69, 0x5C, 0x26, 0x08, 0x22, 0xA2, 0xA8, 0xE5 };
    const BYTE cDat[16] = { 0x02, 0xBB, 0x2A, 0x18, 0xBA, 0x7C, 0x4B, 0x54, 0xFC, 0x9F, 0x8C, 0xAE, 0x37, 0x7F, 0x5C, 0x1A };
#endif

    BYTE* pBuf;
//    int i, j;
    WORD Size;
	BYTE InV[16];
    BYTE WriteMode;
    BYTE BlockCount;
    char Status = 0;

#if MFP_DEBUG
    byte2 = 0;
    TCL_PCB = 0x0a;
    TCL_CID = 0x0c;
    TransactID = 0x24ccbbaa;
    W_Ctr = 1;
    R_Ctr = 0;
    memcpy( MfP_K_ENC, Kenc, 16 );
    memcpy( MfP_K_MAC, Kmac, 16 );
#endif

    WriteMode = 0xa0 | ( Param[0] & 3 );
    pBuf = CryptoPlain;
    pBuf[6] = WriteMode;
    MfPlusAddressFromType( Param[1], Param[2], Param[3], &pBuf[7] );
    BlockCount = Param[4];  // 1, 2 or 3
    Size = 5 + BlockCount * 16;
// 1. Encrypt the data
    if( ( WriteMode & 2 ) == 0 )
    {
        MfPlusAES128InitVector( 1, InV );
        AES_128_CBC_En( &Param[5], BlockCount,
                        MfP_K_ENC,      // Key
                        InV,            // InV
                        &pBuf[9] );
    }
    else
        memcpy( &pBuf[9], &Param[5], BlockCount * 16 );
#if MFP_DEBUG
    byte2 = BlockCount * 16;// 10 3E 4B 9E 73 38 4F 8F B7 9B 02 F0 63 1B 4B 45
    memcpy( Answer, &pBuf[9], byte2 );
//    return 0;
#endif  // MFP_DEBUG
// 2. Calculate the MAC
    MfPlusMacOnWriteCommand( pBuf, Size, &pBuf[Size+4] );
    Size += 8;
/*
1 CC    0      1 PCB  4
2 W_Ctr 1      1 CID  5
4 TI    3      1 CC   6
2 BNr   7      2 BNr  7
X Dat   9      X Dat  9
  Pad  10      8 MAC 10 
*/
// 3. Send the command to the PICC
    pBuf[4] = TCL_PCB;
    pBuf[5] = TCL_CID;
    pBuf[6] = WriteMode;
#if MFP_DEBUG
    memcpy( &Answer[byte2], &Size, 2 );
    byte2 += 2;
    memcpy( &Answer[byte2], &pBuf[4], Size );
    byte2 += Size;
//    return 0;
//                                                                      10 3E 4B 9E 73 38 4F 8F B7 9B 02 F0 63 1B 4B 45
// 10 3E 4B 9E 73 38 4F 8F B7 9B 02 F0 63 1B 4B 45 1D 00 0A 0C A1 0A 00 10 3E 4B 9E 73 38 4F 8F B7 9B 02 F0 63 1B 4B 45 E5 88 17 68 83 D3 90 8F
#else
    Status = Mf500PiccExchangeBlock( &pBuf[4],          // 111
                                     Size,
                                     Answer,
                                     &byte2,
                                     TCL_Timeout );
    TCL_PCB ^= 0x01;
    if( Status )
        return Status;
    Status = MfPlusStatusCode( Answer[2] );
    if( Status )
        return Status;
#endif
// 4. Increment the W_Ctr
    W_Ctr++;
// 5. Receive the response from the PICC
// 6. Verify the response and check the MAC
    if( WriteMode & 1 )
    {
#if MFP_DEBUG
        Answer[0] = 0x0A;
        Answer[1] = 0x0C;
        Answer[2] = 0x90;
        Answer[3] = 0xc2;
        Answer[4] = 0xfb;
        Answer[5] = 0x0e;
        Answer[6] = 0x11;
        Answer[7] = 0x94;
        Answer[8] = 0x70;
        Answer[9] = 0xdf;
        Answer[10] = 0x1c;
#endif
        memcpy( CryptoCoded, &Answer[3], 8 );
        MfPlusMacOnWriteResponse( Answer, &Answer[18] );
#if MFP_DEBUG
    memcpy( &Answer[26], CryptoCoded, 8 );
    byte2 = 34;
// 0A 0C   90 02 00 AA BB CC 24 80 00 00 00 00 00 00 00 00   C2 FB 0E 11 94 70 DF 1C   C2 FB 0E 11 94 70 DF 1C
//    return 0;
#else
        if( memcmp( CryptoCoded, &Answer[18], 8 ) )
            return MI_MFP_INVALID_CARD_MAC;
#endif  // MFP_DEBUG
    }

    return Status;
}
//===================================== VALUE ================================================
/*
            1   1   1   2    16    8
            4   5   6   7     9   25
B0 Inc     PCB TCL B0 Src2 EncDat MAC
B1 Inc Mr
B2 Dec
B3 Dec Mr

            1   1   1   2   8
            4   5   6   7   9
B4 Tra     PCB TCL B0 Tgt2 MAC
B5 Tra Mr

            1   1   1   2   2     16    8
            4   5   6   7   9     11   27
B6 InT     PCB TCL B0 Src2 Tgt2 EncDat MAC
B7 InT Mr
B8 DeT
B9 DeT Mr

            1   1   1   2   8
            4   5   6   7   9
C2 Res     PCB TCL B0 Src2 MAC
C3 Res Mr
*/
char MfPlusValueOperation( BYTE* Param )
 //   0       1       2      3       4        8       9
{// Cmd[1] Crypto[1] SIn[1] BIn[1] Value[4] SOut[1] BOut[1]
    BYTE* pBuf;
    int i, iDataTo;
    WORD Size;
	BYTE InV[16];
    BYTE Oper;
    char Status = 0;

    iDataTo = 0;
    Oper = Param[0];
    pBuf = CryptoPlain;
    MfPlusAddressFromType( 0, Param[2], Param[3], &pBuf[7] );
    if( Oper < 0xB4 )
    {
        iDataTo = 9;
        Size = 21;
    }
    else if( Oper < 0xB6 )
    {
        Size = 5;
    }
    else if( Oper < 0xC2 )
    {
        iDataTo = 11;
        MfPlusAddressFromType( 0, Param[8], Param[9], &pBuf[9] );
        Size = 23;
    }
    else
    {
        Size = 5;
    }
    Oper |= Param[1] & 1;
    pBuf[6] = Oper;
    if( iDataTo )
    {
        Param[8] = 0x80;
        for( i = 9; i < 20; i++ )
            Param[i] = 0;
// 1. Encrypt the data
            MfPlusAES128InitVector( 1, InV );
            AES_128_CBC_En( &Param[4], 1,
                            MfP_K_ENC,      // Key
                            InV,            // InV
                            &pBuf[iDataTo] );
    }
// 2. Calculate the MAC
    MfPlusMacOnWriteCommand( pBuf, Size, &pBuf[Size+4] );
    Size += 8;
/*
1 CC    0      1 PCB  4
2 W_Ctr 1      1 CID  5
4 TI    3      1 CC   6
2 BNr   7      2 BNr  7
X Dat   9      X Dat  9
  Pad  10      8 MAC 10 
*/
// 3. Send the command to the PICC
    pBuf[4] = TCL_PCB;
    pBuf[5] = TCL_CID;
    pBuf[6] = Oper;
    Status = Mf500PiccExchangeBlock( &pBuf[4],          // 111
                                     Size,
                                     Answer,
                                     &byte2,
                                     TCL_Timeout );
    TCL_PCB ^= 0x01;
    if( Status )
        return Status;
    Status = MfPlusStatusCode( Answer[2] );
    if( Status )
        return Status;

// 4. Increment the W_Ctr
    W_Ctr++;
// 5. Receive the response from the PICC
// 6. Verify the response and check the MAC
    if( Oper & 1 )
    {
        memcpy( CryptoCoded, &Answer[3], 8 );
        MfPlusMacOnWriteResponse( Answer, &Answer[18] );
        if( memcmp( CryptoCoded, &Answer[18], 8 ) )
            return MI_MFP_INVALID_CARD_MAC;
    }

    return Status;
}

//===================================== VIRTUAL ================================================

char MfPlusVirtualSupport( BYTE* Param )
{
// a2 00 ffffffffffffffffffffffffffffffff
// a2 01 ffffffffffffffffffffffffffffffff 0000 0000 00 000000
//          0      1        17         19         21         22 
// IN:   Last[1] IID[16] KencNmr[2] KmacNmr[2] LenCap[1] PCDCap[LenCap]
//     2      3       19        31         32
//   CC[1]  IID[16] RndQ[12] LenCap[1] PCDCap[LenCap]
//           0        2         4
// OUT:   Info[1] PICCCAP[2] UID[4/7]

// C4 9A B4 68 0D CA BE 11 8F A1 C2 7F 03 15 5C 7E 
// 90 00 00 00 
// D0 C9 74 D4 04 B1 B8 23 3F D3 C4 E8 C8 A4 EE FD 
// A1 A2 A3 A4 A5 A6 A7 A8 A9 AA AB AC 
// 9A 68 CA 11 A1 7F 15 7E


    BYTE* pBuf;
    int i, j;
    WORD Size;
	BYTE MAC16[16];
    BYTE AesKey[16];
    char Status = 0;

    pBuf = CryptoPlain;
    pBuf[0] = TCL_PCB;
    pBuf[1] = TCL_CID;
    memcpy( &pBuf[3], &Param[1], 16 );
    
    if( Param[0] )
    {
        pBuf[2] = 0x4b;
        for( i = 0; i < 12; i++ )
            pBuf[i+19] = i + 0xa1;
        Size = Param[21];
        pBuf[31] = Size;
        pBuf[32] = 0;
        pBuf[33] = 0;
        pBuf[34] = 0;
        if( Size )
            memcpy( &pBuf[32], &Param[22], Size );
        Size += 32;
    }
    else
    {
        pBuf[2] = 0x42;
        Size = 19;
    }

    Status = Mf500PiccExchangeBlock( pBuf,          // 111
                                     Size,
                                     &CryptoCoded[1],
                                     &byte2,
                                     TCL_Timeout );
    TCL_PCB ^= 0x01;
#if MFP_DEBUG
    memcpy( Answer, &byte2, 2 );
    memcpy( &Answer[2], &CryptoCoded[1], byte2 );
    byte2 += 2;
    return 0;
#endif  // MFP_DEBUG
//           3      4        20
//          SC  Payload[16] MAC
    if( Status )
        return Status;
    Status = MfPlusStatusCode( CryptoCoded[3] );
    if( Status )
        return Status;


    if( Param[0] == 0 )
    {
        byte2 = 0;
        return Status;
    }

//xxxxxxxxxxxxxxx Del
/*
    memcpy( &byte2, &Param[17], 2 );
    Status = LPC_FlashRead16( byte2, AesKey );   // 0xa080
    if( Status )
        return Status;

    aes_128_decrypt16( &CryptoCoded[4], AesKey, MAC16 );
    memcpy( &CryptoCoded[4], MAC16, 16 );
*/
//xxxxxxxxxxxxxxx

    memcpy( &byte2, &Param[19], 2 );
    Status = LPC_FlashRead16( byte2, AesKey );  // 0xa081
    if( Status )
        return Status;

    memcpy( &Answer[48], &CryptoCoded[20], 8 );
    CryptoCoded[0] = CryptoCoded[3];
    memcpy( &CryptoCoded[1], &pBuf[32], 3 );
    memcpy( &CryptoCoded[20], &pBuf[19], 12 );
    AES_128_CBC_En_MAC( CryptoCoded, 32, AesKey, MAC16 );
    for( i = 0, j = 1; i < 8; i++, j+=2 )
        MAC16[i] = MAC16[j];
//xxxxxxxxxxxxxxx Del
/*
    byte2 = 56;
//kbh    memcpy( &Answer[8], MAC16, 8 );
    memcpy( &Answer[0], MAC16, 16 );
    memcpy( &Answer[16], CryptoCoded, 32 );
    return 0;
*/
//xxxxxxxxxxxxxxx
    if( memcmp( &Answer[48], MAC16, 8 ) )
        return MI_MFP_INVALID_CARD_MAC;

    memcpy( &byte2, &Param[17], 2 );
    Status = LPC_FlashRead16( byte2, AesKey );   // 0xa080
    if( Status )
        return Status;

    aes_128_decrypt16( &CryptoCoded[4], AesKey, &Answer[2] );
    byte2 = 16;
    memcpy( Answer, &byte2, 2 );
    byte2 += 2;
//           0        2         4
// OUT:      Info[1] PICCCAP[2]        UID[4/7]            Random Number
// 00 10 00    03     00 0B      04 08 4E A9 61 28 80    1D E7 89 1B 04 27
    return Status;
}// MfPlusVirtualSupport

char MfPlusVirtualSelect( BYTE* Param )
{
//          0        1          3         5
// IN:   Oper[1] KselNmr[2] PICCCAP[2] UID+RndP[13]
    BYTE* pBuf;
    int i, j;
    WORD Size;
	BYTE MAC16[16];
    BYTE AesKey[16];
    char Status = 0;

    pBuf = CryptoPlain;
    pBuf[0] = TCL_PCB;
    pBuf[1] = TCL_CID;
    memcpy( &pBuf[3], &Param[3], 16 );
    
    if( Param[0] )
    {
        memcpy( &byte2, &Param[1], 2 );
        Status = LPC_FlashRead16( byte2, AesKey );  // 0xa000
        if( Status )
            return Status;
    
        CryptoCoded[0] = 0x40;
        memcpy( &CryptoCoded[1], &Param[3], 2 );
        memcpy( &CryptoCoded[3], &Param[5], 13 );
        AES_128_CBC_En_MAC( CryptoCoded, 16, AesKey, MAC16 );
        for( i = 0, j = 1; i < 8; i++, j+=2 )
            pBuf[3+i] = MAC16[j];
        pBuf[2] = 0x40;
        Size = 11;
    }
    else
    {
        pBuf[2] = 0x48;
        Size = 3;
    }

    Status = Mf500PiccExchangeBlock( pBuf,
                                     Size,
                                     Answer,
                                     &byte2,
                                     TCL_Timeout );
    TCL_PCB ^= 0x01;
    if( Status )
        return Status;
    Status = MfPlusStatusCode( Answer[2] );
    return Status;
}

//===================================== PROXIMITY ================================================

char MfPlusProximityCheck( BYTE* Param )
{
    BYTE* pBuf;
    DWORD Timeout;
    int i, j, jn;
    WORD CRC, Size;
	BYTE MAC16[16];
    BYTE AesKey[16];
    char Status = 0;

//Timeout = Param[0];
    Timeout = 106;
    TCL_PCB &= ~0x08;

    pBuf = CryptoPlain;
    pBuf[0] = TCL_PCB;
    pBuf[1] = 0xF0;

    Status = Mf500PiccExchangeBlock( pBuf,
                                     2,
                                     Answer,
                                     &byte2,
                                     TCL_Timeout );
    TCL_PCB ^= 0x01;
    if( Status )
        return Status;
    Status = MfPlusStatusCode( Answer[1] );
    if( Status )
        return Status;

    WriteRC( RegChannelRedundancy, 0x03 );
	PcdSetTmo( Timeout );

    byte2 = 0;
    for( i = 0; i < 7; i++ )
    {
        jn = byte2;
        Answer[jn] = TCL_PCB;
        Answer[jn+1] = 0xF2;
        Answer[jn+2] = 1;
        Answer[jn+3] = i + 0xa1;
        byte2 += 4;
        CRC = 0x6363;
        Status = ExchangeByteStream( PCD_TRANSCEIVE,
                                     &Answer[jn],
                                     4,
                                     &Answer[byte2],
                                     &Size );
        TCL_PCB ^= 0x01;
        if( Status )
            return Status;

        byte2 += Size;
        for( j = jn; j < byte2; j++ )
            CalculateFCS( &CRC, Answer[j] );
        if( CRC != 0 )
            return MI_CRCERR;
    }
/*
03 F2 01 A1 C4 14 3F 
02 F2 01 A2 24 36 F9 
03 F2 01 A3 2E F0 44 
02 F2 01 A4 67 79 DD 
03 F2 01 A5 14 F9 8E 
02 F2 01 A6 42 66 98 
03 F2 01 A7 4B 3B 17
*/
    WriteRC(RegChannelRedundancy,0x0f);

    if( m_CryptoSuite.f )
        memcpy( AesKey, MfP_K_MAC, 16 );
    else
    {
//      memcpy( &Size, &Param[1], 2 );
        memcpy( &Size, &Param[0], 2 );
        Status = LPC_FlashRead16( Size, AesKey );  // 0xa001 Proximity Check Key
        if( Status )
            return Status;
    }

    CryptoCoded[0] = 0xFD;
    CryptoCoded[1] = TCL_PPS1;
    for( i = 0, j = 2; i < byte2; i += 7 )
    {
        CryptoCoded[j++] = Answer[i+4];
        CryptoCoded[j++] = Answer[i+3];
    }
    AES_128_CBC_En_MAC( CryptoCoded, 16, AesKey, MAC16 );
    pBuf[0] = TCL_PCB;
    pBuf[1] = 0xFD;
    for( i = 0, j = 1; i < 8; i++, j+=2 )
        pBuf[2+i] = MAC16[j];
    Status = Mf500PiccExchangeBlock( pBuf,
                                     10,
                                     Answer,
                                     &byte2,
                                     TCL_Timeout );
    TCL_PCB ^= 0x01;
    TCL_PCB |= 0x08;
    if( Status )
        return Status;
    Status = MfPlusStatusCode( Answer[1] );
    if( Status )
        return Status;

    CryptoCoded[0] = Answer[1];
    AES_128_CBC_En_MAC( CryptoCoded, 16, AesKey, MAC16 );
    for( i = 0, j = 1; i < 8; i++, j+=2 )
        Answer[byte2+i] = MAC16[j];
//    byte2 += 8;
    if( memcmp( &Answer[2], &Answer[byte2], 8 ) )
        return MI_MFP_INVALID_CARD_MAC;
/*
02 90 D2 56 A6 24 18 52 7D 40 
      D2 56 A6 24 18 52 7D 40
*/
    return Status;
}

//===================================== SL2 MULTIBLOCK ================================================

char MfPlusMultiReadData( BYTE* Param, BYTE* data )
{
//             0            1
// IN:   AbsBlockNmr[1] BlockCnt[1]
   char status = MI_OK;
   unsigned int bitsExpected;
   char tmp    = 0;
   unsigned char i, datalen;

    datalen = 16 * Param[1];

   FlushFIFO();    // empty FIFO

   PcdSetTmo(640);  // long timeout 

   WriteRC(RegChannelRedundancy,0x0F); // RxCRC, TxCRC, Parity enable
   
   // ************* Cmd Sequence ********************************** 
   ResetInfo(MInfo);   
   MSndBuffer[0] = 0x38;   // read command code
   MSndBuffer[1] = Param[0];
   MSndBuffer[2] = Param[1];
   MInfo.nBytesToSend   = 3;   
   status = PcdSingleResponseCmd(PCD_TRANSCEIVE,
                       MSndBuffer,
                       MRcvBuffer,
                       &MInfo);

   if (status != MI_OK)
   {
      if (status != MI_NOTAGERR ) // no timeout occured
      {
         if (MInfo.nBitsReceived == 8 && (ReadRC(RegDecoderControl) & 0x01))
         {
            // upper nibble should be equal to lower nibble, otherwise 
            // there is a coding error on card side.
            if ((MRcvBuffer[0] & 0x0f) == ((MRcvBuffer[0] >> 4) & 0x0f))
               bitsExpected = 8;
            else
            {
               status = MI_CODINGERR;
               bitsExpected = 0; // data is copied to output
            }
         }
         else
            bitsExpected = 4;  
         if (MInfo.nBitsReceived == bitsExpected)  // NACK
         {
             MRcvBuffer[0] &= 0x0f;  // mask out upper nibble
             if ((MRcvBuffer[0] & 0x0a) == 0)
             {
                status = MI_NOTAUTHERR;
             }
             else
             {
                status = MI_CODEERR;
             }
          }
          else
          {
             // return data - even if an error occured - for debugging reasons
             if (MInfo.nBytesReceived >= (unsigned short)datalen)
                memcpy(data,MRcvBuffer,datalen);
             else
             {
                memcpy(data,MRcvBuffer,MInfo.nBytesReceived);
                for (i = MInfo.nBytesReceived; i < datalen; i++)
                {
                   data[i] = 0x00;
                }
             }
          }
             
      }
      else
         memcpy(data,"0000000000000000",datalen); // in case of an error initialise 
                                             // data
   }
   else   // Response Processing
   {
      if (MInfo.nBytesReceived != (unsigned short)datalen)
      {
         status = MI_BYTECOUNTERR;
         // return data, even if an error occured
         if (MInfo.nBytesReceived >= (unsigned short)datalen)
            memcpy(data,MRcvBuffer,datalen);
         else
         {
            memcpy(data,MRcvBuffer,MInfo.nBytesReceived);             
            for (i = MInfo.nBytesReceived; i < datalen; i++)
            {
               data[i] = 0x00;
            }
         }   
      }
      else
      {
         memcpy(data,MRcvBuffer,datalen);
      }
   }
   return status; 
}

char MfPlusMultiWriteData( BYTE* Param )
{
//             0            1          2
// IN:   AbsBlockNmr[1] BlockCnt[1]  Data[]
   unsigned int bitsExpected;
   BYTE datalen;
   char status = MI_OK;

    datalen = 16 * Param[1];

     // ************* Cmd Sequence ********************************** 
   PcdSetTmo(1000);     // long timeout 

   WriteRC(RegChannelRedundancy,0x07); //  TxCRC, Parity enable
     
   ResetInfo(MInfo);   
   MSndBuffer[0] = 0xa8;        // Write command code
   MSndBuffer[1] = Param[0];
   MSndBuffer[2] = Param[1];
   MInfo.nBytesToSend   = 3;
   status = PcdSingleResponseCmd(PCD_TRANSCEIVE,
                         MSndBuffer,
                         MRcvBuffer,
                         &MInfo);

   if (status != MI_NOTAGERR)   // no timeout error
   {
      if (MInfo.nBitsReceived == 8 && (ReadRC(RegDecoderControl) & 0x01))
      {
         // upper nibble should be equal to lower nibble, otherwise 
         // there is a coding error on card side.
         if ((MRcvBuffer[0] & 0x0f) == ((MRcvBuffer[0] >> 4) & 0x0f))
            bitsExpected = 8;
         else
         {
            bitsExpected = 0;
         }
      }
      else
         bitsExpected = 4;  
      if (MInfo.nBitsReceived != bitsExpected)  // ACK / NACK expected
      {
         if (bitsExpected == 0)
            status = MI_CODINGERR;
         else
            status = MI_BITCOUNTERR;
      }
      else                     // 4 bit received
      {
         MRcvBuffer[0] &= 0x0f; // mask out upper nibble
         switch(MRcvBuffer[0])
         {
            case 0x00: 
               status = MI_NOTAUTHERR;
               break;
            case 0x0a:
               status = MI_OK;
               break;
            default:
               status = MI_CODEERR;
               break;
         }
      }
   }

   if ( status == MI_OK)
   {
      ResetInfo(MInfo);   
      memcpy(MSndBuffer,&Param[2],datalen);
      MInfo.nBytesToSend   = datalen;
      status = PcdSingleResponseCmd(PCD_TRANSCEIVE,
                            MSndBuffer,
                            MRcvBuffer,
                            &MInfo);
        
      if (status != MI_NOTAGERR)    // no timeout occured
      {
         if (MInfo.nBitsReceived == 8 && (ReadRC(RegDecoderControl) & 0x01))
         {
            // upper nibble should be equal to lower nibble, otherwise 
            // there is a coding error on card side.
            if ((MRcvBuffer[0] & 0x0f) == ((MRcvBuffer[0] >> 4) & 0x0f))
               bitsExpected = 8;
            else
            {
               bitsExpected = 0;
            }
         }
         else
            bitsExpected = 4;  
         if (MInfo.nBitsReceived != bitsExpected)  // ACK / NACK expected
         {
            if (bitsExpected == 0)
               status = MI_CODINGERR;
            else
               status = MI_BITCOUNTERR;
         }
         else                     // 4 bit received
         {
            MRcvBuffer[0] &= 0x0f; // mask out upper nibble
            switch(MRcvBuffer[0])
            {
               case 0x00: 
                  status = MI_WRITEERR;
                  break;
               case 0x0a:
                  status = MI_OK;
                  break;
               default:
                  status = MI_CODEERR;
                  break;
            }
         }
      }        
   }
   return status;
}

#endif //(FEATURES & B_MIFARE_PLUS_LPC)



