#include "StdAfx.h"

#include "Clscrfl.h"

std::vector<CLSCRF_CARD*> m_CardList;
CLSCRF_CARD m_DefaultCard = { 0 };
CLSCRF_CARD*  m_pCurCard = &m_DefaultCard;

BYTE m_CID_busy[16];

bool GetCID(PBYTE pCID)
{
	for (int i = 0; i < 15; i++)
	{
		if (m_CID_busy[i])
			continue;
		else
		{
			m_CID_busy[i] = 1;
			*pCID = (BYTE)i;
			return true;
		}
	}
	return false;
}

void FreeCID(BYTE CID)
{
	if (CID < 16)
		m_CID_busy[CID] = 0;
}


CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Cards_Add(IN CLSCRF_CARD*  pCard)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pCard)
		m_CardList.push_back(pCard);

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Cards_GetFirst(OUT CLSCRF_CARD**  ppCard)
{
	LONG Status = SCARD_S_SUCCESS;

	if (ppCard)
		*ppCard = m_CardList.front();

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Cards_GetLast(OUT CLSCRF_CARD**  ppCard)
{
	LONG Status = SCARD_S_SUCCESS;

	if (ppCard)
		*ppCard = m_CardList.back();

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Cards_GetCount(OUT int*  piCardsCount)
{
	LONG Status = SCARD_S_SUCCESS;

	*piCardsCount = (int)m_CardList.size();

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Cards_Get(IN int iCardIndex, OUT CLSCRF_CARD**  ppCard)
{
	LONG Status = SCARD_S_SUCCESS;

	if (ppCard)
		*ppCard = m_CardList.at(iCardIndex);

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Cards_Clear()
{
	LONG Status = SCARD_S_SUCCESS;

	//for each (CLSCRF_CARD* pCard in m_CardList)
	for (int i = 0; i < m_CardList.size(); i++)
	{
		CLSCRF_CARD* pCard = m_CardList[i];
		FreeCID(pCard->CID);
		try
		{
			delete pCard;
		}
		catch (int){}
	}

	CLSCRF_Cards_SetCurrent(&m_DefaultCard);

	m_CardList.clear();

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Cards_SetCurrent(IN CLSCRF_CARD*  pCard)
{
	LONG Status = SCARD_S_SUCCESS;

	m_pCurCard = pCard;

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Cards_GetCurrent(OUT CLSCRF_CARD**  ppCard)
{
	LONG Status = SCARD_S_SUCCESS;

	if (ppCard)
		*ppCard = m_pCurCard;

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Cards_Search(IN int UID_Len, IN PBYTE pUID, OUT CLSCRF_CARD** ppCard)
{
	LONG Status = SCARD_E_UNKNOWN_CARD;

	//for each (CLSCRF_CARD* pCard in m_CardList)
	for (int i = 0; i < m_CardList.size(); i++)
	{
		CLSCRF_CARD* pCard = m_CardList[i];
		if (UID_Len != pCard->UID_Len)
			continue;
		if (memcmp(pUID, pCard->UID, UID_Len) == 0)
		{
			*ppCard = pCard;
			return SCARD_S_SUCCESS;
		}
	}

	return Status;
}








CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Cards_CreateNew_A(IN BYTE bRfMode,
IN PBYTE pbATQ,
IN PBYTE pbSAK,
IN PBYTE pbUID,
IN DWORD UID_Len,
OUT CLSCRF_CARD** ppCard)
{
	LONG Status = SCARD_S_SUCCESS;

	CLSCRF_CARD* pSrchCard = NULL;
	CLSCRF_Cards_Search(UID_Len, pbUID, &pSrchCard);
	if (pSrchCard == NULL)
	{
		m_pCurCard = new CLSCRF_CARD;
		m_pCurCard->Timeout = 6800;     // Timeout = 64 ms
		m_pCurCard->CardType = CLSCRF_ICC_TYPE_ISO14443A;
		m_pCurCard->UID_Len = UID_Len;
		memcpy(&m_pCurCard->UID, pbUID, UID_Len);

		memcpy(&m_pCurCard->ATQ, pbATQ, 2);
		m_pCurCard->ATQ.SAK = *pbSAK;
		GetCID(&m_pCurCard->CID);
		CLSCRF_Cards_Add(m_pCurCard);
		m_pCurCard->Active = 1;
		m_pCurCard->RfMode = bRfMode;
		m_pCurCard->BlockNumberBit = 0;
		*ppCard = m_pCurCard;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Cards_CreateNew_B(IN BYTE bRfMode,
IN DWORD* pdwPUPI,
IN DWORD* pdwAppData,
IN DWORD* pdwProtInfo,
IN int    CardCount,
OUT CLSCRF_CARD** ppCard)
{
	LONG Status = SCARD_S_SUCCESS;

	//CLSCRF_CARD* pCard = NULL;

	for (int i = 0; i < CardCount; i++)
	{
		CLSCRF_CARD* pSrchCard = NULL;
		CLSCRF_Cards_Search(CLSCRF_UID_LENGTH_ISO14443_B, (unsigned char*)&pdwPUPI[i], &pSrchCard);
		if (pSrchCard == NULL)
		{
			m_pCurCard = new CLSCRF_CARD;
			m_pCurCard->Timeout = 6800;     // Timeout = 64 ms
			m_pCurCard->CardType = CLSCRF_ICC_TYPE_ISO14443B;
			m_pCurCard->UID_Len = CLSCRF_UID_LENGTH_ISO14443_B;
			memcpy(&m_pCurCard->UID, &pdwPUPI[i], CLSCRF_UID_LENGTH_ISO14443_B);

			memcpy(&m_pCurCard->ATQ.AppData, &pdwAppData[i], 4);
			memcpy(&m_pCurCard->ATQ.ProtInf, &pdwProtInfo[i], 3);
			if ((m_pCurCard->ATQ.ProtInf[1] & 0x0f) == 0x01)         // ISO 14443-4
				GetCID(&m_pCurCard->CID);
			else
				m_pCurCard->CID = 0xff;
			m_pCurCard->Timeout = 47 << ((m_pCurCard->ATQ.ProtInf[2] >> 4) & 0x0f);

			CLSCRF_Cards_Add(m_pCurCard);
			m_pCurCard->Active = 1;
			m_pCurCard->RfMode = bRfMode;
			m_pCurCard->BlockNumberBit = 0;
			*ppCard = m_pCurCard;
		}
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Cards_CreateNew15693(IN BYTE bRfMode,
IN BYTE* pbData,
IN DWORD dwCount,
OUT CLSCRF_CARD** ppCard)
{
	LONG Status = SCARD_S_SUCCESS;

	//CLSCRF_CARD* pCard = NULL;
	BYTE Size = 0;
	DWORD Index = 0;
	do
	{
		Size = pbData[Index + 1];
		Index += 2;

		CLSCRF_CARD* pSrchCard = NULL;
		CLSCRF_Cards_Search(CLSCRF_UID_LENGTH_ISO15693, &pbData[Index + 2], &pSrchCard);
		if (pbData[Index] == 0
			&& Size
			&& pSrchCard == NULL)
		{
			m_pCurCard = new CLSCRF_CARD;
			m_pCurCard->Timeout = 6800;     // Timeout = 64 ms
			m_pCurCard->CardType = CLSCRF_ICC_TYPE_ISO15693;
			m_pCurCard->UID_Len = CLSCRF_UID_LENGTH_ISO15693;
			memcpy(&m_pCurCard->UID, &pbData[Index + 2], CLSCRF_UID_LENGTH_ISO15693);
			m_pCurCard->ATQ.DSFID = pbData[Index + 1];

			m_pCurCard->CID = 0xff;

			CLSCRF_Cards_Add(m_pCurCard);
			m_pCurCard->Active = 1;
			m_pCurCard->RfMode = bRfMode;
			m_pCurCard->BlockNumberBit = 0;
			*ppCard = m_pCurCard;
		}
		Index += Size;

	} while (Index < dwCount);

	return Status;
}
/*
CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Cards_CreateNewEPCUID(	IN BYTE bRfMode,
IN DWORD dwUID_Len,
IN BYTE* pbData,
IN DWORD dwCount,
OUT CLSCRF_CARD** ppCard )
{
LONG Status = SCARD_S_SUCCESS;

//CLSCRF_CARD* pCard = NULL;
BYTE Size = 0;
DWORD Index = 1;
int i, n;

do
{
Size = pbData[Index+1];
if( dwUID_Len == CLSCRF_UID_LENGTH_EPC && pbData[Index] == 0xFE )
{
pbData[Index] = 0;
n = Index + 2 + dwUID_Len + 2;
for( i = Index + 2; i < n ; i++ )
if( pbData[i] )
{
pbData[Index] = 0xFE;
break;
}
}

CLSCRF_CARD* pSrchCard = NULL;
CLSCRF_Cards_Search(dwUID_Len, &pbData[Index+2], &pSrchCard);
if(   ( pbData[Index] == 0 )
&& ( Size == dwUID_Len + 2 )
&& pSrchCard == NULL )
{
m_pCurCard = new CLSCRF_CARD;
m_pCurCard->Timeout = 6800;     // Timeout = 64 ms
if( dwUID_Len == CLSCRF_UID_LENGTH_EPC )
m_pCurCard->CardType = CLSCRF_ICC_TYPE_ICODE_EPC;
else if( dwUID_Len == CLSCRF_UID_LENGTH_UID )
m_pCurCard->CardType = CLSCRF_ICC_TYPE_ICODE_UID;
else
m_pCurCard->CardType = 0;
m_pCurCard->UID_Len = dwUID_Len;
memcpy( &m_pCurCard->UID, &pbData[Index+2], dwUID_Len );
memcpy( &m_pCurCard->ATQ.CRC16, &pbData[Index+2+dwUID_Len], 2 );

m_pCurCard->CID = 0xff;

CLSCRF_Cards_Add( m_pCurCard );
m_pCurCard->Active = 1;
m_pCurCard->RfMode = bRfMode;
*ppCard = m_pCurCard;
}

Index += Size + 2;

} while( Index < dwCount );

return Status;
}
*/
CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Cards_Delete(IN int UID_Len,
IN PBYTE pUID)
{
	LONG Status = SCARD_E_UNKNOWN_CARD;

	for (std::vector<CLSCRF_CARD*>::iterator iterCard = m_CardList.begin(); iterCard != m_CardList.end(); iterCard++)
	{
		if (UID_Len == (*iterCard)->UID_Len
			&& memcmp(pUID, (*iterCard)->UID, UID_Len) == 0)
		{
			FreeCID((*iterCard)->CID);
			if ((*iterCard) == m_pCurCard)
				m_pCurCard = NULL;
			delete (*iterCard);
			m_CardList.erase(iterCard);
			return SCARD_S_SUCCESS;
		}
	}
	return Status;
}

LPSTR FindInMaskedList(LPCSTR szMaskedList, LPCSTR szATR)
{
	PCHAR szCurrentPos = (PCHAR)szMaskedList;
	int iATRSize = strlen(szATR);
	if (iATRSize == 0)
		return NULL;
	int i;
	while (szCurrentPos && (szCurrentPos = (PCHAR)strchr(szCurrentPos, szATR[0])))
	{
		for (i = 1; i < iATRSize; i++)
			if (
				(szCurrentPos[i] != szATR[i])
				&& (szCurrentPos[i] != '.')
				)
				break;

		if (i == iATRSize)
			return szCurrentPos;

		if (i > 5)
		{
			i++;
			i--;
		}
		szCurrentPos += i;
		szCurrentPos = (PCHAR)strchr(szCurrentPos, '\n');
	}

	return NULL;
}

/*HRSRC hResource = NULL;
HGLOBAL hLoadedResource = NULL;
CHAR ATRMaskedList[200 * 1024] = { 0 };
DWORD dwATRMaskedListSize = 0;

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_Cards_CheckATR(IN LPCSTR szATR,
IN LPSTR szCardInfo,
IN DWORD dwCardInfoMaxSize)
{
	LONG Status = SCARD_S_SUCCESS;
	//HMODULE hi = _AtlBaseModule.m_hInst;//GetCurrModuleHandle();
	//AFX_MANAGE_STATE(AfxGetStaticModuleState( ));
	//HINSTANCE dll_instance = AfxGetInstanceHandle();
	//HINSTANCE dll_instance1 = GetModuleHandle("Clscrfl.dll");
	if (hResource == NULL)
	{
		//hResource = FindResource(_AtlBaseModule.m_hInst, MAKEINTRESOURCE(IDR_HTML_ATR_LIST), RT_HTML);
		hResource = FindResource(_AtlBaseModule.m_hInst, MAKEINTRESOURCE(IDR_TEXT_ATR_LIST), _T("TEXTLIST"));
		//hSrc = FindResource((HMODULE)h, MAKEINTRESOURCE(IDR_TEXT_ATR_LIST), "TEXTLIST");	
		if (hResource)
		{
			if (hLoadedResource == NULL)
				hLoadedResource = LoadResource(_AtlBaseModule.m_hInst, hResource);

			if (hLoadedResource)
			{
				LPVOID pLockedResource = LockResource(hLoadedResource);

				if (pLockedResource)
				{
					dwATRMaskedListSize = SizeofResource(_AtlBaseModule.m_hInst, hResource);

					if (0 != dwATRMaskedListSize)
					{
						memcpy(
							ATRMaskedList,
							pLockedResource,
							min(dwATRMaskedListSize, sizeof(ATRMaskedList))
							);
					}

					UnlockResource(pLockedResource);
				}
			}
		}
	}

	if (dwATRMaskedListSize > 0)
	{
		PCHAR pATR = FindInMaskedList(ATRMaskedList, szATR + 1);
		if (pATR != NULL)
		{
			PCHAR pATRnl = strstr(pATR, _T("\r\n"));
			if (pATRnl == NULL)
				pATRnl = strstr(pATR, _T("\n"));
			if (pATRnl != NULL)
			{
				PCHAR pCardInfo = pATRnl + 2;

				PCHAR pCardInfoEnd = strstr(pCardInfo, _T("\r\n\r\n"));
				if (pCardInfoEnd == NULL)
					pCardInfoEnd = strstr(pCardInfo, _T("\n\n"));

				if (pCardInfoEnd != NULL)
				{
					memcpy(
						szCardInfo,
						pCardInfo,
						min(pCardInfoEnd - pCardInfo, dwCardInfoMaxSize - 1)
						);
					szCardInfo[min(pCardInfoEnd - pCardInfo, dwCardInfoMaxSize - 1)] = 0x00;
				}
			}

		}
	}
	// Use pLockedResource and dwResourceSize however you want


	return Status;
}*/
