#include "StdAfx.h"

#include "Clscrfl.h"

extern std::map<PVOID, READER_SETTINGS> m_ReaderSettings;
extern CLSCRF_CARD*  m_pCurCard;


LONG Transmit_DESFire_Cmd(IN PVOID pReader,
	IN LPCBYTE pbSendBuffer,
	IN DWORD cbSendLength,
	OUT LPBYTE pbRecvBuffer,
	IN OUT LPDWORD pcbRecvLength,
	IN bool fMacSend = true,
	IN bool fMacReceive = true)
{
	CLSCRF_CARD* pCard = m_pCurCard;

	if (pCard == NULL)
	{
		*pcbRecvLength = 0;
		return SCARD_E_NO_SMARTCARD;
	}

	const int BUF_SIZE = 300;
	BYTE SendBuffer[1024*16] = { 0 };
	BYTE RecvBuffer[1024*16];
	//    DWORD SendLength;
	DWORD RecvLength;
	LONG Status = SCARD_S_SUCCESS;

	//    unsigned char PCB;
	int TCardLength = 0;
	int ParamLength = 0;
	int RChainLength = 0;
	int TChainPointer;
	int TChainRemainder;
	int BufferWorkSize = cCLSCRF_PCD_BufSize;
	DWORD CardBufSize = cCLSCRF_PCD_BufSize;

	if (pCard->ATQ.ATS[0] > 0)
		CardBufSize = cCLSCRF_DecodeBufSize[pCard->ATQ.ATS[1] & 0x0f];
	else
		CardBufSize = 32;   // Default

	if (CardBufSize < cCLSCRF_PCD_BufSize)
		BufferWorkSize = CardBufSize;

	if (cbSendLength < 1)
	{
		*pcbRecvLength = 0;
		return SCARD_E_INVALID_PARAMETER;
	}
	TChainPointer = 0;
	TChainRemainder = cbSendLength;
	TCardLength = TChainRemainder + 2;  // PCB, CID + Params
	if (TCardLength + 2 > BufferWorkSize)    // 2 bytes CRC trailer
		TCardLength = BufferWorkSize - 2;
	SendBuffer[0] = 0x0a | pCard->BlockNumberBit;
	SendBuffer[1] = pCard->CID;
	ParamLength = TCardLength - 2;
	if (ParamLength > 0)  // Size of Params
	{
		memcpy(&SendBuffer[2], &pbSendBuffer[TChainPointer], ParamLength);
		TChainRemainder -= ParamLength;
		TChainPointer += ParamLength;
	}

	if (fMacSend)
		if (
			(m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType == CLSCRF_DF_AUTH_TYPE_3K3DES)
			|| (m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType == CLSCRF_DF_AUTH_TYPE_AES))
		{
			DFEncrypt(
				pReader,
				SendBuffer + 2,
				m_ReaderSettings[pReader].CurrentDFSettings()->IV,
				TCardLength - 2,
				NULL,
				true,
				true);
		}

	*pcbRecvLength = 0;
	bool bComplete = true;
	do{
		RecvLength = BUF_SIZE;

		Status = CLSCRF_DirectIO_Card(pReader,
			SendBuffer, TCardLength,
			pCard->Timeout,
			RecvBuffer, &RecvLength);
		if (Status)
		{
			bComplete = true;
		}
		else if (RecvLength)
		{
			pCard->BlockNumberBit ^= 0x01;

			if (((RecvBuffer[0] & 0xfe) == 0x0a) && ((SendBuffer[2] & 0xFF) != 0x0A) && ((SendBuffer[2] & 0xFF) != 0x1A) && ((SendBuffer[2] & 0xFF) != 0xAA)) // 0x0A or 0x0B (I-block)
			{
				if (RecvBuffer[2] == 0x00 || RecvBuffer[2] == 0xaf)  // Normal or Chaining
				{
					RChainLength = 0;
					memcpy(&RChainLength, &RecvLength, 2);
					RChainLength -= 3;
					if (RChainLength > 0)
					{
						memcpy(&pbRecvBuffer[*pcbRecvLength], &RecvBuffer[3], RChainLength);
						*pcbRecvLength += RChainLength;
					}
					if (RecvBuffer[2] == 0xaf)
					{
						TCardLength = TChainRemainder + 3;  // PCB, CID, 0xaf  + Params
						if (TCardLength + 2 > BufferWorkSize)    // 2 bytes CRC trailer
							TCardLength = BufferWorkSize - 2;
						SendBuffer[0] = 0x0a | pCard->BlockNumberBit;
						SendBuffer[1] = pCard->CID;
						SendBuffer[2] = 0xaf;
						ParamLength = TCardLength - 3;
						if (ParamLength > 0)  // Size of Params
						{
							memcpy(&SendBuffer[3], &pbSendBuffer[TChainPointer], ParamLength);
							TChainRemainder -= ParamLength;
							TChainPointer += ParamLength;
						}
						bComplete = false;
					}
					else
					{
						bComplete = true;
					}
				}
				else
				{
					*pcbRecvLength = 1;
					Status = SCARD_F_INTERNAL_ERROR;
					memcpy(pbRecvBuffer, &RecvBuffer[2], 1);
					CTransfer* pTr = (CTransfer*)pReader;
					pTr->m_RcBuf.Clear();
					pTr->m_RcBuf.AddBulk((LPBYTE)pbRecvBuffer, 1);
					bComplete = true;
				}
			}
			else if (RecvBuffer[0] == 0xfa) // WTX (S-block)
			{
				TCardLength = 3;  // PCB, CID, 0xaf
				memcpy(&SendBuffer[0], &RecvBuffer[0], 3);
				bComplete = false;
			}
			else if ((((SendBuffer[2] & 0xFF) == 0x0A) || ((SendBuffer[2] & 0xFF) == 0x1A) || ((SendBuffer[2] & 0xFF) == 0xAA)) && ((RecvBuffer[2] == 0xAF) || (RecvBuffer[2] == 0x00))) //k)
			{
				memcpy(pbRecvBuffer, &RecvBuffer[3], RecvLength - 3);
				*pcbRecvLength = RecvLength - 3;
				//pCard->BlockNumberBit ^= 0x01;
				bComplete = true;
			}
			else
			{
				*pcbRecvLength = 1;
				Status = SCARD_F_INTERNAL_ERROR;
				memcpy(pbRecvBuffer, &RecvBuffer[2], *pcbRecvLength);
				CTransfer* pTr = (CTransfer*)pReader;
				pTr->m_RcBuf.Clear();
				pTr->m_RcBuf.AddBulk((LPBYTE)pbRecvBuffer, 1);
				bComplete = true;
			}
		}

	} while (!bComplete);

	if (Status == SCARD_S_SUCCESS)
		if (fMacReceive)
			if (*pcbRecvLength >= 8)
				if ((m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType == CLSCRF_DF_AUTH_TYPE_3K3DES)
					|| (m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType == CLSCRF_DF_AUTH_TYPE_AES)
					)
				{
					BYTE MACBuf[256] = { 0 };
					int iDataSize = *pcbRecvLength - 8;
					if (iDataSize > 0)
						memcpy(MACBuf, pbRecvBuffer, iDataSize);
					MACBuf[iDataSize] = 0x00;//RecvBuffer[2];

					DFEncrypt(
						pReader,
						MACBuf,
						m_ReaderSettings[pReader].CurrentDFSettings()->IV,
						iDataSize + 1,
						NULL,
						true,
						true
						);

					if (memcmp(&pbRecvBuffer[*pcbRecvLength - 8], m_ReaderSettings[pReader].CurrentDFSettings()->IV, 8) != 0)
						Status = SCARD_E_INVALID_VALUE;
					else
						*pcbRecvLength -= 8;
				}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_SetTransferType(IN LPVOID pReader,
IN BYTE ucTransferType)
{
	LONG Status = SCARD_S_SUCCESS;

	if (ucTransferType == 0x02)
		ucTransferType = 0x03;

	m_ReaderSettings[pReader].CurrentDFSettings()->TransferType = ucTransferType;

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_ChangeKeySettings(IN LPVOID pReader,
IN CLSCRF_DESFIRE_MASTER_KEY_SETTINGS* pMasterKeySettings)
{
	LONG Status = SCARD_S_SUCCESS;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	//BYTE pKeySettings[16] = {0};
	SendBuffer[SendLength++] = 0x54;
	memcpy(&SendBuffer[SendLength], (PVOID)pMasterKeySettings, 1); SendLength++;

	BYTE TT = m_ReaderSettings[pReader].CurrentDFSettings()->TransferType;
	m_ReaderSettings[pReader].CurrentDFSettings()->TransferType = CLSCRF_DF_TRANSFER_TYPE_ENCRYPTED;

	DFEncrypt(pReader, SendBuffer + 1, SendBuffer + 1, 1, NULL, true, false);//, true);

	m_ReaderSettings[pReader].CurrentDFSettings()->TransferType = TT;

	SendLength = 1 + ((m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType == CLSCRF_DF_AUTH_TYPE_AES) ? 16 : 8);

	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength,
		(m_ReaderSettings[pReader].CurrentDFSettings()->TransferType == CLSCRF_DF_TRANSFER_TYPE_PLAIN), true);

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_GetKeySettings(IN LPVOID pReader,
OUT CLSCRF_DESFIRE_MASTER_KEY_SETTINGS_AND_LENGTH* pKeyData)
{
	LONG Status = SCARD_S_SUCCESS;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	SendBuffer[SendLength++] = 0x45;

	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength);

	if (Status == 0)
	{
		memcpy((PVOID)pKeyData, &RecvBuffer[0], 2);
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_ChangeKey(IN LPVOID pReader,
IN CLSCRF_DESFIRE_KEY_DATA* pKeyData)
{
	LONG Status = SCARD_S_SUCCESS;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	//BYTE pCypheredKeyData[40] = {0};


	//BYTE TestKey[16] = {0xB0, 0xB1, 0xB2, 0xB3, 0xB4, 0xB5, 0xB6, 0xB7, 0xB8, 0xB9, 0xBA, 0xBB, 0xBC, 0xBD, 0xBE, 0xBF};
	//memcpy(pKeyData->Key, TestKey, 16);
	//BYTE TestSess[16] = {0x1C, 0x94, 0xD1, 0x5B, 0x50, 0x7F, 0x86, 0x2C, 0x6D, 0xD3, 0xC3, 0xBE, 0xF2, 0xC8, 0xFA, 0x75};
	//memcpy(m_DF_Settings[pReader].SessionKey, TestSess, 16);


	SendBuffer[SendLength++] = 0xC4;
	memcpy(&SendBuffer[SendLength], (PVOID)&pKeyData->KeyNumber, 1); SendLength++;
	memcpy(&SendBuffer[SendLength], pKeyData->Key, pKeyData->ucKeySize); SendLength += pKeyData->ucKeySize;
	if (pKeyData->fIsAESKey)
		SendBuffer[SendLength++] = pKeyData->ucAESKeyVersion;

	BYTE TT = m_ReaderSettings[pReader].CurrentDFSettings()->TransferType;
	m_ReaderSettings[pReader].CurrentDFSettings()->TransferType = CLSCRF_DF_TRANSFER_TYPE_ENCRYPTED;
	BYTE AT = m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType;
	if (AT == CLSCRF_DF_AUTH_TYPE_DES)
	{
		memcpy(
			&m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey[m_ReaderSettings[pReader].CurrentDFSettings()->SessionKeySize],
			m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey,
			m_ReaderSettings[pReader].CurrentDFSettings()->SessionKeySize
			);
		m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType = CLSCRF_DF_AUTH_TYPE_3DES;
	}
	int iEncSize = 0;
	DFEncrypt(
		pReader,
		SendBuffer + 2,
		SendBuffer + 2,
		SendLength - 2,
		&iEncSize,
		false,
		false,
		SendBuffer,
		2
		);
	SendLength = 2 + iEncSize;
	m_ReaderSettings[pReader].CurrentDFSettings()->TransferType = TT;
	m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType = AT;
	//SendLength = 2+pKeyData->ucKeySize+(pKeyData->fIsAESKey)?16:0;

	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength,
		(m_ReaderSettings[pReader].CurrentDFSettings()->TransferType == CLSCRF_DF_TRANSFER_TYPE_PLAIN), true);

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_GetKeyVersion(IN LPVOID pReader,
IN BYTE ucKeyNumber,
OUT PBYTE pucKeyVersion)
{
	LONG Status = SCARD_S_SUCCESS;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength;
	DWORD RecvLength = BUF_SIZE;

	SendBuffer[0] = 0x64;
	SendBuffer[1] = ucKeyNumber;
	SendLength = 2;

	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength);

	if (Status == 0)
	{
		*pucKeyVersion = RecvBuffer[0];
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_CreateApplication(IN LPVOID pReader,
IN PBYTE pAID,
IN CLSCRF_DESFIRE_APPLICATION_MASTER_KEY_SETTINGS* pKeySett1,
IN CLSCRF_DESFIRE_NEW_APPLICATION_KEY_SETTINGS* pKeySett2,
IN PBYTE pIsoFileID,
IN PBYTE pIso7816DfName,
IN int iIso7816DfNameLength)
{
	LONG Status = SCARD_S_SUCCESS;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	SendBuffer[SendLength] = 0xCA; SendLength++;
	memcpy(&SendBuffer[SendLength], pAID, 3); SendLength += 3;
	memcpy(&SendBuffer[SendLength], pKeySett1, 1); SendLength++;
	memcpy(&SendBuffer[SendLength], pKeySett2, 1); SendLength++;
	if (pIsoFileID)
	{
		memcpy(&SendBuffer[SendLength], pIsoFileID, 2); SendLength += 2;
	}
	if (pIso7816DfName && iIso7816DfNameLength)
	{
		memcpy(&SendBuffer[SendLength], pIso7816DfName, iIso7816DfNameLength); SendLength += iIso7816DfNameLength;
	}

	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength);

	if (Status == 0)
	{
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_DeleteApplication(IN LPVOID pReader,
IN PBYTE pAID)
{
	LONG Status = SCARD_S_SUCCESS;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	SendBuffer[SendLength] = 0xDA; SendLength++;
	memcpy(&SendBuffer[SendLength], pAID, 3); SendLength += 3;

	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength);

	if (Status == 0)
	{
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_GetApplicationIDs(IN LPVOID pReader,
OUT PBYTE pAIDs,
OUT PBYTE pucApplicationsCount)
{
	LONG Status = 0xAF;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	int iAIDsTotal = 0;

	SendBuffer[SendLength] = 0x6A; SendLength++;

	while (Status == 0xAF)
	{
		Status = Transmit_DESFire_Cmd(pReader,
			SendBuffer, SendLength,
			RecvBuffer, &RecvLength);

		if ((Status == 0) || (Status == 0xAF))
		{
			int iAIDsCount = RecvLength / 3;

			for (int i = 0; i < iAIDsCount; i++)
			{
				memcpy(&pAIDs[iAIDsTotal * 3], &RecvBuffer[i * 3], 3);
				iAIDsTotal++;
			}
		}
	}

	*pucApplicationsCount = (BYTE)iAIDsTotal;

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_GetDFNames(IN LPVOID pReader,
OUT CLSCRF_DESFIRE_DFNAME* pDFNames,
OUT PBYTE pNamesCount)
{
	LONG Status = 0xAF;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	int iAIDsTotal = 0;

	SendBuffer[SendLength] = 0x6D; SendLength++;

	while (Status == 0xAF)
	{
		Status = Transmit_DESFire_Cmd(pReader,
			SendBuffer, SendLength,
			RecvBuffer, &RecvLength);

		if ((Status == 0) || (Status == 0xAF))
		{
			memcpy(&pDFNames[iAIDsTotal*sizeof(CLSCRF_DESFIRE_DFNAME)], &RecvBuffer[0], RecvLength);
			pDFNames[iAIDsTotal*sizeof(CLSCRF_DESFIRE_DFNAME)].DFNameSize = (BYTE)(RecvLength - 5);
			iAIDsTotal++;
		}
	}

	if (Status == 0)
		*pNamesCount = (BYTE)iAIDsTotal;

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_SelectApplication(IN LPVOID pReader,
IN PBYTE pAID)
{
	LONG Status = SCARD_S_SUCCESS;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	SendBuffer[SendLength] = 0x5A; SendLength++;
	memcpy(&SendBuffer[SendLength], pAID, 3); SendLength += 3;

	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength);

	if (Status == 0)
	{
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_FormatPICC(IN LPVOID pReader)
{
	LONG Status = SCARD_S_SUCCESS;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	SendBuffer[SendLength] = 0xFC; SendLength++;

	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength);

	if (Status == 0)
	{
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_GetVersion(IN LPVOID pReader,
OUT CLSCRF_DESFIRE_HW_SW_INFO* pHWInfo,
OUT CLSCRF_DESFIRE_HW_SW_INFO* pSWInfo,
OUT CLSCRF_DESFIRE_MORE_INFO* pMoreInfo)
{
	LONG Status = SCARD_S_SUCCESS;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	CLSCRF_MifareDesFire_ClearAuthState(pReader);

	SendBuffer[SendLength] = 0x60; SendLength++;

	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength);

	if (Status == 0xAF)
	{
		if (pHWInfo)
			memcpy(pHWInfo, &RecvBuffer[0], 7);

		SendBuffer[0] = 0xAF;
		Status = Transmit_DESFire_Cmd(pReader,
			SendBuffer, SendLength,
			RecvBuffer, &RecvLength);

		if (Status == 0xAF)
		{
			if (pSWInfo)
				memcpy(pSWInfo, &RecvBuffer[0], 7);

			Status = Transmit_DESFire_Cmd(pReader,
				SendBuffer, SendLength,
				RecvBuffer, &RecvLength);

			if (Status == 0x00)
			{
				if (pMoreInfo)
					memcpy(pMoreInfo, &RecvBuffer[0], 14);
			}
			else
				Status = SCARD_F_INTERNAL_ERROR;
		}
		else
			Status = SCARD_F_INTERNAL_ERROR;
	}
	else if (Status == 0x00)
	{
		if (pHWInfo)
			memcpy(pHWInfo, &RecvBuffer[0], sizeof(CLSCRF_DESFIRE_HW_SW_INFO));
		if (pSWInfo)
			memcpy(pSWInfo, &RecvBuffer[sizeof(CLSCRF_DESFIRE_HW_SW_INFO)], sizeof(CLSCRF_DESFIRE_HW_SW_INFO));
		if (pMoreInfo)
			memcpy(pMoreInfo, &RecvBuffer[sizeof(CLSCRF_DESFIRE_HW_SW_INFO) * 2], sizeof(CLSCRF_DESFIRE_MORE_INFO));
	}
	else
		Status = SCARD_F_INTERNAL_ERROR;

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_FreeMemory(IN LPVOID pReader,
OUT PDWORD pdwMemSize)
{
	LONG Status = 0xAF;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	ZeroMemory(pdwMemSize, 4);

	SendBuffer[SendLength] = 0x6E; SendLength++;

	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength);

	if (Status == 0)
	{
		memcpy(&pdwMemSize, &RecvBuffer[0], 3);
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_SetConfiguration(IN LPVOID pReader,
IN CLSCRF_DESFIRE_CONFIGURATION* pCfg)
{
	LONG Status = SCARD_S_SUCCESS;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	SendBuffer[SendLength] = 0x5C; SendLength++;
	SendBuffer[SendLength] = pCfg->ucOption; SendLength++;

	bool f80Padding = false;

	switch (pCfg->ucOption)
	{
	case 0x00:
		SendBuffer[SendLength] = pCfg->ucConfigurationByte; SendLength++;
		break;
	case 0x01:
		memcpy(&SendBuffer[SendLength], pCfg->Key, 24); SendLength += 24;
		SendBuffer[SendLength] = pCfg->KeyVersion; SendLength++;
		break;
	case 0x02:
		memcpy(&SendBuffer[SendLength], pCfg->ATS, pCfg->ATSSize); SendLength += pCfg->ATSSize;
		f80Padding = true;
		break;
	default:
		break;
	}

	//DWORD dwCRC32 = CRC_32(&SendBuffer[2], SendLength-2);
	//memcpy(&SendBuffer[SendLength], &dwCRC32, 4); SendLength+=4;


	DFEncrypt(pReader, &SendBuffer[2], &SendBuffer[2], SendLength - 2, NULL, f80Padding);

	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength,
		(m_ReaderSettings[pReader].CurrentDFSettings()->TransferType == CLSCRF_DF_TRANSFER_TYPE_PLAIN), true);

	if (Status == 0)
	{
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_GetCardUID(IN LPVOID pReader,
OUT PBYTE pUID)
{
	LONG Status = SCARD_S_SUCCESS;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	SendBuffer[SendLength] = 0x51; SendLength++;

	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength);

	if (Status == 0)
	{
		memcpy(pUID, RecvBuffer, 16);
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_GetFileIDs(IN LPVOID pReader,
OUT PBYTE pFileIDs,
OUT PBYTE pucFilesCount)
{
	LONG Status = SCARD_S_SUCCESS;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	SendBuffer[SendLength] = 0x6F; SendLength++;

	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength);

	if (Status == 0)
	{
		for (DWORD i = 0; i < RecvLength; i++)
		{
			pFileIDs[i] = RecvBuffer[i];
		}

		*pucFilesCount = (BYTE)RecvLength;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_GetISOFileIDs(IN LPVOID pReader,
OUT PWORD pISOFileIDs,
OUT PBYTE pucFilesCount)
{
	LONG Status = 0xAF;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	SendBuffer[SendLength] = 0x61; SendLength++;

	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength);

	int iFilesCount = 0;
	while ((Status == 0xAF) || (Status == 0x00))
	{
		for (DWORD i = 0; i < RecvLength; i++)
		{
			memcpy(&pISOFileIDs[iFilesCount * 2], &RecvBuffer[i * 2], 2);
			iFilesCount++;
		}
	}

	*pucFilesCount = (BYTE)iFilesCount;

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_GetFileSettings(IN LPVOID pReader,
IN BYTE ucFileNumber,
OUT CLSCRF_DESFIRE_FILE_SETTINGS* pFileSettings)
{
	LONG Status = SCARD_S_SUCCESS;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	SendBuffer[SendLength] = 0xF5; SendLength++;
	SendBuffer[SendLength] = ucFileNumber; SendLength++;

	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength);

	if (Status == 0)
	{
		memcpy(pFileSettings, &RecvBuffer[0], RecvLength);
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_ChangeFileSettings(IN LPVOID pReader,
IN BYTE ucFileNumber,
IN BYTE ucCommSettings,
IN CLSCRF_DESFIRE_FILE_ACCESS_RIGHTS* pAccessRights)
{
	LONG Status = SCARD_S_SUCCESS;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	SendBuffer[SendLength] = 0x5F; SendLength++;
	SendBuffer[SendLength] = ucFileNumber; SendLength++;
	SendBuffer[SendLength] = ucCommSettings; SendLength++;
	memcpy(&SendBuffer[SendLength], pAccessRights, 2); SendLength += 2;

	BYTE TT = m_ReaderSettings[pReader].CurrentDFSettings()->TransferType;
	m_ReaderSettings[pReader].CurrentDFSettings()->TransferType = CLSCRF_DF_TRANSFER_TYPE_ENCRYPTED;

	int iResSize = 0;
	DFEncrypt(
		pReader,
		&SendBuffer[2],
		&SendBuffer[2],
		3,
		&iResSize);
	SendLength = SendLength - 3 + iResSize;

	m_ReaderSettings[pReader].CurrentDFSettings()->TransferType = TT;

	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength,
		(m_ReaderSettings[pReader].CurrentDFSettings()->TransferType == CLSCRF_DF_TRANSFER_TYPE_PLAIN), true);

	if (Status == 0)
	{
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_CreateStdDataFile(IN LPVOID pReader,
IN BYTE ucFileNumber,
IN PBYTE pISOFileID,
IN BYTE ucCommSettings,
IN CLSCRF_DESFIRE_FILE_ACCESS_RIGHTS* pAccessRights,
IN DWORD dwFileSize)
{
	LONG Status = SCARD_S_SUCCESS;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	if (ucCommSettings == 0x02)
		ucCommSettings = 0x03;
	SendBuffer[SendLength] = 0xCD; SendLength++;
	SendBuffer[SendLength] = ucFileNumber; SendLength++;
	if (pISOFileID)
	{
		memcpy(&SendBuffer[SendLength], &pISOFileID[0], 2); SendLength += 2;
	}
	SendBuffer[SendLength] = ucCommSettings; SendLength++;
	memcpy(&SendBuffer[SendLength], pAccessRights, 2); SendLength += 2;
	memcpy(&SendBuffer[SendLength], &dwFileSize, 3); SendLength += 3;

	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength);

	if (Status == 0)
	{
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_CreateBackupDataFile(IN LPVOID pReader,
IN BYTE ucFileNumber,
IN PBYTE pISOFileID,
IN BYTE ucCommSettings,
IN CLSCRF_DESFIRE_FILE_ACCESS_RIGHTS* pAccessRights,
IN DWORD dwFileSize)
{
	LONG Status = SCARD_S_SUCCESS;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	if (ucCommSettings == 0x02)
		ucCommSettings = 0x03;

	SendBuffer[SendLength] = 0xCB; SendLength++;
	SendBuffer[SendLength] = ucFileNumber; SendLength++;
	if (pISOFileID)
	{
		memcpy(&SendBuffer[SendLength], &pISOFileID[0], 2); SendLength += 2;
	}
	SendBuffer[SendLength] = ucCommSettings; SendLength++;
	memcpy(&SendBuffer[SendLength], pAccessRights, 2); SendLength += 2;
	memcpy(&SendBuffer[SendLength], &dwFileSize, 3); SendLength += 3;

	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength);

	if (Status == 0)
	{
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_CreateValueFile(IN LPVOID pReader,
IN BYTE ucFileNumber,
IN BYTE ucCommSettings,
IN CLSCRF_DESFIRE_FILE_ACCESS_RIGHTS* pAccessRights,
IN DWORD dwLowerLimit,
IN DWORD dwUpperLimit,
IN DWORD dwValue,
IN CLSCRF_DESFIRE_LIMITED_CREDIT_ENABLED* pLimitedCreditEnabled)
{
	LONG Status = SCARD_S_SUCCESS;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	CLSCRF_DESFIRE_LIMITED_CREDIT_ENABLED lcr = { 0 };
	lcr.fEnableFreeGetValue = 1;
	lcr.fEnableLimitedCredit = 1;

	if (!pLimitedCreditEnabled)
		pLimitedCreditEnabled = &lcr;

	if (ucCommSettings == 0x02)
		ucCommSettings = 0x03;

	SendBuffer[SendLength] = 0xCC; SendLength++;
	SendBuffer[SendLength] = ucFileNumber; SendLength++;
	SendBuffer[SendLength] = ucCommSettings; SendLength++;
	memcpy(&SendBuffer[SendLength], pAccessRights, 2); SendLength += 2;
	memcpy(&SendBuffer[SendLength], &dwLowerLimit, 4); SendLength += 4;
	memcpy(&SendBuffer[SendLength], &dwUpperLimit, 4); SendLength += 4;
	memcpy(&SendBuffer[SendLength], &dwValue, 4); SendLength += 4;
	memcpy(&SendBuffer[SendLength], (void*)pLimitedCreditEnabled, 1); SendLength++;

	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength);

	if (Status == 0)
	{
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_CreateLinearRecordFile(IN LPVOID pReader,
IN BYTE ucFileNumber,
IN PBYTE pISOFileID,
IN BYTE ucCommSettings,
IN CLSCRF_DESFIRE_FILE_ACCESS_RIGHTS* pAccessRights,
IN DWORD dwRecordSize,
IN DWORD dwMaxNumOfRecords)
{
	LONG Status = SCARD_S_SUCCESS;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	if (ucCommSettings == 0x02)
		ucCommSettings = 0x03;

	SendBuffer[SendLength] = 0xC1; SendLength++;
	SendBuffer[SendLength] = ucFileNumber; SendLength++;
	if (pISOFileID)
	{
		memcpy(&SendBuffer[SendLength], &pISOFileID[0], 2); SendLength += 2;
	}
	SendBuffer[SendLength] = ucCommSettings; SendLength++;
	memcpy(&SendBuffer[SendLength], pAccessRights, 2); SendLength += 2;
	memcpy(&SendBuffer[SendLength], &dwRecordSize, 3); SendLength += 3;
	memcpy(&SendBuffer[SendLength], &dwMaxNumOfRecords, 3); SendLength += 3;

	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength);

	if (Status == 0)
	{
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_CreateCyclicRecordFile(IN LPVOID pReader,
IN BYTE ucFileNumber,
IN PBYTE pISOFileID,
IN BYTE ucCommSettings,
IN CLSCRF_DESFIRE_FILE_ACCESS_RIGHTS* pAccessRights,
IN DWORD dwRecordSize,
IN DWORD dwMaxNumOfRecords)
{
	LONG Status = SCARD_S_SUCCESS;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	if (ucCommSettings == 0x02)
		ucCommSettings = 0x03;

	SendBuffer[SendLength] = 0xC0; SendLength++;
	SendBuffer[SendLength] = ucFileNumber; SendLength++;
	if (pISOFileID)
	{
		memcpy(&SendBuffer[SendLength], &pISOFileID[0], 2); SendLength += 2;
	}
	SendBuffer[SendLength] = ucCommSettings; SendLength++;
	memcpy(&SendBuffer[SendLength], pAccessRights, 2); SendLength += 2;
	memcpy(&SendBuffer[SendLength], &dwRecordSize, 3); SendLength += 3;
	memcpy(&SendBuffer[SendLength], &dwMaxNumOfRecords, 3); SendLength += 3;

	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength);

	if (Status == 0)
	{
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_DeleteFile(IN LPVOID pReader,
IN BYTE ucFileNumber)
{
	LONG Status = SCARD_S_SUCCESS;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	SendBuffer[SendLength] = 0xDF; SendLength++;
	SendBuffer[SendLength] = ucFileNumber; SendLength++;

	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength);

	if (Status == 0)
	{
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_ReadData(IN LPVOID pReader,
IN BYTE ucFileNumber,
IN DWORD dwOffset,
IN DWORD dwLength,
OUT PBYTE pData,
OUT PDWORD pdwDataLength)
{
	LONG Status = 0xAF;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[1024*16];
	DWORD SendLength = 0;
	DWORD RecvLength = sizeof(RecvBuffer);

	//BYTE tmpData[4096] = {0};

	//int iCurByte = 0;
	SendBuffer[SendLength] = 0xBD; SendLength++;
	SendBuffer[SendLength] = ucFileNumber; SendLength++;
	memcpy(&SendBuffer[SendLength], &dwOffset, 3); SendLength += 3;
	memcpy(&SendBuffer[SendLength], &dwLength, 3); SendLength += 3;

	//while (Status == 0xAF)
	//{
	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength,
		true, (m_ReaderSettings[pReader].CurrentDFSettings()->TransferType != CLSCRF_DF_TRANSFER_TYPE_ENCRYPTED));

	//	if ((Status == 0xAF) || (Status == 0x00))
	//	{
	//		memcpy(&tmpData[iCurByte], &RecvBuffer[0], RecvLength); iCurByte+=RecvLength;
	//
	//		SendBuffer[0] = 0xAF; SendLength = 1;
	//	}
	//}

	if (Status == SCARD_S_SUCCESS)
	{
		int iResSize = 0;

		//if (!DFDecrypt(pReader, tmpData, pData, iCurByte, dwLength, &iResSize))
		if (!DFDecrypt(pReader, RecvBuffer, pData, RecvLength, dwLength, &iResSize))
		{
			return SCARD_E_INVALID_VALUE;
		}

		if (pdwDataLength)
			*pdwDataLength = iResSize;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_WriteData(IN LPVOID pReader,
IN BYTE ucFileNumber,
IN DWORD dwOffset,
IN DWORD dwLength,
IN PBYTE pData)
{
	LONG Status = 0xAF;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[1024*16];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	BYTE tmpData[1024 * 16] = { 0 };
	int iResSize = 0;

	int iCurByte = 0;
	SendBuffer[SendLength] = 0x3D; SendLength++;
	SendBuffer[SendLength] = ucFileNumber; SendLength++;
	memcpy(&SendBuffer[SendLength], &dwOffset, 3); SendLength += 3;
	memcpy(&SendBuffer[SendLength], &dwLength, 3); SendLength += 3;
	/*
	if (
	(m_DF_Settings[pReader].LastAuthType == CLSCRF_DF_AUTH_TYPE_3K3DES)
	||(m_DF_Settings[pReader].LastAuthType == CLSCRF_DF_AUTH_TYPE_AES)
	)
	DFEncrypt(pReader, SendBuffer, tmpData, dwLength, &iResSize);
	else*/

	DFEncrypt(pReader, pData, tmpData, dwLength, &iResSize, false, false, SendBuffer, SendLength);
	//xw怈
	//ltSq
	//
	//<supyN
	while (iResSize > 0)
	{
		if ((iResSize > 52) && (SendLength == 8))
		{
			memcpy(&SendBuffer[SendLength], &tmpData[iCurByte], 52); SendLength += 52;
			iCurByte += 52;
			iResSize -= 52;
		}
		else if ((iResSize > 59) && (SendLength == 1))
		{
			memcpy(&SendBuffer[SendLength], &tmpData[iCurByte], 52); SendLength += 59;
			iCurByte += 59;
			iResSize -= 59;
		}
		else
		{
			memcpy(&SendBuffer[SendLength], &tmpData[iCurByte], iResSize); SendLength += iResSize;
			iCurByte += iResSize;
			iResSize -= iResSize;
		}

		Status = Transmit_DESFire_Cmd(pReader,
			SendBuffer, SendLength,
			RecvBuffer, &RecvLength,
			(m_ReaderSettings[pReader].CurrentDFSettings()->TransferType == CLSCRF_DF_TRANSFER_TYPE_PLAIN), true);

		if (Status == 0xAF)
		{
			SendBuffer[0] = 0xAF; SendLength = 1;
		}
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_GetValue(IN LPVOID pReader,
IN BYTE ucFileNumber,
OUT PDWORD pdwValue)
{
	LONG Status = 0xAF;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	SendBuffer[SendLength] = 0x6C; SendLength++;
	SendBuffer[SendLength] = ucFileNumber; SendLength++;

	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength,
		true, (m_ReaderSettings[pReader].CurrentDFSettings()->TransferType != CLSCRF_DF_TRANSFER_TYPE_ENCRYPTED));

	if (Status == 0x00)
	{
		if ((m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType == 0x00) && (m_ReaderSettings[pReader].CurrentDFSettings()->TransferType == CLSCRF_DF_TRANSFER_TYPE_PLAIN))
		{
			memcpy(pdwValue, &RecvBuffer[0], RecvLength);
		}
		if ((m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType == 0x00) && (m_ReaderSettings[pReader].CurrentDFSettings()->TransferType == CLSCRF_DF_TRANSFER_TYPE_ENCRYPTED))
		{
			if (RecvLength == 8)
			{
				DFDecrypt(pReader, &RecvBuffer[0], &RecvBuffer[0], 8);
				//WORD wCrc16 = CRC_16(&RecvBuffer[0], 4);
				WORD wCrc16 = 0;
				ComputeCrc(CRC_A, (char*)&RecvBuffer[0], 4, (PBYTE)&wCrc16, ((PBYTE)&wCrc16) + 1);
				if (memcmp(&wCrc16, &RecvBuffer[4], 2) == 0)
				{
					memcpy(pdwValue, &RecvBuffer[0], RecvLength);
				}
				else
				{
					CTransfer* pTr = (CTransfer*)pReader;
					pTr->m_RcBuf.Clear();
					BYTE err = (BYTE)-1;
					pTr->m_RcBuf.AddBulk(&err, 1);

					return SCARD_W_SECURITY_VIOLATION;
				}
			}
			else
			{
				CTransfer* pTr = (CTransfer*)pReader;
				pTr->m_RcBuf.Clear();
				BYTE err = (BYTE)-1;
				pTr->m_RcBuf.AddBulk(&err, 1);

				return SCARD_W_SECURITY_VIOLATION;
			}
		}
		if (
			(m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType == 0x00)
			&& (m_ReaderSettings[pReader].CurrentDFSettings()->TransferType == CLSCRF_DF_TRANSFER_TYPE_MACCED))
		{
			memcpy(pdwValue, &RecvBuffer[0], RecvLength);
		}
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_Credit(IN LPVOID pReader,
IN BYTE ucFileNumber,
IN DWORD dwValue)
{
	LONG Status = SCARD_S_SUCCESS;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	SendBuffer[SendLength] = 0x0C; SendLength++;
	SendBuffer[SendLength] = ucFileNumber; SendLength++;
	int iResSize = 0;
	DFEncrypt(pReader, (PBYTE)&dwValue, &SendBuffer[SendLength], 4, &iResSize); SendLength += iResSize;

	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength,
		(m_ReaderSettings[pReader].CurrentDFSettings()->TransferType == CLSCRF_DF_TRANSFER_TYPE_PLAIN), true);

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_Debit(IN LPVOID pReader,
IN BYTE ucFileNumber,
IN DWORD dwValue)
{
	LONG Status = SCARD_S_SUCCESS;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	SendBuffer[SendLength] = 0xDC; SendLength++;
	SendBuffer[SendLength] = ucFileNumber; SendLength++;
	int iResSize = 0;
	DFEncrypt(pReader, (PBYTE)&dwValue, &SendBuffer[SendLength], 4, &iResSize); SendLength += iResSize;

	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength,
		(m_ReaderSettings[pReader].CurrentDFSettings()->TransferType == CLSCRF_DF_TRANSFER_TYPE_PLAIN), true);

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_LimitedCredit(IN LPVOID pReader,
IN BYTE ucFileNumber,
IN DWORD dwValue)
{
	LONG Status = SCARD_S_SUCCESS;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	SendBuffer[SendLength] = 0x1C; SendLength++;
	SendBuffer[SendLength] = ucFileNumber; SendLength++;
	int iResSize = 0;
	DFEncrypt(pReader, (PBYTE)&dwValue, &SendBuffer[SendLength], 4, &iResSize); SendLength += iResSize;

	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength,
		(m_ReaderSettings[pReader].CurrentDFSettings()->TransferType == CLSCRF_DF_TRANSFER_TYPE_PLAIN), true);

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_WriteRecord(IN LPVOID pReader,
IN BYTE ucFileNumber,
IN DWORD dwOffset,
IN DWORD dwLength,
IN PBYTE pData)
{
	LONG Status = 0xAF;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[1024*16];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	BYTE tmpData[1024 * 16] = { 0 };
	int iResSize = 0;

	int iCurByte = 0;
	SendBuffer[SendLength] = 0x3B; SendLength++;
	SendBuffer[SendLength] = ucFileNumber; SendLength++;
	memcpy(&SendBuffer[SendLength], &dwOffset, 3); SendLength += 3;
	memcpy(&SendBuffer[SendLength], &dwLength, 3); SendLength += 3;

	DFEncrypt(pReader, pData, tmpData, dwLength, &iResSize, false, false, SendBuffer, SendLength);

	while (iResSize > 0)
	{
		if ((iResSize > 52) && (SendLength == 8))
		{
			memcpy(&SendBuffer[SendLength], &tmpData[iCurByte], 52); SendLength += 52;
			iCurByte += 52;
			iResSize -= 52;
		}
		else if ((iResSize > 59) && (SendLength == 1))
		{
			memcpy(&SendBuffer[SendLength], &tmpData[iCurByte], 52); SendLength += 59;
			iCurByte += 59;
			iResSize -= 59;
		}
		else
		{
			memcpy(&SendBuffer[SendLength], &tmpData[iCurByte], iResSize); SendLength += iResSize;
			iCurByte += iResSize;
			iResSize -= iResSize;
		}

		Status = Transmit_DESFire_Cmd(pReader,
			SendBuffer, SendLength,
			RecvBuffer, &RecvLength,
			(m_ReaderSettings[pReader].CurrentDFSettings()->TransferType == CLSCRF_DF_TRANSFER_TYPE_PLAIN), true);

		if (Status == 0xAF)
		{
			SendBuffer[0] = 0xAF; SendLength = 1;
		}
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_ReadRecords(IN LPVOID pReader,
IN BYTE ucFileNumber,
IN DWORD dwRecordOffset,
IN DWORD dwRecordsCount,
IN DWORD dwRecordSize,
OUT PBYTE pData,
OUT PDWORD pdwDataLength)
{
	LONG Status = 0xAF;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[1024*16];
	DWORD SendLength = 0;
	DWORD RecvLength = sizeof(RecvBuffer);

	//int iCurByte = 0;
	SendBuffer[SendLength] = 0xBB; SendLength++;
	SendBuffer[SendLength] = ucFileNumber; SendLength++;
	memcpy(&SendBuffer[SendLength], &dwRecordOffset, 3); SendLength += 3;
	memcpy(&SendBuffer[SendLength], &dwRecordsCount, 3); SendLength += 3;

	//BYTE tmpData[4096] = {0};

	//while (Status == 0xAF)
	//{
	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength,
		true, (m_ReaderSettings[pReader].CurrentDFSettings()->TransferType != CLSCRF_DF_TRANSFER_TYPE_ENCRYPTED));

	//	if ((Status == 0xAF) || (Status == 0x00))
	//	{
	//		memcpy(&tmpData[iCurByte], &RecvBuffer[0], RecvLength); iCurByte+=RecvLength;
	//
	//		SendBuffer[0] = 0xAF; SendLength = 1;
	//	}
	//}

	if (Status == SCARD_S_SUCCESS)
	{
		int iResSize = 0;
		//if (!DFDecrypt(pReader, tmpData, pData, iCurByte, dwRecordsCount*dwRecordSize, &iResSize))
		if (!DFDecrypt(pReader, RecvBuffer, pData, RecvLength, dwRecordsCount*dwRecordSize, &iResSize))
		{
			return SCARD_E_INVALID_VALUE;
		}

		if (pdwDataLength)
			*pdwDataLength = iResSize;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_ClearRecordFile(IN LPVOID pReader,
IN BYTE ucFileNumber)
{
	LONG Status = SCARD_S_SUCCESS;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	SendBuffer[SendLength] = 0xEB; SendLength++;
	SendBuffer[SendLength] = ucFileNumber; SendLength++;

	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength);

	if (Status == 0)
	{
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_CommitTransaction(IN LPVOID pReader)
{
	LONG Status = SCARD_S_SUCCESS;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	SendBuffer[SendLength] = 0xC7; SendLength++;

	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength);

	if (Status == 0)
	{
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_AbortTransaction(IN LPVOID pReader)
{
	LONG Status = SCARD_S_SUCCESS;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength = 0;
	DWORD RecvLength = BUF_SIZE;

	SendBuffer[SendLength] = 0xA7; SendLength++;

	Status = Transmit_DESFire_Cmd(pReader,
		SendBuffer, SendLength,
		RecvBuffer, &RecvLength);

	if (Status == 0)
	{
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_ClearAuthState(IN LPVOID pReader)
{
	m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType = (BYTE)-1;
	ZeroMemory(m_ReaderSettings[pReader].CurrentDFSettings()->IV, 24);
	return SCARD_S_SUCCESS;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareDesFire_Authenticate(IN LPVOID pReader,
IN BYTE ucAuthType,
IN PBYTE pKey,
IN BYTE ucKeyNumber)
{
	/*
	BYTE pInp[0x1D + 3] = {	0x3D, 0x03, 0x00, 0x00, 0x00, 0x15, 0x00, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08,
	0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15,    0x80, 0x00, 0x00};
	BYTE pKey1[16] =       {	0x0A, 0xAF, 0x80, 0x3D, 0xD2, 0xA6, 0x25, 0x2D, 0x85, 0x1B, 0x69, 0xB9, 0xE4, 0xF6, 0x38, 0x01};
	BYTE pOut[16] = {0};


	AES_128_CBC_En_MAC(
	pInp,
	2,
	pKey1,
	pOut );

	pOut[0] = pOut[0];
	*/
	/*
	BYTE pInp[64] = {0x6b, 0xc1, 0xbe, 0xe2, 0x2e, 0x40, 0x9f, 0x96, 0xe9, 0x3d, 0x7e, 0x11, 0x73, 0x93, 0x17, 0x2a,
	0xae, 0x2d, 0x8a, 0x57, 0x1e, 0x03, 0xac, 0x9c, 0x9e, 0xb7, 0x6f, 0xac, 0x45, 0xaf, 0x8e, 0x51,
	0x30, 0xc8, 0x1c, 0x46, 0xa3, 0x5c, 0xe4, 0x11, 0xe5, 0xfb, 0xc1, 0x19, 0x1a, 0x0a, 0x52, 0xef,
	0xf6, 0x9f, 0x24, 0x45, 0xdf, 0x4f, 0x9b, 0x17, 0xad, 0x2b, 0x41, 0x7b, 0xe6, 0x6c, 0x37, 0x10};
	BYTE pKey[16] = {0x2b, 0x7e, 0x15, 0x16, 0x28, 0xae, 0xd2, 0xa6, 0xab, 0xf7, 0x15, 0x88, 0x09, 0xcf, 0x4f, 0x3c};
	BYTE pOut[16] = {0};


	AES_128_CBC_En_MAC(
	pInp,
	4,
	pKey,
	pOut );

	pOut[0] = pOut[0];


	ZeroMemory(pOut, 16);
	AES_128_CBC_En_MAC(
	pInp,
	4,
	pKey,
	pOut );

	pOut[0] = pOut[0];
	*/
	/*
	BYTE pKey[16] = {0x72, 0xA8, 0x2A, 0xEF, 0x1A, 0x1E, 0xA4, 0xB8, 0x69, 0x5C, 0x26, 0x08, 0x22, 0xA2, 0xA8, 0xE5};
	BYTE pOut[16] = {0};

	//BYTE TI[4] = {AABBCC24};
	//BYTE IV[16] = {AABBCC24000001000000010000000100};
	//BYTE pData[16] = {02BB2A18BA7C4B54FC9F8CAE377F5C1A};
	//W_Ctr[2] = {0001};
	//BNr[2] = {0A00};
	//CC = A1
	BYTE pInp[32] = {
	0xA1,//CC
	0x01, 0x00,//W_Ctr
	0xAA, 0xBB, 0xCC, 0x24,//TI
	0x0A, 0x00,//BNr
	0x10, 0x3E, 0x4B, 0x9E, 0x73, 0x38, 0x4F, 0x8F, 0xB7, 0x9B, 0x02, 0xF0, 0x63, 0x1B, 0x4B, 0x45, //E(Kx,Data)
	0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	};

	AES_128_CBC_En_MAC(
	pInp,
	2,
	pKey,
	pOut );

	pOut[0] = pOut[0];


	*/
	/*
	BYTE pInp[16] = {0x02, 0xBB, 0x2A, 0x18, 0xBA, 0x7C, 0x4B, 0x54, 0xFC, 0x9F, 0x8C, 0xAE, 0x37, 0x7F, 0x5C, 0x1A};
	BYTE pIV[16] = {0xAA, 0xBB, 0xCC, 0x24, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x01, 0x00};
	BYTE pKey[16] = {0xD1, 0x3C, 0xDB, 0x09, 0xCC, 0xD2, 0xE4, 0x2C, 0xE0, 0x5E, 0xD7, 0xB8, 0xB6, 0xEB, 0xC6, 0x87};
	BYTE pOut[16] = {0};


	aes_128_encrypt_cbc(pInp, 16, pKey, pIV, pOut);
	//aes_128_encrypt16( pInp, pKey, pOut );

	pOut[0] = pOut[0];






	*/

	/*
	BYTE pOut[16] = {0};

	BYTE pKey[16] = {0x8D, 0xDF, 0xF1, 0x51, 0xA6, 0xEF, 0x6A, 0x7F, 0xE6, 0xD0, 0x33, 0x3A, 0x42, 0xBE, 0x21, 0xEE};
	BYTE pRndA[16] = {0xCF, 0xC1, 0x0C, 0x4F, 0x63, 0x05, 0x3E, 0x15, 0x08, 0x25, 0xC8, 0xC7, 0xE8, 0x05, 0xF3, 0x02};
	BYTE pRndB[16] = {0xB0, 0xE4, 0x0C, 0x79, 0x7C, 0x50, 0xE1, 0xE4, 0x8E, 0x88, 0xBE, 0xD0, 0x4C, 0x9F, 0x95, 0x79};

	BYTE pSessKmac[16] = {0};

	for (int i=0;i<5; i++)
	pSessKmac[10+i] = pRndA[i] ^ pRndB[i];

	memcpy(pSessKmac, pRndA+7, 5);
	memcpy(pSessKmac+5, pRndB+7, 5);

	pSessKmac[15] = 0x22;

	aes_128_encrypt16( pSessKmac, pKey, pOut );
	*/

	/*

	BYTE pOut[16] = {0};

	BYTE pKey[16] = {0x8D, 0xDF, 0xF1, 0x51, 0xA6, 0xEF, 0x6A, 0x7F, 0xE6, 0xD0, 0x33, 0x3A, 0x42, 0xBE, 0x21, 0xEE};
	BYTE pRndA[16] = {0xCF, 0xC1, 0x0C, 0x4F, 0x63, 0x05, 0x3E, 0x15, 0x08, 0x25, 0xC8, 0xC7, 0xE8, 0x05, 0xF3, 0x02};
	BYTE pRndB[16] = {0xB0, 0xE4, 0x0C, 0x79, 0x7C, 0x50, 0xE1, 0xE4, 0x8E, 0x88, 0xBE, 0xD0, 0x4C, 0x9F, 0x95, 0x79};

	BYTE pSessKmac[16] = {0};

	for (int i=0;i<5; i++)
	pSessKmac[10+i] = pRndA[4+i] ^ pRndB[4+i];

	memcpy(pSessKmac, pRndA+11, 5);
	memcpy(pSessKmac+5, pRndB+11, 5);

	pSessKmac[15] = 0x11;

	aes_128_encrypt16( pSessKmac, pKey, pOut );





	pOut[0] = pOut[0];

	*/
	// 16 5B D7 41 3F 59 E3 67 
	// 90 91 C9 C7 E9 72 94 FB  
	/*
	//BYTE pInp[32] = {0x90, 0x00, 0x00, 0x00, 0x5E, 0x5F, 0x4C, 0x7A, 0x36, 0x2C, 0x5D, 0xEE, 0x9A, 0x91, 0x49, 0xB7, 0xC0, 0xBB, 0x21, 0x70, 0xA1, 0xA2, 0xA3, 0xA4, 0xA5, 0xA6, 0xA7, 0xA8, 0xA9, 0xAA, 0xAB, 0xAC};
	BYTE pInp[32] = {0xA1, 0xA2, 0xA3, 0xA4, 0xA5, 0xA6, 0xA7, 0xA8, 0xA9, 0xAA, 0xAB, 0xAC, 0x90, 0x00, 0x00, 0x00, 0x5E, 0x5F, 0x4C, 0x7A, 0x36, 0x2C, 0x5D, 0xEE, 0x9A, 0x91, 0x49, 0xB7, 0xC0, 0xBB, 0x21, 0x70};
	BYTE pKey[16] = {0xA5, 0xA4, 0xA3, 0xA2, 0xA1, 0xA0, 0x18, 0x77, 0x8E, 0x69, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0, 0xF0};
	//BYTE pKey[16] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};
	//BYTE pKey[16] = {0};
	BYTE pOut[16] = {0};

	AES_128_CBC_En_MAC(
	pInp,
	2,
	pKey,
	pOut );

	pOut[0] = pOut[0];
	*/

	LONG Status = SCARD_S_SUCCESS;

	const int BUF_SIZE = 300;
	BYTE SendBuffer[BUF_SIZE];
	BYTE RecvBuffer[BUF_SIZE];
	DWORD SendLength;
	DWORD RecvLength = BUF_SIZE;

	BYTE RndA[16] = { 0 };
	//BYTE RndAs[16] = {0};
	BYTE RndB[16] = { 0 };
	//BYTE RndBs[16] = {0};
	BYTE SessionKey[24] = { 0 };
	BYTE SessionKeySize = 16;
	BYTE ucKeyVer = 0x00;
	int iKeyLength = 16;

	BYTE DESkey[24] = { 0 };
	//BYTE IV_Result[24] = {0};

	BYTE LAT = m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType;
	m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType = (BYTE)-1;
	ZeroMemory(m_ReaderSettings[pReader].CurrentDFSettings()->IV, 24);

	switch (ucAuthType) {
	case CLSCRF_DF_AUTH_TYPE_DES:
		iKeyLength = 16;

		SendBuffer[0] = 0x0A;
		SendBuffer[1] = ucKeyNumber;
		SendLength = 2;

		Status = Transmit_DESFire_Cmd(pReader,
			SendBuffer, SendLength,
			RecvBuffer, &RecvLength);

		if (Status == 0)
		{

			BYTE TmpBuffer[BUF_SIZE] = { 0 };

			memcpy(DESkey, pKey, 8);
			memcpy(DESkey + 8, pKey, 8);
			des_cbc_encrypt(
				DESkey,
				RecvBuffer,
				RecvBuffer,
				8);


			memcpy(&RndB[0], &RecvBuffer[0], 8);

			RecvBuffer[0 + 8] = RecvBuffer[0];

			memmove(
				RecvBuffer + 0,
				RecvBuffer + 1,
				8
				);

			for (int i = 0; i < 8; i++)
				RndA[i] = SendBuffer[i] = (BYTE)rand();

			memcpy(
				&SendBuffer[8],
				&RecvBuffer[0],
				8
				);

			des_cbc_encrypt(
				DESkey,
				SendBuffer,
				TmpBuffer,
				16);

			SendBuffer[0] = 0xAF;
			memcpy(&SendBuffer[1], TmpBuffer, 16);

			SendLength = 17;

			RecvLength = BUF_SIZE;

			Status = Transmit_DESFire_Cmd(pReader,
				SendBuffer, SendLength,
				RecvBuffer, &RecvLength);

			if (Status)
			{
				m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType = LAT;

				return Status;
			}

			des_cbc_encrypt(
				DESkey,
				RecvBuffer,
				RecvBuffer,
				8);

			RecvBuffer[0 + 8] = RecvBuffer[0 + 7];
			memmove(
				RecvBuffer + 1,
				RecvBuffer + 0,
				8
				);
			RecvBuffer[0] = RecvBuffer[0 + 8];

			if (memcmp(RndA, RecvBuffer, 8) != 0)
			{
				m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType = LAT;

				return SCARD_E_CARD_UNSUPPORTED;
			}

		}

		break;
	case CLSCRF_DF_AUTH_TYPE_3DES:
		iKeyLength = 16;

		SendBuffer[0] = 0x0A;
		SendBuffer[1] = ucKeyNumber;
		SendLength = 2;

		Status = Transmit_DESFire_Cmd(pReader,
			SendBuffer, SendLength,
			RecvBuffer, &RecvLength);

		if (Status == 0)
		{
			BYTE TmpBuffer[32] = { 0 };

			memcpy(DESkey, pKey, iKeyLength);

			/*
			//des_cblock key = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};
			//des_cblock key = {0xA7, 0xA6, 0xA5, 0xA4, 0xA3, 0xA2, 0xA1, 0xA0};
			//des_cblock key = {0xA0, 0xA1, 0xA2, 0xA3, 0xA4, 0xA5, 0xA6, 0xA7};
			des_cblock DESkey = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};


			des_key_schedule ks;
			des_set_key(&DESkey, ks);

			DES_LONG bl[2];

			memcpy(&RndB[0], &RecvBuffer[0], 8);

			memcpy(&bl, &RecvBuffer[0], 8);
			des_encrypt(
			bl,
			ks,
			DES_DECRYPT
			);
			memcpy(&RecvBuffer[0], &bl, 8);
			*/
			BYTE EncRndB[16] = { 0 };
			memcpy(&EncRndB[0], &RecvBuffer[0], RecvLength);

			des_3cbc_encipher_by_decrypt(
				DESkey,
				DESkey + 8,
				RecvBuffer,
				TmpBuffer,
				RecvLength
				);
			/*,
			NULL,
			NULL,
			m_DF_Settings[pReader].IV
			);*/
			memcpy(&RecvBuffer[0], TmpBuffer, RecvLength);

			memcpy(&RndB[0], &RecvBuffer[0], RecvLength);

			RecvBuffer[0 + RecvLength] = RecvBuffer[0];

			memmove(
				RecvBuffer + 0,
				RecvBuffer + 1,
				RecvLength
				);

			for (DWORD i = 0; i < RecvLength; i++)
				RndA[i] = SendBuffer[i + 1] = (BYTE)rand();

			memcpy(
				&SendBuffer[RecvLength + 1],
				&RecvBuffer[0],
				RecvLength
				);
			/*
			des_cblock cbl[2];
			des_cblock cblo[2];
			des_cblock IV = {0};
			memcpy(&cbl[0], SendBuffer, 8);
			memcpy(&cbl[1], SendBuffer+8, 8);

			des_3cbc_encrypt(
			cbl,
			cblo,
			2,
			ks,
			ks,
			&IV,
			&IV,
			DES_ENCRYPT
			);

			memcpy(SendBuffer+1, &cblo[0], 8);
			memcpy(SendBuffer+1+8, &cblo[1], 8);
			*/

			des_3cbc_encipher_by_decrypt(
				DESkey,
				DESkey + 8,
				SendBuffer + 1,
				TmpBuffer,
				RecvLength * 2
				);
			/*,
			m_DF_Settings[pReader].IV,
			m_DF_Settings[pReader].IV,
			m_DF_Settings[pReader].IV
			);*/
			memcpy(&SendBuffer[1], TmpBuffer, RecvLength * 2);

			SendBuffer[0] = 0xAF;

			SendLength = RecvLength * 2 + 1;

			RecvLength = BUF_SIZE;

			ZeroMemory(RecvBuffer, 8);

			Status = Transmit_DESFire_Cmd(pReader,
				SendBuffer, SendLength,
				RecvBuffer, &RecvLength);

			if (Status)
			{
				m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType = LAT;

				return Status;
			}
			/*
			memcpy(&IV, &cblo[1], 8);

			memcpy(&cbl[0], &RecvBuffer[0], 8);
			des_cbc_encrypt(
			cbl,
			cblo,
			1,
			ks,
			&IV,
			DES_DECRYPT
			);
			memcpy(&RecvBuffer[0], &cblo[0], 8);
			*/
			/*
			memcpy(&bl, &RecvBuffer[0], 8);

			des_encrypt(
			bl,
			ks,
			DES_DECRYPT
			);

			memcpy(&RecvBuffer[0], &bl, 8);
			*/
			/*des_decrypt3(
			bl,
			ks,
			ks,
			ks
			);*/
			des_3cbc_encipher_by_decrypt(
				DESkey,
				DESkey + 8,
				RecvBuffer,
				TmpBuffer,
				RecvLength
				);
			/*,
			m_DF_Settings[pReader].IV,
			m_DF_Settings[pReader].IV,
			m_DF_Settings[pReader].IV
			);*/
			memcpy(&RecvBuffer[0], TmpBuffer, RecvLength);


			RecvBuffer[0 + RecvLength] = RecvBuffer[0 + RecvLength - 1];
			memmove(
				RecvBuffer + 1,
				RecvBuffer + 0,
				RecvLength
				);
			RecvBuffer[0] = RecvBuffer[0 + RecvLength];

			if (memcmp(RndA, RecvBuffer, RecvLength) != 0)
			{
				m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType = LAT;

				return SCARD_E_CARD_UNSUPPORTED;
			}

		}

		break;
	case CLSCRF_DF_AUTH_TYPE_3K3DES:
		iKeyLength = 24;

		SendBuffer[0] = 0x1A;
		SendBuffer[1] = ucKeyNumber;
		SendLength = 2;

		Status = Transmit_DESFire_Cmd(pReader,
			SendBuffer, SendLength,
			RecvBuffer, &RecvLength);

		if (Status == 0)
		{
			BYTE TmpBuffer[32] = { 0 };

			memcpy(DESkey, pKey, iKeyLength);

			BYTE EncRndB[16] = { 0 };
			memcpy(&EncRndB[0], &RecvBuffer[0], RecvLength);

			des_3k3cbc_decrypt(
				DESkey,
				DESkey + 8,
				DESkey + 16,
				RecvBuffer,
				TmpBuffer,
				RecvLength,
				NULL,
				NULL,
				m_ReaderSettings[pReader].CurrentDFSettings()->IV
				);
			memcpy(&RecvBuffer[0], TmpBuffer, RecvLength);

			memcpy(&RndB[0], &RecvBuffer[0], RecvLength);

			RecvBuffer[0 + RecvLength] = RecvBuffer[0];

			memmove(
				RecvBuffer + 0,
				RecvBuffer + 1,
				RecvLength
				);

			for (DWORD i = 0; i < RecvLength; i++)
				RndA[i] = SendBuffer[i + 1] = (BYTE)rand();

			memcpy(
				&SendBuffer[RecvLength + 1],
				&RecvBuffer[0],
				RecvLength
				);

			des_3k3cbc_encrypt(
				DESkey,
				DESkey + 8,
				DESkey + 16,
				SendBuffer + 1,
				TmpBuffer,
				RecvLength * 2,
				m_ReaderSettings[pReader].CurrentDFSettings()->IV,
				m_ReaderSettings[pReader].CurrentDFSettings()->IV,
				m_ReaderSettings[pReader].CurrentDFSettings()->IV
				);
			memcpy(&SendBuffer[1], TmpBuffer, RecvLength * 2);

			SendBuffer[0] = 0xAF;

			SendLength = RecvLength * 2 + 1;

			RecvLength = BUF_SIZE;

			ZeroMemory(RecvBuffer, 8);

			Status = Transmit_DESFire_Cmd(pReader,
				SendBuffer, SendLength,
				RecvBuffer, &RecvLength);

			if (Status)
			{
				m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType = LAT;

				return Status;
			}

			des_3k3cbc_decrypt(
				DESkey,
				DESkey + 8,
				DESkey + 16,
				RecvBuffer,
				TmpBuffer,
				RecvLength,
				m_ReaderSettings[pReader].CurrentDFSettings()->IV,
				m_ReaderSettings[pReader].CurrentDFSettings()->IV,
				m_ReaderSettings[pReader].CurrentDFSettings()->IV
				);
			memcpy(&RecvBuffer[0], TmpBuffer, RecvLength);


			RecvBuffer[0 + RecvLength] = RecvBuffer[0 + RecvLength - 1];
			memmove(
				RecvBuffer + 1,
				RecvBuffer + 0,
				RecvLength
				);
			RecvBuffer[0] = RecvBuffer[0 + RecvLength];

			if (memcmp(RndA, RecvBuffer, RecvLength) != 0)
			{
				m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType = LAT;

				return SCARD_E_CARD_UNSUPPORTED;
			}

		}

		break;
	case CLSCRF_DF_AUTH_TYPE_AES:
		iKeyLength = 16;

		SendBuffer[0] = 0xAA;
		SendBuffer[1] = ucKeyNumber;
		SendLength = 2;

		Status = Transmit_DESFire_Cmd(pReader,
			SendBuffer, SendLength,
			RecvBuffer, &RecvLength);

		if ((Status == 0) && RecvLength)
		{

			BYTE pOut[33] = { 0 };
			aes_128_decrypt_cbc(
				RecvBuffer,
				16,
				pKey,
				m_ReaderSettings[pReader].CurrentDFSettings()->IV,
				&pOut[16],
				m_ReaderSettings[pReader].CurrentDFSettings()->IV
				);

			memcpy(RndB, &pOut[16], 16);

			pOut[16 + 16] = pOut[16];

			memmove(
				pOut + 16,
				pOut + 17,
				16
				);

			//memcpy(&RndBs[0], &pOut[16], 16);

			for (int i = 0; i < 16; i++)
				RndA[i] = pOut[i] = (BYTE)rand();

			aes_128_encrypt_cbc(
				pOut,
				32,
				pKey,
				m_ReaderSettings[pReader].CurrentDFSettings()->IV,
				&SendBuffer[1],
				m_ReaderSettings[pReader].CurrentDFSettings()->IV
				);

			SendBuffer[0] = 0xAF;

			SendLength = 33;

			RecvLength = BUF_SIZE;

			Status = Transmit_DESFire_Cmd(pReader,
				SendBuffer, SendLength,
				RecvBuffer, &RecvLength);

			if (Status)
			{
				m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType = LAT;

				return Status;
			}

			aes_128_decrypt_cbc(
				RecvBuffer,
				16,
				pKey,
				m_ReaderSettings[pReader].CurrentDFSettings()->IV,
				pOut,
				m_ReaderSettings[pReader].CurrentDFSettings()->IV
				);

			//memcpy(RndAs, pOut, 16);

			pOut[0 + 16] = pOut[0 + 16 - 1];
			memmove(
				pOut + 1,
				pOut + 0,
				16
				);
			pOut[0] = pOut[0 + 16];

			if (memcmp(RndA, pOut, 16) != 0)
			{
				m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType = LAT;

				return SCARD_E_CARD_UNSUPPORTED;
			}

		}

		break;
	default:
		break;
	}


	if (Status == SCARD_S_SUCCESS)
	{
		switch (ucAuthType) {
		case CLSCRF_DF_AUTH_TYPE_DES:
			//des
			memcpy(&SessionKey[0], &RndA[0], 4);
			memcpy(&SessionKey[4], &RndB[0], 4);
			SessionKeySize = 8;

			for (int i = 0; i < 8; i++)
				ucKeyVer |= ((pKey[7 - i] & 0x01) << i);
			break;
		case CLSCRF_DF_AUTH_TYPE_3DES:
			//3des
			memcpy(&SessionKey[0], &RndA[0], 4);
			memcpy(&SessionKey[4], &RndB[0], 4);
			memcpy(&SessionKey[8], &RndA[4], 4);
			memcpy(&SessionKey[12], &RndB[4], 4);
			SessionKeySize = 16;

			for (int i = 0; i < 8; i++)
				ucKeyVer |= ((pKey[7 - i] & 0x01) << i);
			break;
		case CLSCRF_DF_AUTH_TYPE_3K3DES:
			//3k3des
			memcpy(&SessionKey[0], &RndA[0], 4);
			memcpy(&SessionKey[4], &RndB[0], 4);
			memcpy(&SessionKey[8], &RndA[6], 4);
			memcpy(&SessionKey[12], &RndB[6], 4);
			memcpy(&SessionKey[16], &RndA[12], 4);
			memcpy(&SessionKey[20], &RndB[12], 4);
			SessionKeySize = 24;

			for (int i = 0; i < 8; i++)
				ucKeyVer |= ((pKey[7 - i] & 0x01) << i);
			break;
		case CLSCRF_DF_AUTH_TYPE_AES:
			//aes
			memcpy(&SessionKey[0], &RndA[0], 4);
			memcpy(&SessionKey[4], &RndB[0], 4);
			memcpy(&SessionKey[8], &RndA[12], 4);
			memcpy(&SessionKey[12], &RndB[12], 4);
			SessionKeySize = 16;
			break;
		default:
			break;
		}

		ZeroMemory(m_ReaderSettings[pReader].CurrentDFSettings()->IV, 16);

		m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType = ucAuthType;

		memcpy(&m_ReaderSettings[pReader].CurrentDFSettings()->SessionKey, SessionKey, 24);
		m_ReaderSettings[pReader].CurrentDFSettings()->SessionKeySize = SessionKeySize;
		m_ReaderSettings[pReader].CurrentDFSettings()->KeyVersion = ucKeyVer;

		DFSubkeys(pReader);//, SessionKey, SessionKeySize, m_DF_Settings[pReader].K1, m_DF_Settings[pReader].K2);
	}
	else
	{
		m_ReaderSettings[pReader].CurrentDFSettings()->LastAuthType = LAT;
	}

	return Status;
}

/*
CLSCRF_API LONG CLSCRF_DECL
CLSCRF_DESFire_Authenticate( IN LPVOID pReader,
IN DWORD  dwKeyNumber )
{
LONG Status = SCARD_S_SUCCESS;

if( pReader )
{
CTransfer* pTr = (CTransfer*)pReader;
pTr->m_TrBuf.Clear();
pTr->m_TrBuf.AddByte( 0x6c );
pTr->m_TrBuf.AddByte( BYTE(dwKeyNumber & 0xff) );

Status = pTr->SingleTransceive();
}
else
{
Status = SCARD_E_READER_UNAVAILABLE;
}

return Status;
}
*/
