#include "StdAfx.h"

#include "Clscrfl.h"

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareUltralightEV1_GetVersion(IN     LPVOID pReader,
OUT LPBYTE pbVersion)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader == NULL)
		return SCARD_E_INVALID_PARAMETER;

	CTransfer* pTr = (CTransfer*)pReader;
	pTr->m_TrBuf.Clear();
	pTr->m_TrBuf.AddByte(0xC4);
	pTr->m_TrBuf.AddByte(0x60);

	Status = pTr->SingleTransceive();

	if (Status != SCARD_S_SUCCESS)
		return Status;

	DWORD Size = 8;
	if (!pTr->m_RcBuf.GetBulk(1, pbVersion, &Size))
		return SCARD_E_COMM_DATA_LOST;

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareUltralightEV1_Read(IN     LPVOID pReader,
IN     BYTE   ucPageAddress,
OUT LPBYTE pData)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader == NULL)
		return SCARD_E_INVALID_PARAMETER;

	if (pData == NULL)
		return SCARD_E_INVALID_PARAMETER;

	CTransfer* pTr = (CTransfer*)pReader;
	pTr->m_TrBuf.Clear();
	pTr->m_TrBuf.AddByte(0xC4);
	pTr->m_TrBuf.AddByte(0x30);
	pTr->m_TrBuf.AddByte(ucPageAddress);

	Status = pTr->SingleTransceive();

	if (Status != SCARD_S_SUCCESS)
		return Status;

	DWORD Size = 16;
	if (!pTr->m_RcBuf.GetBulk(1, pData, &Size))
		return SCARD_E_COMM_DATA_LOST;

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareUltralightEV1_ReadFast(IN     LPVOID pReader,
IN     BYTE   ucPageStart,
IN     BYTE   ucPageEnd,
IN OUT	PDWORD	pdwSizeRead,
OUT LPBYTE pData)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader == NULL)
		return SCARD_E_INVALID_PARAMETER;

	if (pdwSizeRead == NULL)
		return SCARD_E_INVALID_PARAMETER;

	if (pData == NULL)
		return SCARD_E_INVALID_PARAMETER;

	if (4 * (ucPageEnd - ucPageStart + 1) > *pdwSizeRead)
		return SCARD_E_INSUFFICIENT_BUFFER;

	CTransfer* pTr = (CTransfer*)pReader;
	pTr->m_TrBuf.Clear();
	pTr->m_TrBuf.AddByte(0xC4);
	pTr->m_TrBuf.AddByte(0x3A);
	pTr->m_TrBuf.AddByte(ucPageStart);
	pTr->m_TrBuf.AddByte(ucPageEnd);

	Status = pTr->SingleTransceive();

	if (Status != SCARD_S_SUCCESS)
		return Status;

	DWORD Size = 2;
	DWORD dwSize = 0x00000000;

	if (!pTr->m_RcBuf.GetBulk(1, (PBYTE)&dwSize, &Size))
		return SCARD_E_COMM_DATA_LOST;

	if (dwSize > *pdwSizeRead)
		return SCARD_E_INSUFFICIENT_BUFFER;

	*pdwSizeRead = dwSize;

	Size = *pdwSizeRead;

	if (!pTr->m_RcBuf.GetBulk(3, pData, &Size))
		return SCARD_E_COMM_DATA_LOST;

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareUltralightEV1_Write(IN LPVOID pReader,
IN BYTE   ucPageAddress,
IN LPBYTE pData)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader == NULL)
		return SCARD_E_INVALID_PARAMETER;

	if (pData == NULL)
		return SCARD_E_INVALID_PARAMETER;

	CTransfer* pTr = (CTransfer*)pReader;
	pTr->m_TrBuf.Clear();
	pTr->m_TrBuf.AddByte(0xC4);
	pTr->m_TrBuf.AddByte(0xA2);
	pTr->m_TrBuf.AddByte(ucPageAddress);
	pTr->m_TrBuf.AddBulk(pData, 4);

	Status = pTr->SingleTransceive();

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareUltralightEV1_ReadCounter(IN     LPVOID pReader,
IN     BYTE   ucCounterNumber,
OUT PDWORD pdwCounterValue)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader == NULL)
		return SCARD_E_INVALID_PARAMETER;

	if (pdwCounterValue == NULL)
		return SCARD_E_INVALID_PARAMETER;

	CTransfer* pTr = (CTransfer*)pReader;
	pTr->m_TrBuf.Clear();
	pTr->m_TrBuf.AddByte(0xC4);
	pTr->m_TrBuf.AddByte(0x39);
	pTr->m_TrBuf.AddByte(ucCounterNumber);

	Status = pTr->SingleTransceive();

	if (Status != SCARD_S_SUCCESS)
		return Status;

	DWORD Size = 3;
	*pdwCounterValue = 0x00000000;
	if (!pTr->m_RcBuf.GetBulk(1, (PBYTE)pdwCounterValue, &Size))
		return SCARD_E_COMM_DATA_LOST;

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareUltralightEV1_IncrementCounter(IN LPVOID pReader,
IN BYTE   ucCounterNumber,
IN DWORD dwCounterValue)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader == NULL)
		return SCARD_E_INVALID_PARAMETER;

	CTransfer* pTr = (CTransfer*)pReader;
	pTr->m_TrBuf.Clear();
	pTr->m_TrBuf.AddByte(0xC4);
	pTr->m_TrBuf.AddByte(0xA5);
	pTr->m_TrBuf.AddByte(ucCounterNumber);
	pTr->m_TrBuf.AddBulk((PBYTE)&dwCounterValue, 3);

	Status = pTr->SingleTransceive();

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareUltralightEV1_AuthenticatePassword(IN LPVOID pReader,
IN LPBYTE pPassword,
OUT LPBYTE pPasswordAcknowledge)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader == NULL)
		return SCARD_E_INVALID_PARAMETER;

	if (pPassword == NULL)
		return SCARD_E_INVALID_PARAMETER;

	if (pPasswordAcknowledge == NULL)
		return SCARD_E_INVALID_PARAMETER;

	CTransfer* pTr = (CTransfer*)pReader;
	pTr->m_TrBuf.Clear();
	pTr->m_TrBuf.AddByte(0xC4);
	pTr->m_TrBuf.AddByte(0x1B);
	pTr->m_TrBuf.AddBulk(pPassword, 4);

	Status = pTr->SingleTransceive();

	if (Status != SCARD_S_SUCCESS)
		return Status;

	DWORD Size = 2;

	if (!pTr->m_RcBuf.GetBulk(1, pPasswordAcknowledge, &Size))
		return SCARD_E_UNEXPECTED;

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareUltralightEV1_ReadSignature(IN     LPVOID pReader,
IN     BYTE   ucAddr,
OUT LPBYTE pSignature)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader == NULL)
		return SCARD_E_INVALID_PARAMETER;

	if (pSignature == NULL)
		return SCARD_E_INVALID_PARAMETER;

	ucAddr = 0x00;

	CTransfer* pTr = (CTransfer*)pReader;
	pTr->m_TrBuf.Clear();
	pTr->m_TrBuf.AddByte(0xC4);
	pTr->m_TrBuf.AddByte(0x3C);
	pTr->m_TrBuf.AddByte(ucAddr);

	Status = pTr->SingleTransceive();

	if (Status != SCARD_S_SUCCESS)
		return Status;

	DWORD Size = 32;
	if (!pTr->m_RcBuf.GetBulk(1, pSignature, &Size))
		return SCARD_E_COMM_DATA_LOST;

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareUltralightEV1_CheckTearingEvent(IN     LPVOID pReader,
IN     BYTE   ucCounterNumber,
OUT    PBYTE  pucValidFlagByte)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader == NULL)
		return SCARD_E_INVALID_PARAMETER;

	if (pucValidFlagByte == NULL)
		return SCARD_E_INVALID_PARAMETER;

	CTransfer* pTr = (CTransfer*)pReader;
	pTr->m_TrBuf.Clear();
	pTr->m_TrBuf.AddByte(0xC4);
	pTr->m_TrBuf.AddByte(0x3E);
	pTr->m_TrBuf.AddByte(ucCounterNumber);

	Status = pTr->SingleTransceive();

	if (Status != SCARD_S_SUCCESS)
		return Status;

	DWORD Size = 1;
	*pucValidFlagByte = 0x00;
	if (!pTr->m_RcBuf.GetBulk(1, pucValidFlagByte, &Size))
		return SCARD_E_COMM_DATA_LOST;

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_MifareUltralightEV1_VirtualCardSupportLast(IN LPVOID pReader,
IN LPBYTE pIID,
IN LPBYTE pPCDCAPS,
OUT LPBYTE pucVirtualCardTypeIdentifier)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader == NULL)
		return SCARD_E_INVALID_PARAMETER;

	if (pIID == NULL)
		return SCARD_E_INVALID_PARAMETER;

	if (pPCDCAPS == NULL)
		return SCARD_E_INVALID_PARAMETER;

	if (pucVirtualCardTypeIdentifier == NULL)
		return SCARD_E_INVALID_PARAMETER;

	CTransfer* pTr = (CTransfer*)pReader;
	pTr->m_TrBuf.Clear();
	pTr->m_TrBuf.AddByte(0xC4);
	pTr->m_TrBuf.AddByte(0x4B);
	pTr->m_TrBuf.AddBulk(pIID, 16);
	pTr->m_TrBuf.AddBulk(pPCDCAPS, 4);

	Status = pTr->SingleTransceive();

	if (Status != SCARD_S_SUCCESS)
		return Status;

	DWORD Size = 1;

	if (!pTr->m_RcBuf.GetBulk(1, pucVirtualCardTypeIdentifier, &Size))
		return SCARD_E_UNEXPECTED;

	return Status;
}
