#include "StdAfx.h"

#include "Clscrfl.h"

/** \defgroup phpalI18092mPI ISO/IEC 18092 Mode Passive Initiator
* \brief These Components implement the ISO/IEC ISO18092 contactless protocol. Supported mode is passive initiator, i.e.
* using these components it is possible to communicate with a passive target in peer to peer mode.
* @{
*/
#define PHPAL_I18092MPI_DID_MAX          0x0EU  /**< The last valid DID. */
#define PHPAL_I18092MPI_WT_MAX           0x0EU  /**< The maximum allowed WT value, NFCForum-TS-DigitalProtocol-1.0 section 14.11. */
#define PHPAL_I18092MPI_FSL_MAX          0x03U  /**< The maximum allowed FSL value. */
#define PHPAL_I18092MPI_PNI_MAX          0x03U  /**< The maximum allowed PNI value. */

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC663_ActivateCard(IN  LPVOID  pReader,
IN	LPBYTE	Nfcid3i,
IN	BYTE	Did,
IN	BYTE	NadEnable,
IN	BYTE	Nad,
IN	BYTE	Dsi,
IN	BYTE	Dri,
IN	BYTE	Fsl,
IN	BYTE	GiLength,
IN	LPBYTE	Gi,
OUT LPBYTE  pbRecvBuffer,
IN OUT LPDWORD pdwRecvLength)
{
	LONG Status = SCARD_S_SUCCESS;
	DWORD BufSize = *pdwRecvLength;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();

		pTr->m_TrBuf.AddByte(0x11);
		pTr->m_TrBuf.AddBulk(Nfcid3i, 10);
		pTr->m_TrBuf.AddByte(Did);
		pTr->m_TrBuf.AddByte(NadEnable);
		pTr->m_TrBuf.AddByte(Nad);
		pTr->m_TrBuf.AddByte(Dsi);
		pTr->m_TrBuf.AddByte(Dri);
		pTr->m_TrBuf.AddByte(Fsl);
		pTr->m_TrBuf.AddByte(GiLength);
		pTr->m_TrBuf.AddBulk(Gi, GiLength);

		Status = pTr->SingleTransceive();

		DWORD Size = pTr->m_RcBuf.GetLength();

		if (Size < 2)  // Length is absent
		{
			*pdwRecvLength = 0;
			if (Status == SCARD_S_SUCCESS)
				Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
		}
		else
		{
			Size = 2;
			*pdwRecvLength = 0;
			if (!pTr->m_RcBuf.GetBulk(1, (LPBYTE)pdwRecvLength, &Size))
			{
				*pdwRecvLength = 0;
				if (Status == SCARD_S_SUCCESS)
					Status = SCARD_E_COMM_DATA_LOST;
			}

			if (*pdwRecvLength > BufSize)
			{
				*pdwRecvLength = BufSize;
				if (Status == SCARD_S_SUCCESS)
					Status = SCARD_E_INSUFFICIENT_BUFFER;
			}

			if ((Size = *pdwRecvLength) > 0)
				if (!pTr->m_RcBuf.GetBulk(3, pbRecvBuffer, pdwRecvLength))
				{
					if (Status == SCARD_S_SUCCESS)
						Status = SCARD_E_COMM_DATA_LOST;
				}
		}
	}
	else
	{
		*pdwRecvLength = 0;
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC663_Deselect(IN  LPVOID  pReader,
IN	BYTE	DeselectCommand)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();

		pTr->m_TrBuf.AddByte(0x12);
		pTr->m_TrBuf.AddByte(DeselectCommand);

		Status = pTr->SingleTransceive();
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC663_Exchange(IN  LPVOID  pReader,
IN	DWORD	dwOption,
IN	LPCBYTE pbSendBuffer,
IN	DWORD   dwSendLength,
OUT LPBYTE  pbRecvBuffer,
IN OUT LPDWORD pdwRecvLength)
{
	LONG Status = SCARD_S_SUCCESS;
	DWORD BufSize = *pdwRecvLength;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();

		pTr->m_TrBuf.AddByte(0x13);
		pTr->m_TrBuf.AddBulk((PBYTE)&dwOption, 2);
		pTr->m_TrBuf.AddBulk((PBYTE)&dwSendLength, 2);
		pTr->m_TrBuf.AddBulk(pbSendBuffer, dwSendLength);

		Status = pTr->SingleTransceive();

		DWORD Size = pTr->m_RcBuf.GetLength();

		if (Size < 2)  // Length is absent
		{
			*pdwRecvLength = 0;
			if (Status == SCARD_S_SUCCESS)
				Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
		}
		else
		{
			Size = 2;
			*pdwRecvLength = 0;
			if (!pTr->m_RcBuf.GetBulk(1, (LPBYTE)pdwRecvLength, &Size))
			{
				*pdwRecvLength = 0;
				if (Status == SCARD_S_SUCCESS)
					Status = SCARD_E_COMM_DATA_LOST;
			}

			if (*pdwRecvLength > BufSize)
			{
				*pdwRecvLength = BufSize;
				if (Status == SCARD_S_SUCCESS)
					Status = SCARD_E_INSUFFICIENT_BUFFER;
			}

			if ((Size = *pdwRecvLength) > 0)
				if (!pTr->m_RcBuf.GetBulk(3, pbRecvBuffer, pdwRecvLength))
				{
					if (Status == SCARD_S_SUCCESS)
						Status = SCARD_E_COMM_DATA_LOST;
				}
		}
	}
	else
	{
		*pdwRecvLength = 0;
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC663_ResetProtocol(IN  LPVOID  pReader)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();

		pTr->m_TrBuf.AddByte(0x40);

		Status = pTr->SingleTransceive();
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC663_AttributeRequest(IN  LPVOID  pReader,
IN	LPBYTE	Nfcid3i,
IN	BYTE	Did,
IN	BYTE	NadEnable,
IN	BYTE	Nad,
IN	BYTE	Fsl,
IN	BYTE	GiLength,
IN	LPBYTE	Gi,
OUT LPBYTE  pbRecvBuffer,
IN OUT LPDWORD pdwRecvLength)
{
	LONG Status = SCARD_S_SUCCESS;
	DWORD BufSize = *pdwRecvLength;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();

		pTr->m_TrBuf.AddByte(0x33);
		pTr->m_TrBuf.AddBulk(Nfcid3i, 10);
		pTr->m_TrBuf.AddByte(Did);
		pTr->m_TrBuf.AddByte(NadEnable);
		pTr->m_TrBuf.AddByte(Nad);
		pTr->m_TrBuf.AddByte(Fsl);
		pTr->m_TrBuf.AddByte(GiLength);
		pTr->m_TrBuf.AddBulk(Gi, GiLength);

		Status = pTr->SingleTransceive();

		DWORD Size = pTr->m_RcBuf.GetLength();

		if (Size < 2)  // Length is absent
		{
			*pdwRecvLength = 0;
			if (Status == SCARD_S_SUCCESS)
				Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
		}
		else
		{
			Size = 2;
			*pdwRecvLength = 0;
			if (!pTr->m_RcBuf.GetBulk(1, (LPBYTE)pdwRecvLength, &Size))
			{
				*pdwRecvLength = 0;
				if (Status == SCARD_S_SUCCESS)
					Status = SCARD_E_COMM_DATA_LOST;
			}

			if (*pdwRecvLength > BufSize)
			{
				*pdwRecvLength = BufSize;
				if (Status == SCARD_S_SUCCESS)
					Status = SCARD_E_INSUFFICIENT_BUFFER;
			}

			if ((Size = *pdwRecvLength) > 0)
				if (!pTr->m_RcBuf.GetBulk(3, pbRecvBuffer, pdwRecvLength))
				{
					if (Status == SCARD_S_SUCCESS)
						Status = SCARD_E_COMM_DATA_LOST;
				}
		}
	}
	else
	{
		*pdwRecvLength = 0;
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC663_ParameterSelect(IN  LPVOID  pReader,
IN	BYTE	Dsi,
IN	BYTE	Dri,
IN	BYTE	Fsl)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();

		pTr->m_TrBuf.AddByte(0x34);
		pTr->m_TrBuf.AddByte(Dsi);
		pTr->m_TrBuf.AddByte(Dri);
		pTr->m_TrBuf.AddByte(Fsl);

		Status = pTr->SingleTransceive();
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC663_PresenceCheck(IN  LPVOID  pReader)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();

		pTr->m_TrBuf.AddByte(0x1f);

		Status = pTr->SingleTransceive();
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}


CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC663_SetConfig(IN  LPVOID  pReader,
IN	DWORD	dwParameterNumber,
IN  DWORD   pdwParameterValue)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();

		pTr->m_TrBuf.AddByte(0x41);
		pTr->m_TrBuf.AddBulk((PBYTE)&dwParameterNumber, 2);
		pTr->m_TrBuf.AddBulk((PBYTE)&pdwParameterValue, 2);

		Status = pTr->SingleTransceive();
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC663_GetConfig(IN  LPVOID  pReader,
IN	DWORD	dwParameterNumber,
OUT LPDWORD  pdwParameterValue)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();

		pTr->m_TrBuf.AddByte(0x42);
		pTr->m_TrBuf.AddBulk((PBYTE)&dwParameterNumber, 2);

		Status = pTr->SingleTransceive();

		DWORD Size = pTr->m_RcBuf.GetLength();

		if (Size < 2)  // Length is absent
		{
			if (Status == SCARD_S_SUCCESS)
				Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
		}
		else
		{
			Size = 2;
			if (!pTr->m_RcBuf.GetBulk(1, (LPBYTE)pdwParameterValue, &Size))
			{
				if (Status == SCARD_S_SUCCESS)
					Status = SCARD_E_COMM_DATA_LOST;
			}
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC663_GetSerialNo(IN  LPVOID  pReader,
OUT LPBYTE  NFCID3)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();

		pTr->m_TrBuf.AddByte(0x28);

		Status = pTr->SingleTransceive();

		DWORD Size = pTr->m_RcBuf.GetLength();

		if (Size < CLSCRF_PHPAL_I18092MPI_NFCID3_LENGTH)  // Length is absent
		{
			if (Status == SCARD_S_SUCCESS)
				Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
		}
		else
		{
			Size = CLSCRF_PHPAL_I18092MPI_NFCID3_LENGTH;
			if (!pTr->m_RcBuf.GetBulk(1, NFCID3, &Size))
			{
				if (Status == SCARD_S_SUCCESS)
					Status = SCARD_E_COMM_DATA_LOST;
			}
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC663_E2_Read(IN  LPVOID  pReader,
IN  DWORD dwAddr,
IN	BYTE ucLength,
OUT LPBYTE  pData)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();

		pTr->m_TrBuf.AddByte(0x23);
		pTr->m_TrBuf.AddBulk((PBYTE)&dwAddr, 2);
		pTr->m_TrBuf.AddByte(ucLength);

		Status = pTr->SingleTransceive();

		DWORD Size = pTr->m_RcBuf.GetLength();

		if (Size < 2)  // Length is absent
		{
			if (Status == SCARD_S_SUCCESS)
				Status = SCARD_E_COMM_DATA_LOST;    // 0x8010002F
		}
		else
		{
			if (--Size > ucLength)
			{
				if (Status == SCARD_S_SUCCESS)
					Status = SCARD_E_INSUFFICIENT_BUFFER;
			}
			else if (Size > 0)
				if (!pTr->m_RcBuf.GetBulk(1, pData, &Size))
				{
					if (Status == SCARD_S_SUCCESS)
						Status = SCARD_E_COMM_DATA_LOST;
				}
		}
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC663_E2_Write(IN  LPVOID  pReader,
IN  DWORD dwAddr,
IN	BYTE ucData)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();

		pTr->m_TrBuf.AddByte(0x24);
		pTr->m_TrBuf.AddBulk((PBYTE)&dwAddr, 2);
		pTr->m_TrBuf.AddByte(ucData);

		Status = pTr->SingleTransceive();
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}

CLSCRF_API LONG CLSCRF_DECL
CLSCRF_NFC663_E2_WritePage(IN  LPVOID  pReader,
IN  DWORD dwAddr,
IN	BYTE ucLength,
IN  LPBYTE pData)
{
	LONG Status = SCARD_S_SUCCESS;

	if (pReader)
	{
		CTransfer* pTr = (CTransfer*)pReader;
		pTr->m_TrBuf.Clear();

		pTr->m_TrBuf.AddByte(0x2A);
		pTr->m_TrBuf.AddBulk((PBYTE)&dwAddr, 2);
		pTr->m_TrBuf.AddByte(ucLength);
		pTr->m_TrBuf.AddBulk(pData, ucLength);

		Status = pTr->SingleTransceive();
	}
	else
	{
		Status = SCARD_E_READER_UNAVAILABLE;
	}

	return Status;
}
