#include "../StdAfx.h"

#include <vector>

#include "Tlv.h"



Tlv::Tlv()
	: m_type(0)
{
	//m_value = new CByteArray();
	m_value = new std::vector<BYTE>();
}

//Tlv::Tlv(BYTE type, CByteArray* value)
Tlv::Tlv(BYTE type, std::vector<BYTE>* value)
	: m_type(type)
{
	//m_value = new CByteArray();
	m_value = new std::vector<BYTE>();
	if (value)
		//m_value->Copy(*value);
		//m_value->insert(m_value->end(), (*value).begin(), (*value).end());
		m_value = new std::vector<BYTE>(*value);
}

Tlv::~Tlv()
{
	delete m_value;
}

BYTE Tlv::type() const
{
	return m_type;
}

UINT64 Tlv::length() const
{
	switch (m_type)
	{
	case Tlv::Null:
	case Tlv::Terminator:
		return 0;

	default:
		return m_value->size();
	}
}

//CByteArray* Tlv::value()
std::vector<BYTE>* Tlv::value()
{
	switch (m_type)
	{
	case Tlv::Null:
	case Tlv::Terminator:
		//return new CByteArray();
		return new std::vector<BYTE>();

	default:
		//CByteArray* cba = new CByteArray();
		std::vector<BYTE>* cba = new std::vector<BYTE>(*m_value);
		//cba->Copy(*m_value);
		return cba;
	}
}

//CByteArray* Tlv::toByteArray()
std::vector<BYTE>* Tlv::toByteArray()
{
	//CByteArray* buffer = new CByteArray();
	std::vector<BYTE>* buffer = new std::vector<BYTE>();

	switch (m_type)
	{
	case Tlv::Null:
	case Tlv::Terminator:
		buffer->push_back(m_type);
		break;

	default:
	{
		buffer->push_back(m_type);
		USHORT length = (USHORT)this->length();

		if (length <= 0xFE)
		{
			buffer->push_back((BYTE)length);
		}
		else
		{
			buffer->push_back((BYTE)0xFF);
			buffer->push_back((BYTE)(length >> 8));
			buffer->push_back((BYTE)(length & 0xFF));
		}

		if (this->length() > 0)
			//buffer->Append(*m_value);
			buffer->insert(buffer->end(), (*m_value).begin(), (*m_value).end());
	}
	break;
	}

	return buffer;
}

//TlvList* Tlv::fromByteArray(CByteArray* data, UINT64 offset)
TlvList* Tlv::fromByteArray(std::vector<BYTE>* data, UINT64 offset)
{
	TlvList* list = new TlvList();
	
	//CByteArray buffer;
	std::vector<BYTE> buffer(data->size() - (int)offset);
	//buffer.SetSize(data->size() - (INT_PTR)offset);

	for (int i = 0; i < data->size() - offset; i++)
	{
		buffer[i] = (*data)[i];
	}

	//CByteArray buffer = data.GetData(data..left(data.GetCount() - offset);
	UINT count = buffer.size();
	UINT index = 0;

	while (count > index)
	{
		BYTE type = buffer[index];
		++index;

		switch (type)
		{
		case Tlv::Null:
			break;

		case Tlv::Terminator:
		{
			list->push_back(Tlv::createTerminatorTlv());
			index = count;
		}
		break;

		default:
		{
			if ((count - index) > 0)
			{
				USHORT length = buffer[index];
				++index;

				if (length > 0xFE)
				{
					length = 0;

					if ((count - index) > 2)
					{
						length = (BYTE(buffer[index]) << 8) | BYTE(buffer[index + 1]);
						index += 2;
					}
				}

				if ((count - index) >= length)
				{
					//CByteArray* tmpArr = new CByteArray();
					std::vector<BYTE>* tmpArr = new std::vector<BYTE>(length);
					//tmpArr->SetSize(length);
					for (int i = 0; i < length; i++)
						(*tmpArr)[i] = buffer[index + i];

					list->push_back(new Tlv(type, tmpArr));
					delete tmpArr;
					//list.push_back(Tlv(type, buffer.mid(index, length)));
					index += length;
				}
				else
				{
					index = count;
				}
			}
		}
		break;
		}
	}

	return list;
}

Tlv* Tlv::createNullTlv()
{
	return new Tlv(Tlv::Null);
}

Tlv* Tlv::createTerminatorTlv()
{
	return new Tlv(Tlv::Terminator);
}

//Tlv* Tlv::createNDEFMessageTlv(CByteArray* msg)
Tlv* Tlv::createNDEFMessageTlv(std::vector<BYTE>* msg)
{
	return new Tlv(Tlv::NDEF, msg);
}

Tlv* Tlv::createLockControlTlv()
{
	//CByteArray ucaLockControl;
	std::vector<BYTE> ucaLockControl;
	ucaLockControl.push_back(0xA0);
	ucaLockControl.push_back(0x10);
	ucaLockControl.push_back(0x44);

	return new Tlv(Tlv::NDEF, &ucaLockControl);
}
