/*
	uemcmd.h: Emulated CCID reader commands for UEM non-CCID reader
	Copyright (C) 2011-2012   Advanced Card Systems Ltd.
	Copyright (C) 2021-2021   MicroEM JSC
	
	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with this library; if not, write to the Free Software Foundation,
	Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#ifndef UEMCMD_H
#define UEMCMD_H

//#define UEM_USE_EXTENDED_ACTIVATION

//#define UEM_USE_CLESCAR_CHAINING

//#define UEM_PAUSE_BETWEEN_CHAINING_COMMANDS 1
//#define UEM_PAUSE_BETWEEN_CHAINING_COMMANDS 5
//#define UEM_PAUSE_BETWEEN_CHAINING_COMMANDS 10

#define UEM_ENABLE_DEBUG_OUTPUT_VERBOSE 1
#define UEM_ENABLE_DEBUG_OUTPUT_LIMITED 2

//#define UEM_ENABLE_DEBUG_OUTPUT UEM_ENABLE_DEBUG_OUTPUT_VERBOSE
//#define UEM_ENABLE_DEBUG_OUTPUT UEM_ENABLE_DEBUG_OUTPUT_LIMITED

//#define UEM_READER_IFSD 5
#define UEM_READER_IFSD 8

RESPONSECODE UEM_CmdPowerOn(unsigned int reader_index, unsigned int *nlength,
	/*@out@*/ BYTE buffer[], int voltage);

RESPONSECODE UEM_CmdPowerOff(unsigned int reader_index);

RESPONSECODE UEM_Transcieve(unsigned int reader_index,
	/*@out@*/ BYTE out_buffer[], unsigned int out_length,
	/*@in@*/ BYTE in_buffer[],
	ULONG * in_length
	);
	
RESPONSECODE UEM_TranscieveOld(unsigned int reader_index,
	/*@out@*/ BYTE out_buffer[], unsigned int out_length,
	/*@in@*/ BYTE in_buffer[], unsigned int in_length,
	unsigned int * in_actual_length
	);
	
RESPONSECODE UEM_CmdGetSlotStatus(unsigned int reader_index,
	/*@out@*/ BYTE buffer[]);

RESPONSECODE UEM_CmdXfrBlock(unsigned int reader_index, unsigned int tx_length,
	BYTE tx_buffer[], unsigned int *rx_length,
	BYTE rx_buffer[], int protoccol);

RESPONSECODE UEM_TransmitT0(unsigned int reader_index, unsigned int tx_length,
	const BYTE tx_buffer[], unsigned short rx_length, BYTE bBWI);

RESPONSECODE UEM_TransmitT1(unsigned int reader_index, unsigned int tx_length,
	const BYTE tx_buffer[], unsigned short rx_length, BYTE bBWI);

RESPONSECODE UEM_TransmitPPS(unsigned int reader_index, unsigned int tx_length,
	const BYTE tx_buffer[], unsigned short rx_length, BYTE bBWI);

RESPONSECODE UEM_Receive(unsigned int reader_index,
	/*@out@*/ unsigned int *rx_length,
	/*@out@*/ BYTE rx_buffer[], BYTE *chain_parameter);

RESPONSECODE UEM_SetParameters(unsigned int reader_index, char protocol,
	unsigned int length, BYTE buffer[]);

RESPONSECODE UEM_SetCardVoltage(unsigned int reader_index,
	BYTE TxBuffer[], unsigned int TxLength,
	BYTE RxBuffer[], unsigned int *RxLength);

RESPONSECODE UEM_SetCardType(unsigned int reader_index,
	BYTE TxBuffer[], unsigned int TxLength,
	BYTE RxBuffer[], unsigned int *RxLength);

RESPONSECODE UEM_GetFirmwareVersion(unsigned int reader_index,
	BYTE firmwareVersion[]);

#endif
